// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v6.30.2
// source: hook.proto

package gitalypb

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

// State ...
type ReferenceTransactionHookRequest_State int32

const (
	// PREPARED ...
	ReferenceTransactionHookRequest_PREPARED ReferenceTransactionHookRequest_State = 0 // protolint:disable:this ENUM_FIELD_NAMES_PREFIX ENUM_FIELD_NAMES_ZERO_VALUE_END_WITH
	// COMMITTED ...
	ReferenceTransactionHookRequest_COMMITTED ReferenceTransactionHookRequest_State = 1 // protolint:disable:this ENUM_FIELD_NAMES_PREFIX
	// ABORTED ...
	ReferenceTransactionHookRequest_ABORTED ReferenceTransactionHookRequest_State = 2 // protolint:disable:this ENUM_FIELD_NAMES_PREFIX
	// PREPARING ...
	ReferenceTransactionHookRequest_PREPARING ReferenceTransactionHookRequest_State = 3 // protolint:disable:this ENUM_FIELD_NAMES_PREFIX
)

// Enum value maps for ReferenceTransactionHookRequest_State.
var (
	ReferenceTransactionHookRequest_State_name = map[int32]string{
		0: "PREPARED",
		1: "COMMITTED",
		2: "ABORTED",
		3: "PREPARING",
	}
	ReferenceTransactionHookRequest_State_value = map[string]int32{
		"PREPARED":  0,
		"COMMITTED": 1,
		"ABORTED":   2,
		"PREPARING": 3,
	}
)

func (x ReferenceTransactionHookRequest_State) Enum() *ReferenceTransactionHookRequest_State {
	p := new(ReferenceTransactionHookRequest_State)
	*p = x
	return p
}

func (x ReferenceTransactionHookRequest_State) String() string {
	return protoimpl.X.EnumStringOf(x.Descriptor(), protoreflect.EnumNumber(x))
}

func (ReferenceTransactionHookRequest_State) Descriptor() protoreflect.EnumDescriptor {
	return file_hook_proto_enumTypes[0].Descriptor()
}

func (ReferenceTransactionHookRequest_State) Type() protoreflect.EnumType {
	return &file_hook_proto_enumTypes[0]
}

func (x ReferenceTransactionHookRequest_State) Number() protoreflect.EnumNumber {
	return protoreflect.EnumNumber(x)
}

// Deprecated: Use ReferenceTransactionHookRequest_State.Descriptor instead.
func (ReferenceTransactionHookRequest_State) EnumDescriptor() ([]byte, []int) {
	return file_hook_proto_rawDescGZIP(), []int{6, 0}
}

// PreReceiveHookRequest ...
type PreReceiveHookRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// repository ...
	Repository *Repository `protobuf:"bytes,1,opt,name=repository,proto3" json:"repository,omitempty"`
	// environment_variables ...
	EnvironmentVariables []string `protobuf:"bytes,2,rep,name=environment_variables,json=environmentVariables,proto3" json:"environment_variables,omitempty"`
	// stdin ...
	Stdin []byte `protobuf:"bytes,4,opt,name=stdin,proto3" json:"stdin,omitempty"`
	// git_push_options ...
	GitPushOptions []string `protobuf:"bytes,5,rep,name=git_push_options,json=gitPushOptions,proto3" json:"git_push_options,omitempty"`
	unknownFields  protoimpl.UnknownFields
	sizeCache      protoimpl.SizeCache
}

func (x *PreReceiveHookRequest) Reset() {
	*x = PreReceiveHookRequest{}
	mi := &file_hook_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *PreReceiveHookRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*PreReceiveHookRequest) ProtoMessage() {}

func (x *PreReceiveHookRequest) ProtoReflect() protoreflect.Message {
	mi := &file_hook_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use PreReceiveHookRequest.ProtoReflect.Descriptor instead.
func (*PreReceiveHookRequest) Descriptor() ([]byte, []int) {
	return file_hook_proto_rawDescGZIP(), []int{0}
}

func (x *PreReceiveHookRequest) GetRepository() *Repository {
	if x != nil {
		return x.Repository
	}
	return nil
}

func (x *PreReceiveHookRequest) GetEnvironmentVariables() []string {
	if x != nil {
		return x.EnvironmentVariables
	}
	return nil
}

func (x *PreReceiveHookRequest) GetStdin() []byte {
	if x != nil {
		return x.Stdin
	}
	return nil
}

func (x *PreReceiveHookRequest) GetGitPushOptions() []string {
	if x != nil {
		return x.GitPushOptions
	}
	return nil
}

// PreReceiveHookResponse ...
type PreReceiveHookResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// stdout ...
	Stdout []byte `protobuf:"bytes,1,opt,name=stdout,proto3" json:"stdout,omitempty"`
	// stderr ...
	Stderr []byte `protobuf:"bytes,2,opt,name=stderr,proto3" json:"stderr,omitempty"`
	// exit_status ...
	ExitStatus    *ExitStatus `protobuf:"bytes,3,opt,name=exit_status,json=exitStatus,proto3" json:"exit_status,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *PreReceiveHookResponse) Reset() {
	*x = PreReceiveHookResponse{}
	mi := &file_hook_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *PreReceiveHookResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*PreReceiveHookResponse) ProtoMessage() {}

func (x *PreReceiveHookResponse) ProtoReflect() protoreflect.Message {
	mi := &file_hook_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use PreReceiveHookResponse.ProtoReflect.Descriptor instead.
func (*PreReceiveHookResponse) Descriptor() ([]byte, []int) {
	return file_hook_proto_rawDescGZIP(), []int{1}
}

func (x *PreReceiveHookResponse) GetStdout() []byte {
	if x != nil {
		return x.Stdout
	}
	return nil
}

func (x *PreReceiveHookResponse) GetStderr() []byte {
	if x != nil {
		return x.Stderr
	}
	return nil
}

func (x *PreReceiveHookResponse) GetExitStatus() *ExitStatus {
	if x != nil {
		return x.ExitStatus
	}
	return nil
}

// PostReceiveHookRequest ...
type PostReceiveHookRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// repository ...
	Repository *Repository `protobuf:"bytes,1,opt,name=repository,proto3" json:"repository,omitempty"`
	// environment_variables ...
	EnvironmentVariables []string `protobuf:"bytes,2,rep,name=environment_variables,json=environmentVariables,proto3" json:"environment_variables,omitempty"`
	// stdin ...
	Stdin []byte `protobuf:"bytes,3,opt,name=stdin,proto3" json:"stdin,omitempty"`
	// git_push_options ...
	GitPushOptions []string `protobuf:"bytes,4,rep,name=git_push_options,json=gitPushOptions,proto3" json:"git_push_options,omitempty"`
	unknownFields  protoimpl.UnknownFields
	sizeCache      protoimpl.SizeCache
}

func (x *PostReceiveHookRequest) Reset() {
	*x = PostReceiveHookRequest{}
	mi := &file_hook_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *PostReceiveHookRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*PostReceiveHookRequest) ProtoMessage() {}

func (x *PostReceiveHookRequest) ProtoReflect() protoreflect.Message {
	mi := &file_hook_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use PostReceiveHookRequest.ProtoReflect.Descriptor instead.
func (*PostReceiveHookRequest) Descriptor() ([]byte, []int) {
	return file_hook_proto_rawDescGZIP(), []int{2}
}

func (x *PostReceiveHookRequest) GetRepository() *Repository {
	if x != nil {
		return x.Repository
	}
	return nil
}

func (x *PostReceiveHookRequest) GetEnvironmentVariables() []string {
	if x != nil {
		return x.EnvironmentVariables
	}
	return nil
}

func (x *PostReceiveHookRequest) GetStdin() []byte {
	if x != nil {
		return x.Stdin
	}
	return nil
}

func (x *PostReceiveHookRequest) GetGitPushOptions() []string {
	if x != nil {
		return x.GitPushOptions
	}
	return nil
}

// PostReceiveHookResponse ...
type PostReceiveHookResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// stdout ...
	Stdout []byte `protobuf:"bytes,1,opt,name=stdout,proto3" json:"stdout,omitempty"`
	// stderr ...
	Stderr []byte `protobuf:"bytes,2,opt,name=stderr,proto3" json:"stderr,omitempty"`
	// exit_status ...
	ExitStatus    *ExitStatus `protobuf:"bytes,3,opt,name=exit_status,json=exitStatus,proto3" json:"exit_status,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *PostReceiveHookResponse) Reset() {
	*x = PostReceiveHookResponse{}
	mi := &file_hook_proto_msgTypes[3]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *PostReceiveHookResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*PostReceiveHookResponse) ProtoMessage() {}

func (x *PostReceiveHookResponse) ProtoReflect() protoreflect.Message {
	mi := &file_hook_proto_msgTypes[3]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use PostReceiveHookResponse.ProtoReflect.Descriptor instead.
func (*PostReceiveHookResponse) Descriptor() ([]byte, []int) {
	return file_hook_proto_rawDescGZIP(), []int{3}
}

func (x *PostReceiveHookResponse) GetStdout() []byte {
	if x != nil {
		return x.Stdout
	}
	return nil
}

func (x *PostReceiveHookResponse) GetStderr() []byte {
	if x != nil {
		return x.Stderr
	}
	return nil
}

func (x *PostReceiveHookResponse) GetExitStatus() *ExitStatus {
	if x != nil {
		return x.ExitStatus
	}
	return nil
}

// UpdateHookRequest ...
type UpdateHookRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// repository ...
	Repository *Repository `protobuf:"bytes,1,opt,name=repository,proto3" json:"repository,omitempty"`
	// environment_variables ...
	EnvironmentVariables []string `protobuf:"bytes,2,rep,name=environment_variables,json=environmentVariables,proto3" json:"environment_variables,omitempty"`
	// ref ...
	Ref []byte `protobuf:"bytes,3,opt,name=ref,proto3" json:"ref,omitempty"`
	// old_value ...
	OldValue string `protobuf:"bytes,4,opt,name=old_value,json=oldValue,proto3" json:"old_value,omitempty"`
	// new_value ...
	NewValue      string `protobuf:"bytes,5,opt,name=new_value,json=newValue,proto3" json:"new_value,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *UpdateHookRequest) Reset() {
	*x = UpdateHookRequest{}
	mi := &file_hook_proto_msgTypes[4]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *UpdateHookRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*UpdateHookRequest) ProtoMessage() {}

func (x *UpdateHookRequest) ProtoReflect() protoreflect.Message {
	mi := &file_hook_proto_msgTypes[4]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use UpdateHookRequest.ProtoReflect.Descriptor instead.
func (*UpdateHookRequest) Descriptor() ([]byte, []int) {
	return file_hook_proto_rawDescGZIP(), []int{4}
}

func (x *UpdateHookRequest) GetRepository() *Repository {
	if x != nil {
		return x.Repository
	}
	return nil
}

func (x *UpdateHookRequest) GetEnvironmentVariables() []string {
	if x != nil {
		return x.EnvironmentVariables
	}
	return nil
}

func (x *UpdateHookRequest) GetRef() []byte {
	if x != nil {
		return x.Ref
	}
	return nil
}

func (x *UpdateHookRequest) GetOldValue() string {
	if x != nil {
		return x.OldValue
	}
	return ""
}

func (x *UpdateHookRequest) GetNewValue() string {
	if x != nil {
		return x.NewValue
	}
	return ""
}

// UpdateHookResponse ...
type UpdateHookResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// stdout ...
	Stdout []byte `protobuf:"bytes,1,opt,name=stdout,proto3" json:"stdout,omitempty"`
	// stderr ...
	Stderr []byte `protobuf:"bytes,2,opt,name=stderr,proto3" json:"stderr,omitempty"`
	// exit_status ...
	ExitStatus    *ExitStatus `protobuf:"bytes,3,opt,name=exit_status,json=exitStatus,proto3" json:"exit_status,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *UpdateHookResponse) Reset() {
	*x = UpdateHookResponse{}
	mi := &file_hook_proto_msgTypes[5]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *UpdateHookResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*UpdateHookResponse) ProtoMessage() {}

func (x *UpdateHookResponse) ProtoReflect() protoreflect.Message {
	mi := &file_hook_proto_msgTypes[5]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use UpdateHookResponse.ProtoReflect.Descriptor instead.
func (*UpdateHookResponse) Descriptor() ([]byte, []int) {
	return file_hook_proto_rawDescGZIP(), []int{5}
}

func (x *UpdateHookResponse) GetStdout() []byte {
	if x != nil {
		return x.Stdout
	}
	return nil
}

func (x *UpdateHookResponse) GetStderr() []byte {
	if x != nil {
		return x.Stderr
	}
	return nil
}

func (x *UpdateHookResponse) GetExitStatus() *ExitStatus {
	if x != nil {
		return x.ExitStatus
	}
	return nil
}

// ReferenceTransactionHookRequest ...
type ReferenceTransactionHookRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// repository ...
	Repository *Repository `protobuf:"bytes,1,opt,name=repository,proto3" json:"repository,omitempty"`
	// environment_variables ...
	EnvironmentVariables []string `protobuf:"bytes,2,rep,name=environment_variables,json=environmentVariables,proto3" json:"environment_variables,omitempty"`
	// stdin ...
	Stdin []byte `protobuf:"bytes,3,opt,name=stdin,proto3" json:"stdin,omitempty"`
	// state ...
	State         ReferenceTransactionHookRequest_State `protobuf:"varint,4,opt,name=state,proto3,enum=gitaly.ReferenceTransactionHookRequest_State" json:"state,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ReferenceTransactionHookRequest) Reset() {
	*x = ReferenceTransactionHookRequest{}
	mi := &file_hook_proto_msgTypes[6]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ReferenceTransactionHookRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ReferenceTransactionHookRequest) ProtoMessage() {}

func (x *ReferenceTransactionHookRequest) ProtoReflect() protoreflect.Message {
	mi := &file_hook_proto_msgTypes[6]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ReferenceTransactionHookRequest.ProtoReflect.Descriptor instead.
func (*ReferenceTransactionHookRequest) Descriptor() ([]byte, []int) {
	return file_hook_proto_rawDescGZIP(), []int{6}
}

func (x *ReferenceTransactionHookRequest) GetRepository() *Repository {
	if x != nil {
		return x.Repository
	}
	return nil
}

func (x *ReferenceTransactionHookRequest) GetEnvironmentVariables() []string {
	if x != nil {
		return x.EnvironmentVariables
	}
	return nil
}

func (x *ReferenceTransactionHookRequest) GetStdin() []byte {
	if x != nil {
		return x.Stdin
	}
	return nil
}

func (x *ReferenceTransactionHookRequest) GetState() ReferenceTransactionHookRequest_State {
	if x != nil {
		return x.State
	}
	return ReferenceTransactionHookRequest_PREPARED
}

// ReferenceTransactionHookResponse ...
type ReferenceTransactionHookResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// stdout ...
	Stdout []byte `protobuf:"bytes,1,opt,name=stdout,proto3" json:"stdout,omitempty"`
	// stderr ...
	Stderr []byte `protobuf:"bytes,2,opt,name=stderr,proto3" json:"stderr,omitempty"`
	// exit_status ...
	ExitStatus    *ExitStatus `protobuf:"bytes,3,opt,name=exit_status,json=exitStatus,proto3" json:"exit_status,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ReferenceTransactionHookResponse) Reset() {
	*x = ReferenceTransactionHookResponse{}
	mi := &file_hook_proto_msgTypes[7]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ReferenceTransactionHookResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ReferenceTransactionHookResponse) ProtoMessage() {}

func (x *ReferenceTransactionHookResponse) ProtoReflect() protoreflect.Message {
	mi := &file_hook_proto_msgTypes[7]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ReferenceTransactionHookResponse.ProtoReflect.Descriptor instead.
func (*ReferenceTransactionHookResponse) Descriptor() ([]byte, []int) {
	return file_hook_proto_rawDescGZIP(), []int{7}
}

func (x *ReferenceTransactionHookResponse) GetStdout() []byte {
	if x != nil {
		return x.Stdout
	}
	return nil
}

func (x *ReferenceTransactionHookResponse) GetStderr() []byte {
	if x != nil {
		return x.Stderr
	}
	return nil
}

func (x *ReferenceTransactionHookResponse) GetExitStatus() *ExitStatus {
	if x != nil {
		return x.ExitStatus
	}
	return nil
}

// PackObjectsHookWithSidechannelRequest ...
type PackObjectsHookWithSidechannelRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// repository ...
	Repository *Repository `protobuf:"bytes,1,opt,name=repository,proto3" json:"repository,omitempty"`
	// args contains the arguments passed to the pack-objects hook, without the leading "git"
	Args []string `protobuf:"bytes,2,rep,name=args,proto3" json:"args,omitempty"`
	// gl_id is the user id of the initator of the fetch
	GlId string `protobuf:"bytes,3,opt,name=gl_id,json=glId,proto3" json:"gl_id,omitempty"`
	// gl_username is the username of the initator of the fetch
	GlUsername string `protobuf:"bytes,5,opt,name=gl_username,json=glUsername,proto3" json:"gl_username,omitempty"`
	// git_protocol is the protocol used for the fetch
	GitProtocol string `protobuf:"bytes,6,opt,name=git_protocol,json=gitProtocol,proto3" json:"git_protocol,omitempty"`
	// remote_ip is the original IP of the initator of the fetch. This field is used as the key for
	// concurrency limiting before we trigger git-pack-objects command
	RemoteIp string `protobuf:"bytes,7,opt,name=remote_ip,json=remoteIp,proto3" json:"remote_ip,omitempty"`
	// environment_variables contains the environment variables passed to the hook invocation.
	EnvironmentVariables []string `protobuf:"bytes,8,rep,name=environment_variables,json=environmentVariables,proto3" json:"environment_variables,omitempty"`
	unknownFields        protoimpl.UnknownFields
	sizeCache            protoimpl.SizeCache
}

func (x *PackObjectsHookWithSidechannelRequest) Reset() {
	*x = PackObjectsHookWithSidechannelRequest{}
	mi := &file_hook_proto_msgTypes[8]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *PackObjectsHookWithSidechannelRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*PackObjectsHookWithSidechannelRequest) ProtoMessage() {}

func (x *PackObjectsHookWithSidechannelRequest) ProtoReflect() protoreflect.Message {
	mi := &file_hook_proto_msgTypes[8]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use PackObjectsHookWithSidechannelRequest.ProtoReflect.Descriptor instead.
func (*PackObjectsHookWithSidechannelRequest) Descriptor() ([]byte, []int) {
	return file_hook_proto_rawDescGZIP(), []int{8}
}

func (x *PackObjectsHookWithSidechannelRequest) GetRepository() *Repository {
	if x != nil {
		return x.Repository
	}
	return nil
}

func (x *PackObjectsHookWithSidechannelRequest) GetArgs() []string {
	if x != nil {
		return x.Args
	}
	return nil
}

func (x *PackObjectsHookWithSidechannelRequest) GetGlId() string {
	if x != nil {
		return x.GlId
	}
	return ""
}

func (x *PackObjectsHookWithSidechannelRequest) GetGlUsername() string {
	if x != nil {
		return x.GlUsername
	}
	return ""
}

func (x *PackObjectsHookWithSidechannelRequest) GetGitProtocol() string {
	if x != nil {
		return x.GitProtocol
	}
	return ""
}

func (x *PackObjectsHookWithSidechannelRequest) GetRemoteIp() string {
	if x != nil {
		return x.RemoteIp
	}
	return ""
}

func (x *PackObjectsHookWithSidechannelRequest) GetEnvironmentVariables() []string {
	if x != nil {
		return x.EnvironmentVariables
	}
	return nil
}

// PackObjectsHookWithSidechannelResponse ...
type PackObjectsHookWithSidechannelResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *PackObjectsHookWithSidechannelResponse) Reset() {
	*x = PackObjectsHookWithSidechannelResponse{}
	mi := &file_hook_proto_msgTypes[9]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *PackObjectsHookWithSidechannelResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*PackObjectsHookWithSidechannelResponse) ProtoMessage() {}

func (x *PackObjectsHookWithSidechannelResponse) ProtoReflect() protoreflect.Message {
	mi := &file_hook_proto_msgTypes[9]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use PackObjectsHookWithSidechannelResponse.ProtoReflect.Descriptor instead.
func (*PackObjectsHookWithSidechannelResponse) Descriptor() ([]byte, []int) {
	return file_hook_proto_rawDescGZIP(), []int{9}
}

// ProcReceiveHookRequest is the request for the ProcReceiveHook RPC.
type ProcReceiveHookRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// repository is the repository the hook operates on. This field is unused
	// currently since we gather all information from stdin.
	Repository *Repository `protobuf:"bytes,1,opt,name=repository,proto3" json:"repository,omitempty"`
	// environment_variables is the set of env variables to be passed to the RPC.
	EnvironmentVariables []string `protobuf:"bytes,2,rep,name=environment_variables,json=environmentVariables,proto3" json:"environment_variables,omitempty"`
	// stdin is a chunk of raw data which the hook receives.
	Stdin         []byte `protobuf:"bytes,3,opt,name=stdin,proto3" json:"stdin,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ProcReceiveHookRequest) Reset() {
	*x = ProcReceiveHookRequest{}
	mi := &file_hook_proto_msgTypes[10]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ProcReceiveHookRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ProcReceiveHookRequest) ProtoMessage() {}

func (x *ProcReceiveHookRequest) ProtoReflect() protoreflect.Message {
	mi := &file_hook_proto_msgTypes[10]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ProcReceiveHookRequest.ProtoReflect.Descriptor instead.
func (*ProcReceiveHookRequest) Descriptor() ([]byte, []int) {
	return file_hook_proto_rawDescGZIP(), []int{10}
}

func (x *ProcReceiveHookRequest) GetRepository() *Repository {
	if x != nil {
		return x.Repository
	}
	return nil
}

func (x *ProcReceiveHookRequest) GetEnvironmentVariables() []string {
	if x != nil {
		return x.EnvironmentVariables
	}
	return nil
}

func (x *ProcReceiveHookRequest) GetStdin() []byte {
	if x != nil {
		return x.Stdin
	}
	return nil
}

// ProcReceiveHookResponse is the response for the ProcReceiveHook RPC.
type ProcReceiveHookResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// stdout is a chunk of raw data to be output to stdout of the hook.
	Stdout []byte `protobuf:"bytes,1,opt,name=stdout,proto3" json:"stdout,omitempty"`
	// stderr is a chunk of raw data to be output to stderr of the hook.
	Stderr []byte `protobuf:"bytes,2,opt,name=stderr,proto3" json:"stderr,omitempty"`
	// exit_status is the exit status that the hooks outputs, this dictates the
	// success/failure of git-receive-pack(1).
	ExitStatus    *ExitStatus `protobuf:"bytes,3,opt,name=exit_status,json=exitStatus,proto3" json:"exit_status,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ProcReceiveHookResponse) Reset() {
	*x = ProcReceiveHookResponse{}
	mi := &file_hook_proto_msgTypes[11]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ProcReceiveHookResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ProcReceiveHookResponse) ProtoMessage() {}

func (x *ProcReceiveHookResponse) ProtoReflect() protoreflect.Message {
	mi := &file_hook_proto_msgTypes[11]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ProcReceiveHookResponse.ProtoReflect.Descriptor instead.
func (*ProcReceiveHookResponse) Descriptor() ([]byte, []int) {
	return file_hook_proto_rawDescGZIP(), []int{11}
}

func (x *ProcReceiveHookResponse) GetStdout() []byte {
	if x != nil {
		return x.Stdout
	}
	return nil
}

func (x *ProcReceiveHookResponse) GetStderr() []byte {
	if x != nil {
		return x.Stderr
	}
	return nil
}

func (x *ProcReceiveHookResponse) GetExitStatus() *ExitStatus {
	if x != nil {
		return x.ExitStatus
	}
	return nil
}

var File_hook_proto protoreflect.FileDescriptor

const file_hook_proto_rawDesc = "" +
	"\n" +
	"\n" +
	"hook.proto\x12\x06gitaly\x1a\n" +
	"lint.proto\x1a\fshared.proto\"\xc6\x01\n" +
	"\x15PreReceiveHookRequest\x128\n" +
	"\n" +
	"repository\x18\x01 \x01(\v2\x12.gitaly.RepositoryB\x04\x98\xc6,\x01R\n" +
	"repository\x123\n" +
	"\x15environment_variables\x18\x02 \x03(\tR\x14environmentVariables\x12\x14\n" +
	"\x05stdin\x18\x04 \x01(\fR\x05stdin\x12(\n" +
	"\x10git_push_options\x18\x05 \x03(\tR\x0egitPushOptions\"}\n" +
	"\x16PreReceiveHookResponse\x12\x16\n" +
	"\x06stdout\x18\x01 \x01(\fR\x06stdout\x12\x16\n" +
	"\x06stderr\x18\x02 \x01(\fR\x06stderr\x123\n" +
	"\vexit_status\x18\x03 \x01(\v2\x12.gitaly.ExitStatusR\n" +
	"exitStatus\"\xc7\x01\n" +
	"\x16PostReceiveHookRequest\x128\n" +
	"\n" +
	"repository\x18\x01 \x01(\v2\x12.gitaly.RepositoryB\x04\x98\xc6,\x01R\n" +
	"repository\x123\n" +
	"\x15environment_variables\x18\x02 \x03(\tR\x14environmentVariables\x12\x14\n" +
	"\x05stdin\x18\x03 \x01(\fR\x05stdin\x12(\n" +
	"\x10git_push_options\x18\x04 \x03(\tR\x0egitPushOptions\"~\n" +
	"\x17PostReceiveHookResponse\x12\x16\n" +
	"\x06stdout\x18\x01 \x01(\fR\x06stdout\x12\x16\n" +
	"\x06stderr\x18\x02 \x01(\fR\x06stderr\x123\n" +
	"\vexit_status\x18\x03 \x01(\v2\x12.gitaly.ExitStatusR\n" +
	"exitStatus\"\xce\x01\n" +
	"\x11UpdateHookRequest\x128\n" +
	"\n" +
	"repository\x18\x01 \x01(\v2\x12.gitaly.RepositoryB\x04\x98\xc6,\x01R\n" +
	"repository\x123\n" +
	"\x15environment_variables\x18\x02 \x03(\tR\x14environmentVariables\x12\x10\n" +
	"\x03ref\x18\x03 \x01(\fR\x03ref\x12\x1b\n" +
	"\told_value\x18\x04 \x01(\tR\boldValue\x12\x1b\n" +
	"\tnew_value\x18\x05 \x01(\tR\bnewValue\"y\n" +
	"\x12UpdateHookResponse\x12\x16\n" +
	"\x06stdout\x18\x01 \x01(\fR\x06stdout\x12\x16\n" +
	"\x06stderr\x18\x02 \x01(\fR\x06stderr\x123\n" +
	"\vexit_status\x18\x03 \x01(\v2\x12.gitaly.ExitStatusR\n" +
	"exitStatus\"\xad\x02\n" +
	"\x1fReferenceTransactionHookRequest\x128\n" +
	"\n" +
	"repository\x18\x01 \x01(\v2\x12.gitaly.RepositoryB\x04\x98\xc6,\x01R\n" +
	"repository\x123\n" +
	"\x15environment_variables\x18\x02 \x03(\tR\x14environmentVariables\x12\x14\n" +
	"\x05stdin\x18\x03 \x01(\fR\x05stdin\x12C\n" +
	"\x05state\x18\x04 \x01(\x0e2-.gitaly.ReferenceTransactionHookRequest.StateR\x05state\"@\n" +
	"\x05State\x12\f\n" +
	"\bPREPARED\x10\x00\x12\r\n" +
	"\tCOMMITTED\x10\x01\x12\v\n" +
	"\aABORTED\x10\x02\x12\r\n" +
	"\tPREPARING\x10\x03\"\x87\x01\n" +
	" ReferenceTransactionHookResponse\x12\x16\n" +
	"\x06stdout\x18\x01 \x01(\fR\x06stdout\x12\x16\n" +
	"\x06stderr\x18\x02 \x01(\fR\x06stderr\x123\n" +
	"\vexit_status\x18\x03 \x01(\v2\x12.gitaly.ExitStatusR\n" +
	"exitStatus\"\xa0\x02\n" +
	"%PackObjectsHookWithSidechannelRequest\x128\n" +
	"\n" +
	"repository\x18\x01 \x01(\v2\x12.gitaly.RepositoryB\x04\x98\xc6,\x01R\n" +
	"repository\x12\x12\n" +
	"\x04args\x18\x02 \x03(\tR\x04args\x12\x13\n" +
	"\x05gl_id\x18\x03 \x01(\tR\x04glId\x12\x1f\n" +
	"\vgl_username\x18\x05 \x01(\tR\n" +
	"glUsername\x12!\n" +
	"\fgit_protocol\x18\x06 \x01(\tR\vgitProtocol\x12\x1b\n" +
	"\tremote_ip\x18\a \x01(\tR\bremoteIp\x123\n" +
	"\x15environment_variables\x18\b \x03(\tR\x14environmentVariables\"(\n" +
	"&PackObjectsHookWithSidechannelResponse\"\x9d\x01\n" +
	"\x16ProcReceiveHookRequest\x128\n" +
	"\n" +
	"repository\x18\x01 \x01(\v2\x12.gitaly.RepositoryB\x04\x98\xc6,\x01R\n" +
	"repository\x123\n" +
	"\x15environment_variables\x18\x02 \x03(\tR\x14environmentVariables\x12\x14\n" +
	"\x05stdin\x18\x03 \x01(\fR\x05stdin\"~\n" +
	"\x17ProcReceiveHookResponse\x12\x16\n" +
	"\x06stdout\x18\x01 \x01(\fR\x06stdout\x12\x16\n" +
	"\x06stderr\x18\x02 \x01(\fR\x06stderr\x123\n" +
	"\vexit_status\x18\x03 \x01(\v2\x12.gitaly.ExitStatusR\n" +
	"exitStatus2\xfe\x04\n" +
	"\vHookService\x12[\n" +
	"\x0ePreReceiveHook\x12\x1d.gitaly.PreReceiveHookRequest\x1a\x1e.gitaly.PreReceiveHookResponse\"\x06\xfa\x97(\x02\b\x02(\x010\x01\x12^\n" +
	"\x0fPostReceiveHook\x12\x1e.gitaly.PostReceiveHookRequest\x1a\x1f.gitaly.PostReceiveHookResponse\"\x06\xfa\x97(\x02\b\x02(\x010\x01\x12M\n" +
	"\n" +
	"UpdateHook\x12\x19.gitaly.UpdateHookRequest\x1a\x1a.gitaly.UpdateHookResponse\"\x06\xfa\x97(\x02\b\x020\x01\x12y\n" +
	"\x18ReferenceTransactionHook\x12'.gitaly.ReferenceTransactionHookRequest\x1a(.gitaly.ReferenceTransactionHookResponse\"\x06\xfa\x97(\x02\b\x02(\x010\x01\x12\x87\x01\n" +
	"\x1ePackObjectsHookWithSidechannel\x12-.gitaly.PackObjectsHookWithSidechannelRequest\x1a..gitaly.PackObjectsHookWithSidechannelResponse\"\x06\xfa\x97(\x02\b\x02\x12^\n" +
	"\x0fProcReceiveHook\x12\x1e.gitaly.ProcReceiveHookRequest\x1a\x1f.gitaly.ProcReceiveHookResponse\"\x06\xfa\x97(\x02\b\x01(\x010\x01B4Z2gitlab.com/gitlab-org/gitaly/v18/proto/go/gitalypbb\x06proto3"

var (
	file_hook_proto_rawDescOnce sync.Once
	file_hook_proto_rawDescData []byte
)

func file_hook_proto_rawDescGZIP() []byte {
	file_hook_proto_rawDescOnce.Do(func() {
		file_hook_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_hook_proto_rawDesc), len(file_hook_proto_rawDesc)))
	})
	return file_hook_proto_rawDescData
}

var file_hook_proto_enumTypes = make([]protoimpl.EnumInfo, 1)
var file_hook_proto_msgTypes = make([]protoimpl.MessageInfo, 12)
var file_hook_proto_goTypes = []any{
	(ReferenceTransactionHookRequest_State)(0),     // 0: gitaly.ReferenceTransactionHookRequest.State
	(*PreReceiveHookRequest)(nil),                  // 1: gitaly.PreReceiveHookRequest
	(*PreReceiveHookResponse)(nil),                 // 2: gitaly.PreReceiveHookResponse
	(*PostReceiveHookRequest)(nil),                 // 3: gitaly.PostReceiveHookRequest
	(*PostReceiveHookResponse)(nil),                // 4: gitaly.PostReceiveHookResponse
	(*UpdateHookRequest)(nil),                      // 5: gitaly.UpdateHookRequest
	(*UpdateHookResponse)(nil),                     // 6: gitaly.UpdateHookResponse
	(*ReferenceTransactionHookRequest)(nil),        // 7: gitaly.ReferenceTransactionHookRequest
	(*ReferenceTransactionHookResponse)(nil),       // 8: gitaly.ReferenceTransactionHookResponse
	(*PackObjectsHookWithSidechannelRequest)(nil),  // 9: gitaly.PackObjectsHookWithSidechannelRequest
	(*PackObjectsHookWithSidechannelResponse)(nil), // 10: gitaly.PackObjectsHookWithSidechannelResponse
	(*ProcReceiveHookRequest)(nil),                 // 11: gitaly.ProcReceiveHookRequest
	(*ProcReceiveHookResponse)(nil),                // 12: gitaly.ProcReceiveHookResponse
	(*Repository)(nil),                             // 13: gitaly.Repository
	(*ExitStatus)(nil),                             // 14: gitaly.ExitStatus
}
var file_hook_proto_depIdxs = []int32{
	13, // 0: gitaly.PreReceiveHookRequest.repository:type_name -> gitaly.Repository
	14, // 1: gitaly.PreReceiveHookResponse.exit_status:type_name -> gitaly.ExitStatus
	13, // 2: gitaly.PostReceiveHookRequest.repository:type_name -> gitaly.Repository
	14, // 3: gitaly.PostReceiveHookResponse.exit_status:type_name -> gitaly.ExitStatus
	13, // 4: gitaly.UpdateHookRequest.repository:type_name -> gitaly.Repository
	14, // 5: gitaly.UpdateHookResponse.exit_status:type_name -> gitaly.ExitStatus
	13, // 6: gitaly.ReferenceTransactionHookRequest.repository:type_name -> gitaly.Repository
	0,  // 7: gitaly.ReferenceTransactionHookRequest.state:type_name -> gitaly.ReferenceTransactionHookRequest.State
	14, // 8: gitaly.ReferenceTransactionHookResponse.exit_status:type_name -> gitaly.ExitStatus
	13, // 9: gitaly.PackObjectsHookWithSidechannelRequest.repository:type_name -> gitaly.Repository
	13, // 10: gitaly.ProcReceiveHookRequest.repository:type_name -> gitaly.Repository
	14, // 11: gitaly.ProcReceiveHookResponse.exit_status:type_name -> gitaly.ExitStatus
	1,  // 12: gitaly.HookService.PreReceiveHook:input_type -> gitaly.PreReceiveHookRequest
	3,  // 13: gitaly.HookService.PostReceiveHook:input_type -> gitaly.PostReceiveHookRequest
	5,  // 14: gitaly.HookService.UpdateHook:input_type -> gitaly.UpdateHookRequest
	7,  // 15: gitaly.HookService.ReferenceTransactionHook:input_type -> gitaly.ReferenceTransactionHookRequest
	9,  // 16: gitaly.HookService.PackObjectsHookWithSidechannel:input_type -> gitaly.PackObjectsHookWithSidechannelRequest
	11, // 17: gitaly.HookService.ProcReceiveHook:input_type -> gitaly.ProcReceiveHookRequest
	2,  // 18: gitaly.HookService.PreReceiveHook:output_type -> gitaly.PreReceiveHookResponse
	4,  // 19: gitaly.HookService.PostReceiveHook:output_type -> gitaly.PostReceiveHookResponse
	6,  // 20: gitaly.HookService.UpdateHook:output_type -> gitaly.UpdateHookResponse
	8,  // 21: gitaly.HookService.ReferenceTransactionHook:output_type -> gitaly.ReferenceTransactionHookResponse
	10, // 22: gitaly.HookService.PackObjectsHookWithSidechannel:output_type -> gitaly.PackObjectsHookWithSidechannelResponse
	12, // 23: gitaly.HookService.ProcReceiveHook:output_type -> gitaly.ProcReceiveHookResponse
	18, // [18:24] is the sub-list for method output_type
	12, // [12:18] is the sub-list for method input_type
	12, // [12:12] is the sub-list for extension type_name
	12, // [12:12] is the sub-list for extension extendee
	0,  // [0:12] is the sub-list for field type_name
}

func init() { file_hook_proto_init() }
func file_hook_proto_init() {
	if File_hook_proto != nil {
		return
	}
	file_lint_proto_init()
	file_shared_proto_init()
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_hook_proto_rawDesc), len(file_hook_proto_rawDesc)),
			NumEnums:      1,
			NumMessages:   12,
			NumExtensions: 0,
			NumServices:   1,
		},
		GoTypes:           file_hook_proto_goTypes,
		DependencyIndexes: file_hook_proto_depIdxs,
		EnumInfos:         file_hook_proto_enumTypes,
		MessageInfos:      file_hook_proto_msgTypes,
	}.Build()
	File_hook_proto = out.File
	file_hook_proto_goTypes = nil
	file_hook_proto_depIdxs = nil
}
