// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v6.30.2
// source: objectpool.proto

package gitalypb

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

// CreateObjectPoolRequest is a request for the CreateObjectPool RPC.
type CreateObjectPoolRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// object_pool is the object pool to create. This field controls where exactly the object pool will
	// be created.
	ObjectPool *ObjectPool `protobuf:"bytes,1,opt,name=object_pool,json=objectPool,proto3" json:"object_pool,omitempty"`
	// origin is the repository from which the object pool shall be created.
	Origin        *Repository `protobuf:"bytes,2,opt,name=origin,proto3" json:"origin,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *CreateObjectPoolRequest) Reset() {
	*x = CreateObjectPoolRequest{}
	mi := &file_objectpool_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *CreateObjectPoolRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*CreateObjectPoolRequest) ProtoMessage() {}

func (x *CreateObjectPoolRequest) ProtoReflect() protoreflect.Message {
	mi := &file_objectpool_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use CreateObjectPoolRequest.ProtoReflect.Descriptor instead.
func (*CreateObjectPoolRequest) Descriptor() ([]byte, []int) {
	return file_objectpool_proto_rawDescGZIP(), []int{0}
}

func (x *CreateObjectPoolRequest) GetObjectPool() *ObjectPool {
	if x != nil {
		return x.ObjectPool
	}
	return nil
}

func (x *CreateObjectPoolRequest) GetOrigin() *Repository {
	if x != nil {
		return x.Origin
	}
	return nil
}

// CreateObjectPoolResponse is a response for the CreateObjectPool RPC.
type CreateObjectPoolResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *CreateObjectPoolResponse) Reset() {
	*x = CreateObjectPoolResponse{}
	mi := &file_objectpool_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *CreateObjectPoolResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*CreateObjectPoolResponse) ProtoMessage() {}

func (x *CreateObjectPoolResponse) ProtoReflect() protoreflect.Message {
	mi := &file_objectpool_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use CreateObjectPoolResponse.ProtoReflect.Descriptor instead.
func (*CreateObjectPoolResponse) Descriptor() ([]byte, []int) {
	return file_objectpool_proto_rawDescGZIP(), []int{1}
}

// DeleteObjectPoolRequest is a request for the DeleteObjectPool RPC.
type DeleteObjectPoolRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// object_pool is the object pool that shall be deleted.
	ObjectPool    *ObjectPool `protobuf:"bytes,1,opt,name=object_pool,json=objectPool,proto3" json:"object_pool,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *DeleteObjectPoolRequest) Reset() {
	*x = DeleteObjectPoolRequest{}
	mi := &file_objectpool_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *DeleteObjectPoolRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*DeleteObjectPoolRequest) ProtoMessage() {}

func (x *DeleteObjectPoolRequest) ProtoReflect() protoreflect.Message {
	mi := &file_objectpool_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use DeleteObjectPoolRequest.ProtoReflect.Descriptor instead.
func (*DeleteObjectPoolRequest) Descriptor() ([]byte, []int) {
	return file_objectpool_proto_rawDescGZIP(), []int{2}
}

func (x *DeleteObjectPoolRequest) GetObjectPool() *ObjectPool {
	if x != nil {
		return x.ObjectPool
	}
	return nil
}

// DeleteObjectPoolResponse is a response for the DeleteObjectPool RPC.
type DeleteObjectPoolResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *DeleteObjectPoolResponse) Reset() {
	*x = DeleteObjectPoolResponse{}
	mi := &file_objectpool_proto_msgTypes[3]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *DeleteObjectPoolResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*DeleteObjectPoolResponse) ProtoMessage() {}

func (x *DeleteObjectPoolResponse) ProtoReflect() protoreflect.Message {
	mi := &file_objectpool_proto_msgTypes[3]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use DeleteObjectPoolResponse.ProtoReflect.Descriptor instead.
func (*DeleteObjectPoolResponse) Descriptor() ([]byte, []int) {
	return file_objectpool_proto_rawDescGZIP(), []int{3}
}

// LinkRepositoryToObjectPoolRequest is a request for the LinkRepositoryToObjectPool RPC.
type LinkRepositoryToObjectPoolRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// object_pool is the object pool to which the repository shall be linked to.
	ObjectPool *ObjectPool `protobuf:"bytes,1,opt,name=object_pool,json=objectPool,proto3" json:"object_pool,omitempty"`
	// repository is the repository that shall be linked to the object pool.
	Repository    *Repository `protobuf:"bytes,2,opt,name=repository,proto3" json:"repository,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *LinkRepositoryToObjectPoolRequest) Reset() {
	*x = LinkRepositoryToObjectPoolRequest{}
	mi := &file_objectpool_proto_msgTypes[4]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *LinkRepositoryToObjectPoolRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*LinkRepositoryToObjectPoolRequest) ProtoMessage() {}

func (x *LinkRepositoryToObjectPoolRequest) ProtoReflect() protoreflect.Message {
	mi := &file_objectpool_proto_msgTypes[4]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use LinkRepositoryToObjectPoolRequest.ProtoReflect.Descriptor instead.
func (*LinkRepositoryToObjectPoolRequest) Descriptor() ([]byte, []int) {
	return file_objectpool_proto_rawDescGZIP(), []int{4}
}

func (x *LinkRepositoryToObjectPoolRequest) GetObjectPool() *ObjectPool {
	if x != nil {
		return x.ObjectPool
	}
	return nil
}

func (x *LinkRepositoryToObjectPoolRequest) GetRepository() *Repository {
	if x != nil {
		return x.Repository
	}
	return nil
}

// LinkRepositoryToObjectPoolResponse is a response for the LinkRepositoryToObjectPool RPC.
type LinkRepositoryToObjectPoolResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *LinkRepositoryToObjectPoolResponse) Reset() {
	*x = LinkRepositoryToObjectPoolResponse{}
	mi := &file_objectpool_proto_msgTypes[5]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *LinkRepositoryToObjectPoolResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*LinkRepositoryToObjectPoolResponse) ProtoMessage() {}

func (x *LinkRepositoryToObjectPoolResponse) ProtoReflect() protoreflect.Message {
	mi := &file_objectpool_proto_msgTypes[5]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use LinkRepositoryToObjectPoolResponse.ProtoReflect.Descriptor instead.
func (*LinkRepositoryToObjectPoolResponse) Descriptor() ([]byte, []int) {
	return file_objectpool_proto_rawDescGZIP(), []int{5}
}

// DisconnectGitAlternatesRequest is a request for the DisconnectGitAlternates RPC.
type DisconnectGitAlternatesRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// repository is th repository that shall be disconnected from its object pool.
	Repository    *Repository `protobuf:"bytes,1,opt,name=repository,proto3" json:"repository,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *DisconnectGitAlternatesRequest) Reset() {
	*x = DisconnectGitAlternatesRequest{}
	mi := &file_objectpool_proto_msgTypes[6]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *DisconnectGitAlternatesRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*DisconnectGitAlternatesRequest) ProtoMessage() {}

func (x *DisconnectGitAlternatesRequest) ProtoReflect() protoreflect.Message {
	mi := &file_objectpool_proto_msgTypes[6]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use DisconnectGitAlternatesRequest.ProtoReflect.Descriptor instead.
func (*DisconnectGitAlternatesRequest) Descriptor() ([]byte, []int) {
	return file_objectpool_proto_rawDescGZIP(), []int{6}
}

func (x *DisconnectGitAlternatesRequest) GetRepository() *Repository {
	if x != nil {
		return x.Repository
	}
	return nil
}

// DisconnectGitAlternatesResponse is a response for the DisconnectGitAlternates RPC.
type DisconnectGitAlternatesResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *DisconnectGitAlternatesResponse) Reset() {
	*x = DisconnectGitAlternatesResponse{}
	mi := &file_objectpool_proto_msgTypes[7]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *DisconnectGitAlternatesResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*DisconnectGitAlternatesResponse) ProtoMessage() {}

func (x *DisconnectGitAlternatesResponse) ProtoReflect() protoreflect.Message {
	mi := &file_objectpool_proto_msgTypes[7]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use DisconnectGitAlternatesResponse.ProtoReflect.Descriptor instead.
func (*DisconnectGitAlternatesResponse) Descriptor() ([]byte, []int) {
	return file_objectpool_proto_rawDescGZIP(), []int{7}
}

// FetchIntoObjectPoolRequest is a request for the FetchIntoObjectPool RPC.
type FetchIntoObjectPoolRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// origin is the repository to fetch changes from.
	Origin *Repository `protobuf:"bytes,1,opt,name=origin,proto3" json:"origin,omitempty"`
	// object_pool is the repository to fetch changes into.
	ObjectPool    *ObjectPool `protobuf:"bytes,2,opt,name=object_pool,json=objectPool,proto3" json:"object_pool,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *FetchIntoObjectPoolRequest) Reset() {
	*x = FetchIntoObjectPoolRequest{}
	mi := &file_objectpool_proto_msgTypes[8]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *FetchIntoObjectPoolRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*FetchIntoObjectPoolRequest) ProtoMessage() {}

func (x *FetchIntoObjectPoolRequest) ProtoReflect() protoreflect.Message {
	mi := &file_objectpool_proto_msgTypes[8]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use FetchIntoObjectPoolRequest.ProtoReflect.Descriptor instead.
func (*FetchIntoObjectPoolRequest) Descriptor() ([]byte, []int) {
	return file_objectpool_proto_rawDescGZIP(), []int{8}
}

func (x *FetchIntoObjectPoolRequest) GetOrigin() *Repository {
	if x != nil {
		return x.Origin
	}
	return nil
}

func (x *FetchIntoObjectPoolRequest) GetObjectPool() *ObjectPool {
	if x != nil {
		return x.ObjectPool
	}
	return nil
}

// FetchIntoObjectPoolResponse is a response for the FetchIntoObjectPool RPC.
type FetchIntoObjectPoolResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *FetchIntoObjectPoolResponse) Reset() {
	*x = FetchIntoObjectPoolResponse{}
	mi := &file_objectpool_proto_msgTypes[9]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *FetchIntoObjectPoolResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*FetchIntoObjectPoolResponse) ProtoMessage() {}

func (x *FetchIntoObjectPoolResponse) ProtoReflect() protoreflect.Message {
	mi := &file_objectpool_proto_msgTypes[9]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use FetchIntoObjectPoolResponse.ProtoReflect.Descriptor instead.
func (*FetchIntoObjectPoolResponse) Descriptor() ([]byte, []int) {
	return file_objectpool_proto_rawDescGZIP(), []int{9}
}

// GetObjectPoolRequest is a request for the GetObjectPool RPC.
type GetObjectPoolRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// repository is the repository for which the object pool shall be retrieved.
	Repository    *Repository `protobuf:"bytes,1,opt,name=repository,proto3" json:"repository,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *GetObjectPoolRequest) Reset() {
	*x = GetObjectPoolRequest{}
	mi := &file_objectpool_proto_msgTypes[10]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *GetObjectPoolRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*GetObjectPoolRequest) ProtoMessage() {}

func (x *GetObjectPoolRequest) ProtoReflect() protoreflect.Message {
	mi := &file_objectpool_proto_msgTypes[10]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use GetObjectPoolRequest.ProtoReflect.Descriptor instead.
func (*GetObjectPoolRequest) Descriptor() ([]byte, []int) {
	return file_objectpool_proto_rawDescGZIP(), []int{10}
}

func (x *GetObjectPoolRequest) GetRepository() *Repository {
	if x != nil {
		return x.Repository
	}
	return nil
}

// GetObjectPoolResponse is a response for the GetObjectPool RPC.
type GetObjectPoolResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// object_pool is the object pool the repository is connected to. If the repository is not
	// connected to any object pool, then this field will be empty.
	ObjectPool    *ObjectPool `protobuf:"bytes,1,opt,name=object_pool,json=objectPool,proto3" json:"object_pool,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *GetObjectPoolResponse) Reset() {
	*x = GetObjectPoolResponse{}
	mi := &file_objectpool_proto_msgTypes[11]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *GetObjectPoolResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*GetObjectPoolResponse) ProtoMessage() {}

func (x *GetObjectPoolResponse) ProtoReflect() protoreflect.Message {
	mi := &file_objectpool_proto_msgTypes[11]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use GetObjectPoolResponse.ProtoReflect.Descriptor instead.
func (*GetObjectPoolResponse) Descriptor() ([]byte, []int) {
	return file_objectpool_proto_rawDescGZIP(), []int{11}
}

func (x *GetObjectPoolResponse) GetObjectPool() *ObjectPool {
	if x != nil {
		return x.ObjectPool
	}
	return nil
}

var File_objectpool_proto protoreflect.FileDescriptor

const file_objectpool_proto_rawDesc = "" +
	"\n" +
	"\x10objectpool.proto\x12\x06gitaly\x1a\n" +
	"lint.proto\x1a\fshared.proto\"\x86\x01\n" +
	"\x17CreateObjectPoolRequest\x129\n" +
	"\vobject_pool\x18\x01 \x01(\v2\x12.gitaly.ObjectPoolB\x04\x98\xc6,\x01R\n" +
	"objectPool\x120\n" +
	"\x06origin\x18\x02 \x01(\v2\x12.gitaly.RepositoryB\x04\xa0\xc6,\x01R\x06origin\"\x1a\n" +
	"\x18CreateObjectPoolResponse\"T\n" +
	"\x17DeleteObjectPoolRequest\x129\n" +
	"\vobject_pool\x18\x01 \x01(\v2\x12.gitaly.ObjectPoolB\x04\x98\xc6,\x01R\n" +
	"objectPool\"\x1a\n" +
	"\x18DeleteObjectPoolResponse\"\x98\x01\n" +
	"!LinkRepositoryToObjectPoolRequest\x129\n" +
	"\vobject_pool\x18\x01 \x01(\v2\x12.gitaly.ObjectPoolB\x04\xa0\xc6,\x01R\n" +
	"objectPool\x128\n" +
	"\n" +
	"repository\x18\x02 \x01(\v2\x12.gitaly.RepositoryB\x04\x98\xc6,\x01R\n" +
	"repository\"$\n" +
	"\"LinkRepositoryToObjectPoolResponse\"Z\n" +
	"\x1eDisconnectGitAlternatesRequest\x128\n" +
	"\n" +
	"repository\x18\x01 \x01(\v2\x12.gitaly.RepositoryB\x04\x98\xc6,\x01R\n" +
	"repository\"!\n" +
	"\x1fDisconnectGitAlternatesResponse\"\x97\x01\n" +
	"\x1aFetchIntoObjectPoolRequest\x120\n" +
	"\x06origin\x18\x01 \x01(\v2\x12.gitaly.RepositoryB\x04\xa0\xc6,\x01R\x06origin\x129\n" +
	"\vobject_pool\x18\x02 \x01(\v2\x12.gitaly.ObjectPoolB\x04\x98\xc6,\x01R\n" +
	"objectPoolJ\x04\b\x03\x10\x04R\x06repack\"\x1d\n" +
	"\x1bFetchIntoObjectPoolResponse\"P\n" +
	"\x14GetObjectPoolRequest\x128\n" +
	"\n" +
	"repository\x18\x01 \x01(\v2\x12.gitaly.RepositoryB\x04\x98\xc6,\x01R\n" +
	"repository\"L\n" +
	"\x15GetObjectPoolResponse\x123\n" +
	"\vobject_pool\x18\x01 \x01(\v2\x12.gitaly.ObjectPoolR\n" +
	"objectPool2\x80\x05\n" +
	"\x11ObjectPoolService\x12]\n" +
	"\x10CreateObjectPool\x12\x1f.gitaly.CreateObjectPoolRequest\x1a .gitaly.CreateObjectPoolResponse\"\x06\xfa\x97(\x02\b\x01\x12]\n" +
	"\x10DeleteObjectPool\x12\x1f.gitaly.DeleteObjectPoolRequest\x1a .gitaly.DeleteObjectPoolResponse\"\x06\xfa\x97(\x02\b\x01\x12{\n" +
	"\x1aLinkRepositoryToObjectPool\x12).gitaly.LinkRepositoryToObjectPoolRequest\x1a*.gitaly.LinkRepositoryToObjectPoolResponse\"\x06\xfa\x97(\x02\b\x01\x12r\n" +
	"\x17DisconnectGitAlternates\x12&.gitaly.DisconnectGitAlternatesRequest\x1a'.gitaly.DisconnectGitAlternatesResponse\"\x06\xfa\x97(\x02\b\x01\x12f\n" +
	"\x13FetchIntoObjectPool\x12\".gitaly.FetchIntoObjectPoolRequest\x1a#.gitaly.FetchIntoObjectPoolResponse\"\x06\xfa\x97(\x02\b\x01\x12T\n" +
	"\rGetObjectPool\x12\x1c.gitaly.GetObjectPoolRequest\x1a\x1d.gitaly.GetObjectPoolResponse\"\x06\xfa\x97(\x02\b\x02B4Z2gitlab.com/gitlab-org/gitaly/v18/proto/go/gitalypbb\x06proto3"

var (
	file_objectpool_proto_rawDescOnce sync.Once
	file_objectpool_proto_rawDescData []byte
)

func file_objectpool_proto_rawDescGZIP() []byte {
	file_objectpool_proto_rawDescOnce.Do(func() {
		file_objectpool_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_objectpool_proto_rawDesc), len(file_objectpool_proto_rawDesc)))
	})
	return file_objectpool_proto_rawDescData
}

var file_objectpool_proto_msgTypes = make([]protoimpl.MessageInfo, 12)
var file_objectpool_proto_goTypes = []any{
	(*CreateObjectPoolRequest)(nil),            // 0: gitaly.CreateObjectPoolRequest
	(*CreateObjectPoolResponse)(nil),           // 1: gitaly.CreateObjectPoolResponse
	(*DeleteObjectPoolRequest)(nil),            // 2: gitaly.DeleteObjectPoolRequest
	(*DeleteObjectPoolResponse)(nil),           // 3: gitaly.DeleteObjectPoolResponse
	(*LinkRepositoryToObjectPoolRequest)(nil),  // 4: gitaly.LinkRepositoryToObjectPoolRequest
	(*LinkRepositoryToObjectPoolResponse)(nil), // 5: gitaly.LinkRepositoryToObjectPoolResponse
	(*DisconnectGitAlternatesRequest)(nil),     // 6: gitaly.DisconnectGitAlternatesRequest
	(*DisconnectGitAlternatesResponse)(nil),    // 7: gitaly.DisconnectGitAlternatesResponse
	(*FetchIntoObjectPoolRequest)(nil),         // 8: gitaly.FetchIntoObjectPoolRequest
	(*FetchIntoObjectPoolResponse)(nil),        // 9: gitaly.FetchIntoObjectPoolResponse
	(*GetObjectPoolRequest)(nil),               // 10: gitaly.GetObjectPoolRequest
	(*GetObjectPoolResponse)(nil),              // 11: gitaly.GetObjectPoolResponse
	(*ObjectPool)(nil),                         // 12: gitaly.ObjectPool
	(*Repository)(nil),                         // 13: gitaly.Repository
}
var file_objectpool_proto_depIdxs = []int32{
	12, // 0: gitaly.CreateObjectPoolRequest.object_pool:type_name -> gitaly.ObjectPool
	13, // 1: gitaly.CreateObjectPoolRequest.origin:type_name -> gitaly.Repository
	12, // 2: gitaly.DeleteObjectPoolRequest.object_pool:type_name -> gitaly.ObjectPool
	12, // 3: gitaly.LinkRepositoryToObjectPoolRequest.object_pool:type_name -> gitaly.ObjectPool
	13, // 4: gitaly.LinkRepositoryToObjectPoolRequest.repository:type_name -> gitaly.Repository
	13, // 5: gitaly.DisconnectGitAlternatesRequest.repository:type_name -> gitaly.Repository
	13, // 6: gitaly.FetchIntoObjectPoolRequest.origin:type_name -> gitaly.Repository
	12, // 7: gitaly.FetchIntoObjectPoolRequest.object_pool:type_name -> gitaly.ObjectPool
	13, // 8: gitaly.GetObjectPoolRequest.repository:type_name -> gitaly.Repository
	12, // 9: gitaly.GetObjectPoolResponse.object_pool:type_name -> gitaly.ObjectPool
	0,  // 10: gitaly.ObjectPoolService.CreateObjectPool:input_type -> gitaly.CreateObjectPoolRequest
	2,  // 11: gitaly.ObjectPoolService.DeleteObjectPool:input_type -> gitaly.DeleteObjectPoolRequest
	4,  // 12: gitaly.ObjectPoolService.LinkRepositoryToObjectPool:input_type -> gitaly.LinkRepositoryToObjectPoolRequest
	6,  // 13: gitaly.ObjectPoolService.DisconnectGitAlternates:input_type -> gitaly.DisconnectGitAlternatesRequest
	8,  // 14: gitaly.ObjectPoolService.FetchIntoObjectPool:input_type -> gitaly.FetchIntoObjectPoolRequest
	10, // 15: gitaly.ObjectPoolService.GetObjectPool:input_type -> gitaly.GetObjectPoolRequest
	1,  // 16: gitaly.ObjectPoolService.CreateObjectPool:output_type -> gitaly.CreateObjectPoolResponse
	3,  // 17: gitaly.ObjectPoolService.DeleteObjectPool:output_type -> gitaly.DeleteObjectPoolResponse
	5,  // 18: gitaly.ObjectPoolService.LinkRepositoryToObjectPool:output_type -> gitaly.LinkRepositoryToObjectPoolResponse
	7,  // 19: gitaly.ObjectPoolService.DisconnectGitAlternates:output_type -> gitaly.DisconnectGitAlternatesResponse
	9,  // 20: gitaly.ObjectPoolService.FetchIntoObjectPool:output_type -> gitaly.FetchIntoObjectPoolResponse
	11, // 21: gitaly.ObjectPoolService.GetObjectPool:output_type -> gitaly.GetObjectPoolResponse
	16, // [16:22] is the sub-list for method output_type
	10, // [10:16] is the sub-list for method input_type
	10, // [10:10] is the sub-list for extension type_name
	10, // [10:10] is the sub-list for extension extendee
	0,  // [0:10] is the sub-list for field type_name
}

func init() { file_objectpool_proto_init() }
func file_objectpool_proto_init() {
	if File_objectpool_proto != nil {
		return
	}
	file_lint_proto_init()
	file_shared_proto_init()
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_objectpool_proto_rawDesc), len(file_objectpool_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   12,
			NumExtensions: 0,
			NumServices:   1,
		},
		GoTypes:           file_objectpool_proto_goTypes,
		DependencyIndexes: file_objectpool_proto_depIdxs,
		MessageInfos:      file_objectpool_proto_msgTypes,
	}.Build()
	File_objectpool_proto = out.File
	file_objectpool_proto_goTypes = nil
	file_objectpool_proto_depIdxs = nil
}
