// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v6.30.2
// source: server.proto

package gitalypb

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	durationpb "google.golang.org/protobuf/types/known/durationpb"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

// ServerInfoRequest ...
type ServerInfoRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ServerInfoRequest) Reset() {
	*x = ServerInfoRequest{}
	mi := &file_server_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ServerInfoRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ServerInfoRequest) ProtoMessage() {}

func (x *ServerInfoRequest) ProtoReflect() protoreflect.Message {
	mi := &file_server_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ServerInfoRequest.ProtoReflect.Descriptor instead.
func (*ServerInfoRequest) Descriptor() ([]byte, []int) {
	return file_server_proto_rawDescGZIP(), []int{0}
}

// ServerInfoResponse ...
type ServerInfoResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// server_version ...
	ServerVersion string `protobuf:"bytes,1,opt,name=server_version,json=serverVersion,proto3" json:"server_version,omitempty"`
	// git_version ...
	GitVersion string `protobuf:"bytes,2,opt,name=git_version,json=gitVersion,proto3" json:"git_version,omitempty"`
	// storage_statuses ...
	StorageStatuses []*ServerInfoResponse_StorageStatus `protobuf:"bytes,3,rep,name=storage_statuses,json=storageStatuses,proto3" json:"storage_statuses,omitempty"`
	unknownFields   protoimpl.UnknownFields
	sizeCache       protoimpl.SizeCache
}

func (x *ServerInfoResponse) Reset() {
	*x = ServerInfoResponse{}
	mi := &file_server_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ServerInfoResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ServerInfoResponse) ProtoMessage() {}

func (x *ServerInfoResponse) ProtoReflect() protoreflect.Message {
	mi := &file_server_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ServerInfoResponse.ProtoReflect.Descriptor instead.
func (*ServerInfoResponse) Descriptor() ([]byte, []int) {
	return file_server_proto_rawDescGZIP(), []int{1}
}

func (x *ServerInfoResponse) GetServerVersion() string {
	if x != nil {
		return x.ServerVersion
	}
	return ""
}

func (x *ServerInfoResponse) GetGitVersion() string {
	if x != nil {
		return x.GitVersion
	}
	return ""
}

func (x *ServerInfoResponse) GetStorageStatuses() []*ServerInfoResponse_StorageStatus {
	if x != nil {
		return x.StorageStatuses
	}
	return nil
}

// DiskStatisticsRequest ...
type DiskStatisticsRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *DiskStatisticsRequest) Reset() {
	*x = DiskStatisticsRequest{}
	mi := &file_server_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *DiskStatisticsRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*DiskStatisticsRequest) ProtoMessage() {}

func (x *DiskStatisticsRequest) ProtoReflect() protoreflect.Message {
	mi := &file_server_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use DiskStatisticsRequest.ProtoReflect.Descriptor instead.
func (*DiskStatisticsRequest) Descriptor() ([]byte, []int) {
	return file_server_proto_rawDescGZIP(), []int{2}
}

// DiskStatisticsResponse ...
type DiskStatisticsResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// storage_statuses ...
	StorageStatuses []*DiskStatisticsResponse_StorageStatus `protobuf:"bytes,1,rep,name=storage_statuses,json=storageStatuses,proto3" json:"storage_statuses,omitempty"`
	unknownFields   protoimpl.UnknownFields
	sizeCache       protoimpl.SizeCache
}

func (x *DiskStatisticsResponse) Reset() {
	*x = DiskStatisticsResponse{}
	mi := &file_server_proto_msgTypes[3]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *DiskStatisticsResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*DiskStatisticsResponse) ProtoMessage() {}

func (x *DiskStatisticsResponse) ProtoReflect() protoreflect.Message {
	mi := &file_server_proto_msgTypes[3]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use DiskStatisticsResponse.ProtoReflect.Descriptor instead.
func (*DiskStatisticsResponse) Descriptor() ([]byte, []int) {
	return file_server_proto_rawDescGZIP(), []int{3}
}

func (x *DiskStatisticsResponse) GetStorageStatuses() []*DiskStatisticsResponse_StorageStatus {
	if x != nil {
		return x.StorageStatuses
	}
	return nil
}

// ReadinessCheckRequest is used to verify if the service is in operational state.
type ReadinessCheckRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// timeout is an amount of milliseconds for the check to run before give up and mark as failed.
	Timeout       *durationpb.Duration `protobuf:"bytes,1,opt,name=timeout,proto3" json:"timeout,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ReadinessCheckRequest) Reset() {
	*x = ReadinessCheckRequest{}
	mi := &file_server_proto_msgTypes[4]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ReadinessCheckRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ReadinessCheckRequest) ProtoMessage() {}

func (x *ReadinessCheckRequest) ProtoReflect() protoreflect.Message {
	mi := &file_server_proto_msgTypes[4]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ReadinessCheckRequest.ProtoReflect.Descriptor instead.
func (*ReadinessCheckRequest) Descriptor() ([]byte, []int) {
	return file_server_proto_rawDescGZIP(), []int{4}
}

func (x *ReadinessCheckRequest) GetTimeout() *durationpb.Duration {
	if x != nil {
		return x.Timeout
	}
	return nil
}

// ReadinessCheckResponse is just a stub now and contains no information.
// If the service is not in the operational state the error will be returned instead.
type ReadinessCheckResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// Types that are valid to be assigned to Result:
	//
	//	*ReadinessCheckResponse_OkResponse
	//	*ReadinessCheckResponse_FailureResponse
	Result        isReadinessCheckResponse_Result `protobuf_oneof:"Result"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ReadinessCheckResponse) Reset() {
	*x = ReadinessCheckResponse{}
	mi := &file_server_proto_msgTypes[5]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ReadinessCheckResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ReadinessCheckResponse) ProtoMessage() {}

func (x *ReadinessCheckResponse) ProtoReflect() protoreflect.Message {
	mi := &file_server_proto_msgTypes[5]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ReadinessCheckResponse.ProtoReflect.Descriptor instead.
func (*ReadinessCheckResponse) Descriptor() ([]byte, []int) {
	return file_server_proto_rawDescGZIP(), []int{5}
}

func (x *ReadinessCheckResponse) GetResult() isReadinessCheckResponse_Result {
	if x != nil {
		return x.Result
	}
	return nil
}

func (x *ReadinessCheckResponse) GetOkResponse() *ReadinessCheckResponse_Ok {
	if x != nil {
		if x, ok := x.Result.(*ReadinessCheckResponse_OkResponse); ok {
			return x.OkResponse
		}
	}
	return nil
}

func (x *ReadinessCheckResponse) GetFailureResponse() *ReadinessCheckResponse_Failure {
	if x != nil {
		if x, ok := x.Result.(*ReadinessCheckResponse_FailureResponse); ok {
			return x.FailureResponse
		}
	}
	return nil
}

type isReadinessCheckResponse_Result interface {
	isReadinessCheckResponse_Result()
}

type ReadinessCheckResponse_OkResponse struct {
	// ok_response is set when all checks pass.
	OkResponse *ReadinessCheckResponse_Ok `protobuf:"bytes,1,opt,name=ok_response,json=okResponse,proto3,oneof"`
}

type ReadinessCheckResponse_FailureResponse struct {
	// failure_response is set if at least one check failed.
	FailureResponse *ReadinessCheckResponse_Failure `protobuf:"bytes,2,opt,name=failure_response,json=failureResponse,proto3,oneof"`
}

func (*ReadinessCheckResponse_OkResponse) isReadinessCheckResponse_Result() {}

func (*ReadinessCheckResponse_FailureResponse) isReadinessCheckResponse_Result() {}

// ServerSignatureRequest is the request for the ServerSignature RPC.
type ServerSignatureRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ServerSignatureRequest) Reset() {
	*x = ServerSignatureRequest{}
	mi := &file_server_proto_msgTypes[6]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ServerSignatureRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ServerSignatureRequest) ProtoMessage() {}

func (x *ServerSignatureRequest) ProtoReflect() protoreflect.Message {
	mi := &file_server_proto_msgTypes[6]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ServerSignatureRequest.ProtoReflect.Descriptor instead.
func (*ServerSignatureRequest) Descriptor() ([]byte, []int) {
	return file_server_proto_rawDescGZIP(), []int{6}
}

// ServerSignatureResponse is the response for the ServerSignature RPC.
type ServerSignatureResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// public_key is the string with a content of the public key.
	PublicKey     []byte `protobuf:"bytes,1,opt,name=public_key,json=publicKey,proto3" json:"public_key,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ServerSignatureResponse) Reset() {
	*x = ServerSignatureResponse{}
	mi := &file_server_proto_msgTypes[7]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ServerSignatureResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ServerSignatureResponse) ProtoMessage() {}

func (x *ServerSignatureResponse) ProtoReflect() protoreflect.Message {
	mi := &file_server_proto_msgTypes[7]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ServerSignatureResponse.ProtoReflect.Descriptor instead.
func (*ServerSignatureResponse) Descriptor() ([]byte, []int) {
	return file_server_proto_rawDescGZIP(), []int{7}
}

func (x *ServerSignatureResponse) GetPublicKey() []byte {
	if x != nil {
		return x.PublicKey
	}
	return nil
}

// StorageStatus ...
type ServerInfoResponse_StorageStatus struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// storage_name ...
	StorageName string `protobuf:"bytes,1,opt,name=storage_name,json=storageName,proto3" json:"storage_name,omitempty"`
	// readable ...
	Readable bool `protobuf:"varint,2,opt,name=readable,proto3" json:"readable,omitempty"`
	// writeable ...
	Writeable bool `protobuf:"varint,3,opt,name=writeable,proto3" json:"writeable,omitempty"`
	// fs_type ...
	FsType string `protobuf:"bytes,4,opt,name=fs_type,json=fsType,proto3" json:"fs_type,omitempty"`
	// filesystem_id ...
	FilesystemId string `protobuf:"bytes,5,opt,name=filesystem_id,json=filesystemId,proto3" json:"filesystem_id,omitempty"`
	// replication_factor ...
	ReplicationFactor uint32 `protobuf:"varint,6,opt,name=replication_factor,json=replicationFactor,proto3" json:"replication_factor,omitempty"`
	unknownFields     protoimpl.UnknownFields
	sizeCache         protoimpl.SizeCache
}

func (x *ServerInfoResponse_StorageStatus) Reset() {
	*x = ServerInfoResponse_StorageStatus{}
	mi := &file_server_proto_msgTypes[8]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ServerInfoResponse_StorageStatus) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ServerInfoResponse_StorageStatus) ProtoMessage() {}

func (x *ServerInfoResponse_StorageStatus) ProtoReflect() protoreflect.Message {
	mi := &file_server_proto_msgTypes[8]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ServerInfoResponse_StorageStatus.ProtoReflect.Descriptor instead.
func (*ServerInfoResponse_StorageStatus) Descriptor() ([]byte, []int) {
	return file_server_proto_rawDescGZIP(), []int{1, 0}
}

func (x *ServerInfoResponse_StorageStatus) GetStorageName() string {
	if x != nil {
		return x.StorageName
	}
	return ""
}

func (x *ServerInfoResponse_StorageStatus) GetReadable() bool {
	if x != nil {
		return x.Readable
	}
	return false
}

func (x *ServerInfoResponse_StorageStatus) GetWriteable() bool {
	if x != nil {
		return x.Writeable
	}
	return false
}

func (x *ServerInfoResponse_StorageStatus) GetFsType() string {
	if x != nil {
		return x.FsType
	}
	return ""
}

func (x *ServerInfoResponse_StorageStatus) GetFilesystemId() string {
	if x != nil {
		return x.FilesystemId
	}
	return ""
}

func (x *ServerInfoResponse_StorageStatus) GetReplicationFactor() uint32 {
	if x != nil {
		return x.ReplicationFactor
	}
	return 0
}

// StorageStatus ...
// When both available and used fields are equal 0 that means that
// Gitaly was unable to determine storage stats.
type DiskStatisticsResponse_StorageStatus struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// storage_name ...
	StorageName string `protobuf:"bytes,1,opt,name=storage_name,json=storageName,proto3" json:"storage_name,omitempty"`
	// available ...
	Available int64 `protobuf:"varint,2,opt,name=available,proto3" json:"available,omitempty"`
	// used ...
	Used          int64 `protobuf:"varint,3,opt,name=used,proto3" json:"used,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *DiskStatisticsResponse_StorageStatus) Reset() {
	*x = DiskStatisticsResponse_StorageStatus{}
	mi := &file_server_proto_msgTypes[9]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *DiskStatisticsResponse_StorageStatus) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*DiskStatisticsResponse_StorageStatus) ProtoMessage() {}

func (x *DiskStatisticsResponse_StorageStatus) ProtoReflect() protoreflect.Message {
	mi := &file_server_proto_msgTypes[9]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use DiskStatisticsResponse_StorageStatus.ProtoReflect.Descriptor instead.
func (*DiskStatisticsResponse_StorageStatus) Descriptor() ([]byte, []int) {
	return file_server_proto_rawDescGZIP(), []int{3, 0}
}

func (x *DiskStatisticsResponse_StorageStatus) GetStorageName() string {
	if x != nil {
		return x.StorageName
	}
	return ""
}

func (x *DiskStatisticsResponse_StorageStatus) GetAvailable() int64 {
	if x != nil {
		return x.Available
	}
	return 0
}

func (x *DiskStatisticsResponse_StorageStatus) GetUsed() int64 {
	if x != nil {
		return x.Used
	}
	return 0
}

// Ok represents response if none checks failed.
type ReadinessCheckResponse_Ok struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ReadinessCheckResponse_Ok) Reset() {
	*x = ReadinessCheckResponse_Ok{}
	mi := &file_server_proto_msgTypes[10]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ReadinessCheckResponse_Ok) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ReadinessCheckResponse_Ok) ProtoMessage() {}

func (x *ReadinessCheckResponse_Ok) ProtoReflect() protoreflect.Message {
	mi := &file_server_proto_msgTypes[10]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ReadinessCheckResponse_Ok.ProtoReflect.Descriptor instead.
func (*ReadinessCheckResponse_Ok) Descriptor() ([]byte, []int) {
	return file_server_proto_rawDescGZIP(), []int{5, 0}
}

// Failure represents response if at least one check failed.
type ReadinessCheckResponse_Failure struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// failed_checks is a list of failed checks.
	FailedChecks  []*ReadinessCheckResponse_Failure_Response `protobuf:"bytes,1,rep,name=failed_checks,json=failedChecks,proto3" json:"failed_checks,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ReadinessCheckResponse_Failure) Reset() {
	*x = ReadinessCheckResponse_Failure{}
	mi := &file_server_proto_msgTypes[11]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ReadinessCheckResponse_Failure) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ReadinessCheckResponse_Failure) ProtoMessage() {}

func (x *ReadinessCheckResponse_Failure) ProtoReflect() protoreflect.Message {
	mi := &file_server_proto_msgTypes[11]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ReadinessCheckResponse_Failure.ProtoReflect.Descriptor instead.
func (*ReadinessCheckResponse_Failure) Descriptor() ([]byte, []int) {
	return file_server_proto_rawDescGZIP(), []int{5, 1}
}

func (x *ReadinessCheckResponse_Failure) GetFailedChecks() []*ReadinessCheckResponse_Failure_Response {
	if x != nil {
		return x.FailedChecks
	}
	return nil
}

// Response contains information about failed check.
type ReadinessCheckResponse_Failure_Response struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// name is a name of the check that was performed.
	Name string `protobuf:"bytes,1,opt,name=name,proto3" json:"name,omitempty"`
	// error_message is a cause of the check failure.
	ErrorMessage  string `protobuf:"bytes,2,opt,name=error_message,json=errorMessage,proto3" json:"error_message,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ReadinessCheckResponse_Failure_Response) Reset() {
	*x = ReadinessCheckResponse_Failure_Response{}
	mi := &file_server_proto_msgTypes[12]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ReadinessCheckResponse_Failure_Response) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ReadinessCheckResponse_Failure_Response) ProtoMessage() {}

func (x *ReadinessCheckResponse_Failure_Response) ProtoReflect() protoreflect.Message {
	mi := &file_server_proto_msgTypes[12]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ReadinessCheckResponse_Failure_Response.ProtoReflect.Descriptor instead.
func (*ReadinessCheckResponse_Failure_Response) Descriptor() ([]byte, []int) {
	return file_server_proto_rawDescGZIP(), []int{5, 1, 0}
}

func (x *ReadinessCheckResponse_Failure_Response) GetName() string {
	if x != nil {
		return x.Name
	}
	return ""
}

func (x *ReadinessCheckResponse_Failure_Response) GetErrorMessage() string {
	if x != nil {
		return x.ErrorMessage
	}
	return ""
}

var File_server_proto protoreflect.FileDescriptor

const file_server_proto_rawDesc = "" +
	"\n" +
	"\fserver.proto\x12\x06gitaly\x1a\x1egoogle/protobuf/duration.proto\x1a\n" +
	"lint.proto\"\x13\n" +
	"\x11ServerInfoRequest\"\x8d\x03\n" +
	"\x12ServerInfoResponse\x12%\n" +
	"\x0eserver_version\x18\x01 \x01(\tR\rserverVersion\x12\x1f\n" +
	"\vgit_version\x18\x02 \x01(\tR\n" +
	"gitVersion\x12S\n" +
	"\x10storage_statuses\x18\x03 \x03(\v2(.gitaly.ServerInfoResponse.StorageStatusR\x0fstorageStatuses\x1a\xd9\x01\n" +
	"\rStorageStatus\x12!\n" +
	"\fstorage_name\x18\x01 \x01(\tR\vstorageName\x12\x1a\n" +
	"\breadable\x18\x02 \x01(\bR\breadable\x12\x1c\n" +
	"\twriteable\x18\x03 \x01(\bR\twriteable\x12\x17\n" +
	"\afs_type\x18\x04 \x01(\tR\x06fsType\x12#\n" +
	"\rfilesystem_id\x18\x05 \x01(\tR\ffilesystemId\x12-\n" +
	"\x12replication_factor\x18\x06 \x01(\rR\x11replicationFactor\"\x17\n" +
	"\x15DiskStatisticsRequest\"\xd7\x01\n" +
	"\x16DiskStatisticsResponse\x12W\n" +
	"\x10storage_statuses\x18\x01 \x03(\v2,.gitaly.DiskStatisticsResponse.StorageStatusR\x0fstorageStatuses\x1ad\n" +
	"\rStorageStatus\x12!\n" +
	"\fstorage_name\x18\x01 \x01(\tR\vstorageName\x12\x1c\n" +
	"\tavailable\x18\x02 \x01(\x03R\tavailable\x12\x12\n" +
	"\x04used\x18\x03 \x01(\x03R\x04used\"L\n" +
	"\x15ReadinessCheckRequest\x123\n" +
	"\atimeout\x18\x01 \x01(\v2\x19.google.protobuf.DurationR\atimeout\"\xea\x02\n" +
	"\x16ReadinessCheckResponse\x12D\n" +
	"\vok_response\x18\x01 \x01(\v2!.gitaly.ReadinessCheckResponse.OkH\x00R\n" +
	"okResponse\x12S\n" +
	"\x10failure_response\x18\x02 \x01(\v2&.gitaly.ReadinessCheckResponse.FailureH\x00R\x0ffailureResponse\x1a\x04\n" +
	"\x02Ok\x1a\xa4\x01\n" +
	"\aFailure\x12T\n" +
	"\rfailed_checks\x18\x01 \x03(\v2/.gitaly.ReadinessCheckResponse.Failure.ResponseR\ffailedChecks\x1aC\n" +
	"\bResponse\x12\x12\n" +
	"\x04name\x18\x01 \x01(\tR\x04name\x12#\n" +
	"\rerror_message\x18\x02 \x01(\tR\ferrorMessageB\b\n" +
	"\x06Result\"\x18\n" +
	"\x16ServerSignatureRequest\"8\n" +
	"\x17ServerSignatureResponse\x12\x1d\n" +
	"\n" +
	"public_key\x18\x01 \x01(\fR\tpublicKey2\xd0\x02\n" +
	"\rServerService\x12C\n" +
	"\n" +
	"ServerInfo\x12\x19.gitaly.ServerInfoRequest\x1a\x1a.gitaly.ServerInfoResponse\x12O\n" +
	"\x0eDiskStatistics\x12\x1d.gitaly.DiskStatisticsRequest\x1a\x1e.gitaly.DiskStatisticsResponse\x12O\n" +
	"\x0eReadinessCheck\x12\x1d.gitaly.ReadinessCheckRequest\x1a\x1e.gitaly.ReadinessCheckResponse\x12R\n" +
	"\x0fServerSignature\x12\x1e.gitaly.ServerSignatureRequest\x1a\x1f.gitaly.ServerSignatureResponse\x1a\x04\xf0\x97(\x01B4Z2gitlab.com/gitlab-org/gitaly/v18/proto/go/gitalypbb\x06proto3"

var (
	file_server_proto_rawDescOnce sync.Once
	file_server_proto_rawDescData []byte
)

func file_server_proto_rawDescGZIP() []byte {
	file_server_proto_rawDescOnce.Do(func() {
		file_server_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_server_proto_rawDesc), len(file_server_proto_rawDesc)))
	})
	return file_server_proto_rawDescData
}

var file_server_proto_msgTypes = make([]protoimpl.MessageInfo, 13)
var file_server_proto_goTypes = []any{
	(*ServerInfoRequest)(nil),                       // 0: gitaly.ServerInfoRequest
	(*ServerInfoResponse)(nil),                      // 1: gitaly.ServerInfoResponse
	(*DiskStatisticsRequest)(nil),                   // 2: gitaly.DiskStatisticsRequest
	(*DiskStatisticsResponse)(nil),                  // 3: gitaly.DiskStatisticsResponse
	(*ReadinessCheckRequest)(nil),                   // 4: gitaly.ReadinessCheckRequest
	(*ReadinessCheckResponse)(nil),                  // 5: gitaly.ReadinessCheckResponse
	(*ServerSignatureRequest)(nil),                  // 6: gitaly.ServerSignatureRequest
	(*ServerSignatureResponse)(nil),                 // 7: gitaly.ServerSignatureResponse
	(*ServerInfoResponse_StorageStatus)(nil),        // 8: gitaly.ServerInfoResponse.StorageStatus
	(*DiskStatisticsResponse_StorageStatus)(nil),    // 9: gitaly.DiskStatisticsResponse.StorageStatus
	(*ReadinessCheckResponse_Ok)(nil),               // 10: gitaly.ReadinessCheckResponse.Ok
	(*ReadinessCheckResponse_Failure)(nil),          // 11: gitaly.ReadinessCheckResponse.Failure
	(*ReadinessCheckResponse_Failure_Response)(nil), // 12: gitaly.ReadinessCheckResponse.Failure.Response
	(*durationpb.Duration)(nil),                     // 13: google.protobuf.Duration
}
var file_server_proto_depIdxs = []int32{
	8,  // 0: gitaly.ServerInfoResponse.storage_statuses:type_name -> gitaly.ServerInfoResponse.StorageStatus
	9,  // 1: gitaly.DiskStatisticsResponse.storage_statuses:type_name -> gitaly.DiskStatisticsResponse.StorageStatus
	13, // 2: gitaly.ReadinessCheckRequest.timeout:type_name -> google.protobuf.Duration
	10, // 3: gitaly.ReadinessCheckResponse.ok_response:type_name -> gitaly.ReadinessCheckResponse.Ok
	11, // 4: gitaly.ReadinessCheckResponse.failure_response:type_name -> gitaly.ReadinessCheckResponse.Failure
	12, // 5: gitaly.ReadinessCheckResponse.Failure.failed_checks:type_name -> gitaly.ReadinessCheckResponse.Failure.Response
	0,  // 6: gitaly.ServerService.ServerInfo:input_type -> gitaly.ServerInfoRequest
	2,  // 7: gitaly.ServerService.DiskStatistics:input_type -> gitaly.DiskStatisticsRequest
	4,  // 8: gitaly.ServerService.ReadinessCheck:input_type -> gitaly.ReadinessCheckRequest
	6,  // 9: gitaly.ServerService.ServerSignature:input_type -> gitaly.ServerSignatureRequest
	1,  // 10: gitaly.ServerService.ServerInfo:output_type -> gitaly.ServerInfoResponse
	3,  // 11: gitaly.ServerService.DiskStatistics:output_type -> gitaly.DiskStatisticsResponse
	5,  // 12: gitaly.ServerService.ReadinessCheck:output_type -> gitaly.ReadinessCheckResponse
	7,  // 13: gitaly.ServerService.ServerSignature:output_type -> gitaly.ServerSignatureResponse
	10, // [10:14] is the sub-list for method output_type
	6,  // [6:10] is the sub-list for method input_type
	6,  // [6:6] is the sub-list for extension type_name
	6,  // [6:6] is the sub-list for extension extendee
	0,  // [0:6] is the sub-list for field type_name
}

func init() { file_server_proto_init() }
func file_server_proto_init() {
	if File_server_proto != nil {
		return
	}
	file_lint_proto_init()
	file_server_proto_msgTypes[5].OneofWrappers = []any{
		(*ReadinessCheckResponse_OkResponse)(nil),
		(*ReadinessCheckResponse_FailureResponse)(nil),
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_server_proto_rawDesc), len(file_server_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   13,
			NumExtensions: 0,
			NumServices:   1,
		},
		GoTypes:           file_server_proto_goTypes,
		DependencyIndexes: file_server_proto_depIdxs,
		MessageInfos:      file_server_proto_msgTypes,
	}.Build()
	File_server_proto = out.File
	file_server_proto_goTypes = nil
	file_server_proto_depIdxs = nil
}
