// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v6.30.2
// source: shared.proto

package gitalypb

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	timestamppb "google.golang.org/protobuf/types/known/timestamppb"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

// ObjectType ...
type ObjectType int32

const (
	// UNKNOWN ...
	ObjectType_UNKNOWN ObjectType = 0 // protolint:disable:this ENUM_FIELD_NAMES_PREFIX ENUM_FIELD_NAMES_ZERO_VALUE_END_WITH
	// COMMIT ...
	ObjectType_COMMIT ObjectType = 1 // protolint:disable:this ENUM_FIELD_NAMES_PREFIX
	// BLOB ...
	ObjectType_BLOB ObjectType = 2 // protolint:disable:this ENUM_FIELD_NAMES_PREFIX
	// TREE ...
	ObjectType_TREE ObjectType = 3 // protolint:disable:this ENUM_FIELD_NAMES_PREFIX
	// TAG ...
	ObjectType_TAG ObjectType = 4 // protolint:disable:this ENUM_FIELD_NAMES_PREFIX
)

// Enum value maps for ObjectType.
var (
	ObjectType_name = map[int32]string{
		0: "UNKNOWN",
		1: "COMMIT",
		2: "BLOB",
		3: "TREE",
		4: "TAG",
	}
	ObjectType_value = map[string]int32{
		"UNKNOWN": 0,
		"COMMIT":  1,
		"BLOB":    2,
		"TREE":    3,
		"TAG":     4,
	}
)

func (x ObjectType) Enum() *ObjectType {
	p := new(ObjectType)
	*p = x
	return p
}

func (x ObjectType) String() string {
	return protoimpl.X.EnumStringOf(x.Descriptor(), protoreflect.EnumNumber(x))
}

func (ObjectType) Descriptor() protoreflect.EnumDescriptor {
	return file_shared_proto_enumTypes[0].Descriptor()
}

func (ObjectType) Type() protoreflect.EnumType {
	return &file_shared_proto_enumTypes[0]
}

func (x ObjectType) Number() protoreflect.EnumNumber {
	return protoreflect.EnumNumber(x)
}

// Deprecated: Use ObjectType.Descriptor instead.
func (ObjectType) EnumDescriptor() ([]byte, []int) {
	return file_shared_proto_rawDescGZIP(), []int{0}
}

// ObjectFormat is the object format that a Git repository can use.
type ObjectFormat int32

const (
	// OBJECT_FORMAT_UNSPECIFIED is the default object type when it has not been explicitly requested
	// by the client. Defaults to OBJECT_FORMAT_SHA1.
	ObjectFormat_OBJECT_FORMAT_UNSPECIFIED ObjectFormat = 0
	// OBJECT_FORMAT_SHA1 is the object format based on the SHA1 hash.
	ObjectFormat_OBJECT_FORMAT_SHA1 ObjectFormat = 1
	// OBJECT_FORMAT_SHA256 is the object format based on the SHA256 hash. This is experimental.
	ObjectFormat_OBJECT_FORMAT_SHA256 ObjectFormat = 2
)

// Enum value maps for ObjectFormat.
var (
	ObjectFormat_name = map[int32]string{
		0: "OBJECT_FORMAT_UNSPECIFIED",
		1: "OBJECT_FORMAT_SHA1",
		2: "OBJECT_FORMAT_SHA256",
	}
	ObjectFormat_value = map[string]int32{
		"OBJECT_FORMAT_UNSPECIFIED": 0,
		"OBJECT_FORMAT_SHA1":        1,
		"OBJECT_FORMAT_SHA256":      2,
	}
)

func (x ObjectFormat) Enum() *ObjectFormat {
	p := new(ObjectFormat)
	*p = x
	return p
}

func (x ObjectFormat) String() string {
	return protoimpl.X.EnumStringOf(x.Descriptor(), protoreflect.EnumNumber(x))
}

func (ObjectFormat) Descriptor() protoreflect.EnumDescriptor {
	return file_shared_proto_enumTypes[1].Descriptor()
}

func (ObjectFormat) Type() protoreflect.EnumType {
	return &file_shared_proto_enumTypes[1]
}

func (x ObjectFormat) Number() protoreflect.EnumNumber {
	return protoreflect.EnumNumber(x)
}

// Deprecated: Use ObjectFormat.Descriptor instead.
func (ObjectFormat) EnumDescriptor() ([]byte, []int) {
	return file_shared_proto_rawDescGZIP(), []int{1}
}

// SignatureType ...
type SignatureType int32

const (
	// NONE ...
	SignatureType_NONE SignatureType = 0 // protolint:disable:this ENUM_FIELD_NAMES_PREFIX ENUM_FIELD_NAMES_ZERO_VALUE_END_WITH
	// PGP ...
	SignatureType_PGP SignatureType = 1 // protolint:disable:this ENUM_FIELD_NAMES_PREFIX
	// X509 ...
	SignatureType_X509 SignatureType = 2 // protolint:disable:this ENUM_FIELD_NAMES_PREFIX
	// SSH ...
	SignatureType_SSH SignatureType = 3 // protolint:disable:this ENUM_FIELD_NAMES_PREFIX
)

// Enum value maps for SignatureType.
var (
	SignatureType_name = map[int32]string{
		0: "NONE",
		1: "PGP",
		2: "X509",
		3: "SSH",
	}
	SignatureType_value = map[string]int32{
		"NONE": 0,
		"PGP":  1,
		"X509": 2,
		"SSH":  3,
	}
)

func (x SignatureType) Enum() *SignatureType {
	p := new(SignatureType)
	*p = x
	return p
}

func (x SignatureType) String() string {
	return protoimpl.X.EnumStringOf(x.Descriptor(), protoreflect.EnumNumber(x))
}

func (SignatureType) Descriptor() protoreflect.EnumDescriptor {
	return file_shared_proto_enumTypes[2].Descriptor()
}

func (SignatureType) Type() protoreflect.EnumType {
	return &file_shared_proto_enumTypes[2]
}

func (x SignatureType) Number() protoreflect.EnumNumber {
	return protoreflect.EnumNumber(x)
}

// Deprecated: Use SignatureType.Descriptor instead.
func (SignatureType) EnumDescriptor() ([]byte, []int) {
	return file_shared_proto_rawDescGZIP(), []int{2}
}

// SortDirection defines the sort direction.
type SortDirection int32

const (
	// ASCENDING sorts by the sort key in ascending order.
	SortDirection_ASCENDING SortDirection = 0 // protolint:disable:this ENUM_FIELD_NAMES_PREFIX ENUM_FIELD_NAMES_ZERO_VALUE_END_WITH
	// DESCENDING sorts by the sort key in descending order.
	SortDirection_DESCENDING SortDirection = 1 // protolint:disable:this ENUM_FIELD_NAMES_PREFIX
)

// Enum value maps for SortDirection.
var (
	SortDirection_name = map[int32]string{
		0: "ASCENDING",
		1: "DESCENDING",
	}
	SortDirection_value = map[string]int32{
		"ASCENDING":  0,
		"DESCENDING": 1,
	}
)

func (x SortDirection) Enum() *SortDirection {
	p := new(SortDirection)
	*p = x
	return p
}

func (x SortDirection) String() string {
	return protoimpl.X.EnumStringOf(x.Descriptor(), protoreflect.EnumNumber(x))
}

func (SortDirection) Descriptor() protoreflect.EnumDescriptor {
	return file_shared_proto_enumTypes[3].Descriptor()
}

func (SortDirection) Type() protoreflect.EnumType {
	return &file_shared_proto_enumTypes[3]
}

func (x SortDirection) Number() protoreflect.EnumNumber {
	return protoreflect.EnumNumber(x)
}

// Deprecated: Use SortDirection.Descriptor instead.
func (SortDirection) EnumDescriptor() ([]byte, []int) {
	return file_shared_proto_rawDescGZIP(), []int{3}
}

// Repository ...
type Repository struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// storage_name ...
	StorageName string `protobuf:"bytes,2,opt,name=storage_name,json=storageName,proto3" json:"storage_name,omitempty"`
	// relative_path ...
	RelativePath string `protobuf:"bytes,3,opt,name=relative_path,json=relativePath,proto3" json:"relative_path,omitempty"`
	// git_object_directory sets the GIT_OBJECT_DIRECTORY envvar on git commands to the value of this field.
	// It influences the object storage directory the SHA1 directories are created underneath.
	GitObjectDirectory string `protobuf:"bytes,4,opt,name=git_object_directory,json=gitObjectDirectory,proto3" json:"git_object_directory,omitempty"`
	// git_alternate_object_directories sets the GIT_ALTERNATE_OBJECT_DIRECTORIES envvar on git commands to
	// the values of this field. It influences the list of Git object directories which can be used to search
	// for Git objects.
	GitAlternateObjectDirectories []string `protobuf:"bytes,5,rep,name=git_alternate_object_directories,json=gitAlternateObjectDirectories,proto3" json:"git_alternate_object_directories,omitempty"`
	// gl_repository is used in callbacks to GitLab so that it knows what repository the event is
	// associated with. May be left empty on RPC's that do not perform callbacks.
	// During project creation, `gl_repository` may not be known.
	GlRepository string `protobuf:"bytes,6,opt,name=gl_repository,json=glRepository,proto3" json:"gl_repository,omitempty"`
	// gl_project_path is the human-readable GitLab project path (e.g. gitlab-org/gitlab-ce).
	// When hashed storage is use, this associates a project path with its
	// path on disk. The name can change over time (e.g. when a project is
	// renamed). This is primarily used for logging/debugging at the
	// moment.
	GlProjectPath string `protobuf:"bytes,8,opt,name=gl_project_path,json=glProjectPath,proto3" json:"gl_project_path,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Repository) Reset() {
	*x = Repository{}
	mi := &file_shared_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Repository) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Repository) ProtoMessage() {}

func (x *Repository) ProtoReflect() protoreflect.Message {
	mi := &file_shared_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Repository.ProtoReflect.Descriptor instead.
func (*Repository) Descriptor() ([]byte, []int) {
	return file_shared_proto_rawDescGZIP(), []int{0}
}

func (x *Repository) GetStorageName() string {
	if x != nil {
		return x.StorageName
	}
	return ""
}

func (x *Repository) GetRelativePath() string {
	if x != nil {
		return x.RelativePath
	}
	return ""
}

func (x *Repository) GetGitObjectDirectory() string {
	if x != nil {
		return x.GitObjectDirectory
	}
	return ""
}

func (x *Repository) GetGitAlternateObjectDirectories() []string {
	if x != nil {
		return x.GitAlternateObjectDirectories
	}
	return nil
}

func (x *Repository) GetGlRepository() string {
	if x != nil {
		return x.GlRepository
	}
	return ""
}

func (x *Repository) GetGlProjectPath() string {
	if x != nil {
		return x.GlProjectPath
	}
	return ""
}

// CommitTrailer is a single Git trailer (https://git-scm.com/docs/git-interpret-trailers)
// key-value pair.
type CommitTrailer struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// key is the key of the trailer, such as `Signed-off-by`.
	Key []byte `protobuf:"bytes,1,opt,name=key,proto3" json:"key,omitempty"`
	// value is the value of the trailer, such as `Alice <alice@gmail.com>`.
	Value         []byte `protobuf:"bytes,2,opt,name=value,proto3" json:"value,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *CommitTrailer) Reset() {
	*x = CommitTrailer{}
	mi := &file_shared_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *CommitTrailer) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*CommitTrailer) ProtoMessage() {}

func (x *CommitTrailer) ProtoReflect() protoreflect.Message {
	mi := &file_shared_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use CommitTrailer.ProtoReflect.Descriptor instead.
func (*CommitTrailer) Descriptor() ([]byte, []int) {
	return file_shared_proto_rawDescGZIP(), []int{1}
}

func (x *CommitTrailer) GetKey() []byte {
	if x != nil {
		return x.Key
	}
	return nil
}

func (x *CommitTrailer) GetValue() []byte {
	if x != nil {
		return x.Value
	}
	return nil
}

// CommitStatInfo includes the number of changed lines and files in the commit.
type CommitStatInfo struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// additions is the number of line additions in the commit.
	Additions int32 `protobuf:"varint,1,opt,name=additions,proto3" json:"additions,omitempty"`
	// deletions is the number of lines deleted in the commit.
	Deletions int32 `protobuf:"varint,2,opt,name=deletions,proto3" json:"deletions,omitempty"`
	// changed_files is the number of files changed in the commit.
	ChangedFiles  int32 `protobuf:"varint,3,opt,name=changed_files,json=changedFiles,proto3" json:"changed_files,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *CommitStatInfo) Reset() {
	*x = CommitStatInfo{}
	mi := &file_shared_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *CommitStatInfo) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*CommitStatInfo) ProtoMessage() {}

func (x *CommitStatInfo) ProtoReflect() protoreflect.Message {
	mi := &file_shared_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use CommitStatInfo.ProtoReflect.Descriptor instead.
func (*CommitStatInfo) Descriptor() ([]byte, []int) {
	return file_shared_proto_rawDescGZIP(), []int{2}
}

func (x *CommitStatInfo) GetAdditions() int32 {
	if x != nil {
		return x.Additions
	}
	return 0
}

func (x *CommitStatInfo) GetDeletions() int32 {
	if x != nil {
		return x.Deletions
	}
	return 0
}

func (x *CommitStatInfo) GetChangedFiles() int32 {
	if x != nil {
		return x.ChangedFiles
	}
	return 0
}

// GitCommit corresponds to Gitlab::Git::Commit
type GitCommit struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// id ...
	Id string `protobuf:"bytes,1,opt,name=id,proto3" json:"id,omitempty"`
	// subject ...
	Subject []byte `protobuf:"bytes,2,opt,name=subject,proto3" json:"subject,omitempty"`
	// body ...
	Body []byte `protobuf:"bytes,3,opt,name=body,proto3" json:"body,omitempty"`
	// author ...
	Author *CommitAuthor `protobuf:"bytes,4,opt,name=author,proto3" json:"author,omitempty"`
	// committer ...
	Committer *CommitAuthor `protobuf:"bytes,5,opt,name=committer,proto3" json:"committer,omitempty"`
	// parent_ids ...
	ParentIds []string `protobuf:"bytes,6,rep,name=parent_ids,json=parentIds,proto3" json:"parent_ids,omitempty"`
	// body_size is the size of the commit body. If body exceeds a certain threshold,
	// it will be nullified, but its size will be set in body_size so we can know if
	// a commit had a body in the first place.
	BodySize int64 `protobuf:"varint,7,opt,name=body_size,json=bodySize,proto3" json:"body_size,omitempty"`
	// signature_type ...
	SignatureType SignatureType `protobuf:"varint,8,opt,name=signature_type,json=signatureType,proto3,enum=gitaly.SignatureType" json:"signature_type,omitempty"`
	// tree_id is the object ID of the tree. The tree ID will always be filled, even
	// if the tree is empty. In that case the value will be `4b825dc642cb6eb9a060e54bf8d69288fbee4904`.
	// That value is equivalent to `git hash-object -t tree /dev/null`.
	TreeId string `protobuf:"bytes,9,opt,name=tree_id,json=treeId,proto3" json:"tree_id,omitempty"`
	// trailers is the list of Git trailers (https://git-scm.com/docs/git-interpret-trailers)
	// found in this commit's message. The number of trailers and their key/value
	// sizes are limited. If a trailer exceeds these size limits, it and any
	// trailers that follow it are not included.
	Trailers []*CommitTrailer `protobuf:"bytes,10,rep,name=trailers,proto3" json:"trailers,omitempty"`
	// short_stats are the git stats including additions, deletions and changed_files,
	// they are only set when `include_shortstat == true`.
	ShortStats *CommitStatInfo `protobuf:"bytes,11,opt,name=short_stats,json=shortStats,proto3" json:"short_stats,omitempty"`
	// referenced_by contains fully-qualified reference names (e.g refs/heads/main)
	// that point to the commit.
	ReferencedBy [][]byte `protobuf:"bytes,12,rep,name=referenced_by,json=referencedBy,proto3" json:"referenced_by,omitempty"` // protolint:disable:this REPEATED_FIELD_NAMES_PLURALIZED
	// encoding is the encoding of the commit message. This field will only be present if
	// `i18n.commitEncoding` was set to a value other than "UTF-8" at the time
	// this commit was made.
	// See: https://git-scm.com/docs/git-commit#_discussion
	Encoding      string `protobuf:"bytes,13,opt,name=encoding,proto3" json:"encoding,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *GitCommit) Reset() {
	*x = GitCommit{}
	mi := &file_shared_proto_msgTypes[3]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *GitCommit) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*GitCommit) ProtoMessage() {}

func (x *GitCommit) ProtoReflect() protoreflect.Message {
	mi := &file_shared_proto_msgTypes[3]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use GitCommit.ProtoReflect.Descriptor instead.
func (*GitCommit) Descriptor() ([]byte, []int) {
	return file_shared_proto_rawDescGZIP(), []int{3}
}

func (x *GitCommit) GetId() string {
	if x != nil {
		return x.Id
	}
	return ""
}

func (x *GitCommit) GetSubject() []byte {
	if x != nil {
		return x.Subject
	}
	return nil
}

func (x *GitCommit) GetBody() []byte {
	if x != nil {
		return x.Body
	}
	return nil
}

func (x *GitCommit) GetAuthor() *CommitAuthor {
	if x != nil {
		return x.Author
	}
	return nil
}

func (x *GitCommit) GetCommitter() *CommitAuthor {
	if x != nil {
		return x.Committer
	}
	return nil
}

func (x *GitCommit) GetParentIds() []string {
	if x != nil {
		return x.ParentIds
	}
	return nil
}

func (x *GitCommit) GetBodySize() int64 {
	if x != nil {
		return x.BodySize
	}
	return 0
}

func (x *GitCommit) GetSignatureType() SignatureType {
	if x != nil {
		return x.SignatureType
	}
	return SignatureType_NONE
}

func (x *GitCommit) GetTreeId() string {
	if x != nil {
		return x.TreeId
	}
	return ""
}

func (x *GitCommit) GetTrailers() []*CommitTrailer {
	if x != nil {
		return x.Trailers
	}
	return nil
}

func (x *GitCommit) GetShortStats() *CommitStatInfo {
	if x != nil {
		return x.ShortStats
	}
	return nil
}

func (x *GitCommit) GetReferencedBy() [][]byte {
	if x != nil {
		return x.ReferencedBy
	}
	return nil
}

func (x *GitCommit) GetEncoding() string {
	if x != nil {
		return x.Encoding
	}
	return ""
}

// CommitAuthor ...
type CommitAuthor struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// name ...
	Name []byte `protobuf:"bytes,1,opt,name=name,proto3" json:"name,omitempty"`
	// email ...
	Email []byte `protobuf:"bytes,2,opt,name=email,proto3" json:"email,omitempty"`
	// date ...
	Date *timestamppb.Timestamp `protobuf:"bytes,3,opt,name=date,proto3" json:"date,omitempty"`
	// timezone ...
	Timezone      []byte `protobuf:"bytes,4,opt,name=timezone,proto3" json:"timezone,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *CommitAuthor) Reset() {
	*x = CommitAuthor{}
	mi := &file_shared_proto_msgTypes[4]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *CommitAuthor) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*CommitAuthor) ProtoMessage() {}

func (x *CommitAuthor) ProtoReflect() protoreflect.Message {
	mi := &file_shared_proto_msgTypes[4]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use CommitAuthor.ProtoReflect.Descriptor instead.
func (*CommitAuthor) Descriptor() ([]byte, []int) {
	return file_shared_proto_rawDescGZIP(), []int{4}
}

func (x *CommitAuthor) GetName() []byte {
	if x != nil {
		return x.Name
	}
	return nil
}

func (x *CommitAuthor) GetEmail() []byte {
	if x != nil {
		return x.Email
	}
	return nil
}

func (x *CommitAuthor) GetDate() *timestamppb.Timestamp {
	if x != nil {
		return x.Date
	}
	return nil
}

func (x *CommitAuthor) GetTimezone() []byte {
	if x != nil {
		return x.Timezone
	}
	return nil
}

// ExitStatus ...
type ExitStatus struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// value ...
	Value         int32 `protobuf:"varint,1,opt,name=value,proto3" json:"value,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ExitStatus) Reset() {
	*x = ExitStatus{}
	mi := &file_shared_proto_msgTypes[5]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ExitStatus) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ExitStatus) ProtoMessage() {}

func (x *ExitStatus) ProtoReflect() protoreflect.Message {
	mi := &file_shared_proto_msgTypes[5]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ExitStatus.ProtoReflect.Descriptor instead.
func (*ExitStatus) Descriptor() ([]byte, []int) {
	return file_shared_proto_rawDescGZIP(), []int{5}
}

func (x *ExitStatus) GetValue() int32 {
	if x != nil {
		return x.Value
	}
	return 0
}

// Branch corresponds to Gitlab::Git::Branch
type Branch struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// name ...
	Name []byte `protobuf:"bytes,1,opt,name=name,proto3" json:"name,omitempty"`
	// target_commit ...
	TargetCommit  *GitCommit `protobuf:"bytes,2,opt,name=target_commit,json=targetCommit,proto3" json:"target_commit,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Branch) Reset() {
	*x = Branch{}
	mi := &file_shared_proto_msgTypes[6]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Branch) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Branch) ProtoMessage() {}

func (x *Branch) ProtoReflect() protoreflect.Message {
	mi := &file_shared_proto_msgTypes[6]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Branch.ProtoReflect.Descriptor instead.
func (*Branch) Descriptor() ([]byte, []int) {
	return file_shared_proto_rawDescGZIP(), []int{6}
}

func (x *Branch) GetName() []byte {
	if x != nil {
		return x.Name
	}
	return nil
}

func (x *Branch) GetTargetCommit() *GitCommit {
	if x != nil {
		return x.TargetCommit
	}
	return nil
}

// Tag ...
type Tag struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// name ...
	Name []byte `protobuf:"bytes,1,opt,name=name,proto3" json:"name,omitempty"`
	// id ...
	Id string `protobuf:"bytes,2,opt,name=id,proto3" json:"id,omitempty"`
	// target_commit ...
	TargetCommit *GitCommit `protobuf:"bytes,3,opt,name=target_commit,json=targetCommit,proto3" json:"target_commit,omitempty"`
	// message is the message of the tag. If message exceeds a certain
	// threshold, it will be nullified, but its size will be set in
	// message_size so we can know if a tag had a message in the first place.
	Message []byte `protobuf:"bytes,4,opt,name=message,proto3" json:"message,omitempty"`
	// message_size ...
	MessageSize int64 `protobuf:"varint,5,opt,name=message_size,json=messageSize,proto3" json:"message_size,omitempty"`
	// tagger ...
	Tagger *CommitAuthor `protobuf:"bytes,6,opt,name=tagger,proto3" json:"tagger,omitempty"`
	// signature_type ...
	SignatureType SignatureType `protobuf:"varint,7,opt,name=signature_type,json=signatureType,proto3,enum=gitaly.SignatureType" json:"signature_type,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Tag) Reset() {
	*x = Tag{}
	mi := &file_shared_proto_msgTypes[7]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Tag) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Tag) ProtoMessage() {}

func (x *Tag) ProtoReflect() protoreflect.Message {
	mi := &file_shared_proto_msgTypes[7]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Tag.ProtoReflect.Descriptor instead.
func (*Tag) Descriptor() ([]byte, []int) {
	return file_shared_proto_rawDescGZIP(), []int{7}
}

func (x *Tag) GetName() []byte {
	if x != nil {
		return x.Name
	}
	return nil
}

func (x *Tag) GetId() string {
	if x != nil {
		return x.Id
	}
	return ""
}

func (x *Tag) GetTargetCommit() *GitCommit {
	if x != nil {
		return x.TargetCommit
	}
	return nil
}

func (x *Tag) GetMessage() []byte {
	if x != nil {
		return x.Message
	}
	return nil
}

func (x *Tag) GetMessageSize() int64 {
	if x != nil {
		return x.MessageSize
	}
	return 0
}

func (x *Tag) GetTagger() *CommitAuthor {
	if x != nil {
		return x.Tagger
	}
	return nil
}

func (x *Tag) GetSignatureType() SignatureType {
	if x != nil {
		return x.SignatureType
	}
	return SignatureType_NONE
}

// User ...
type User struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// gl_id ...
	GlId string `protobuf:"bytes,1,opt,name=gl_id,json=glId,proto3" json:"gl_id,omitempty"`
	// name ...
	Name []byte `protobuf:"bytes,2,opt,name=name,proto3" json:"name,omitempty"`
	// email ...
	Email []byte `protobuf:"bytes,3,opt,name=email,proto3" json:"email,omitempty"`
	// gl_username ...
	GlUsername string `protobuf:"bytes,4,opt,name=gl_username,json=glUsername,proto3" json:"gl_username,omitempty"`
	// timezone is the timezone as configured by the user in the web interface. This
	// timezone may be used when new commits are created via RPC calls.
	Timezone      string `protobuf:"bytes,5,opt,name=timezone,proto3" json:"timezone,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *User) Reset() {
	*x = User{}
	mi := &file_shared_proto_msgTypes[8]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *User) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*User) ProtoMessage() {}

func (x *User) ProtoReflect() protoreflect.Message {
	mi := &file_shared_proto_msgTypes[8]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use User.ProtoReflect.Descriptor instead.
func (*User) Descriptor() ([]byte, []int) {
	return file_shared_proto_rawDescGZIP(), []int{8}
}

func (x *User) GetGlId() string {
	if x != nil {
		return x.GlId
	}
	return ""
}

func (x *User) GetName() []byte {
	if x != nil {
		return x.Name
	}
	return nil
}

func (x *User) GetEmail() []byte {
	if x != nil {
		return x.Email
	}
	return nil
}

func (x *User) GetGlUsername() string {
	if x != nil {
		return x.GlUsername
	}
	return ""
}

func (x *User) GetTimezone() string {
	if x != nil {
		return x.Timezone
	}
	return ""
}

// ObjectPool ...
type ObjectPool struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// repository ...
	Repository    *Repository `protobuf:"bytes,1,opt,name=repository,proto3" json:"repository,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ObjectPool) Reset() {
	*x = ObjectPool{}
	mi := &file_shared_proto_msgTypes[9]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ObjectPool) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ObjectPool) ProtoMessage() {}

func (x *ObjectPool) ProtoReflect() protoreflect.Message {
	mi := &file_shared_proto_msgTypes[9]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ObjectPool.ProtoReflect.Descriptor instead.
func (*ObjectPool) Descriptor() ([]byte, []int) {
	return file_shared_proto_rawDescGZIP(), []int{9}
}

func (x *ObjectPool) GetRepository() *Repository {
	if x != nil {
		return x.Repository
	}
	return nil
}

// PaginationParameter controls pagination within RPCs.
type PaginationParameter struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// page_token instructs pagination to start sending results after the provided page
	// token appears. A page token allows for a generic pattern to uniquely
	// identify a result or 'page'. Each paginated RPC may interpret a page
	// token differently.
	PageToken string `protobuf:"bytes,1,opt,name=page_token,json=pageToken,proto3" json:"page_token,omitempty"`
	// limit is the maximum number of objects the client will receive. When fully consuming
	// the response the client will receive _at most_ `limit` number of resulting objects.
	// Note that the number of response messages might be much lower, as some response
	// messages already send multiple objects per message.
	// When the limit is smaller than 0, it will be normalized to 2147483647
	// on the server side. When limit is not set, it defaults to 0, and no
	// results are send in the response.
	Limit         int32 `protobuf:"varint,2,opt,name=limit,proto3" json:"limit,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *PaginationParameter) Reset() {
	*x = PaginationParameter{}
	mi := &file_shared_proto_msgTypes[10]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *PaginationParameter) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*PaginationParameter) ProtoMessage() {}

func (x *PaginationParameter) ProtoReflect() protoreflect.Message {
	mi := &file_shared_proto_msgTypes[10]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use PaginationParameter.ProtoReflect.Descriptor instead.
func (*PaginationParameter) Descriptor() ([]byte, []int) {
	return file_shared_proto_rawDescGZIP(), []int{10}
}

func (x *PaginationParameter) GetPageToken() string {
	if x != nil {
		return x.PageToken
	}
	return ""
}

func (x *PaginationParameter) GetLimit() int32 {
	if x != nil {
		return x.Limit
	}
	return 0
}

// PaginationCursor defines the page token clients should use to fetch the next page.
type PaginationCursor struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// next_cursor is an opaque token provided to the caller to indicate what the caller
	// should present as a page_token to get subsequent results.
	NextCursor    string `protobuf:"bytes,1,opt,name=next_cursor,json=nextCursor,proto3" json:"next_cursor,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *PaginationCursor) Reset() {
	*x = PaginationCursor{}
	mi := &file_shared_proto_msgTypes[11]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *PaginationCursor) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*PaginationCursor) ProtoMessage() {}

func (x *PaginationCursor) ProtoReflect() protoreflect.Message {
	mi := &file_shared_proto_msgTypes[11]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use PaginationCursor.ProtoReflect.Descriptor instead.
func (*PaginationCursor) Descriptor() ([]byte, []int) {
	return file_shared_proto_rawDescGZIP(), []int{11}
}

func (x *PaginationCursor) GetNextCursor() string {
	if x != nil {
		return x.NextCursor
	}
	return ""
}

// GlobalOptions are additional git options as defined at https://git-scm.com/docs/git/#_options.
type GlobalOptions struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// literal_pathspecs should be set to treat pathspecs literally (i.e. no globbing, no pathspec magic).
	LiteralPathspecs bool `protobuf:"varint,1,opt,name=literal_pathspecs,json=literalPathspecs,proto3" json:"literal_pathspecs,omitempty"`
	unknownFields    protoimpl.UnknownFields
	sizeCache        protoimpl.SizeCache
}

func (x *GlobalOptions) Reset() {
	*x = GlobalOptions{}
	mi := &file_shared_proto_msgTypes[12]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *GlobalOptions) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*GlobalOptions) ProtoMessage() {}

func (x *GlobalOptions) ProtoReflect() protoreflect.Message {
	mi := &file_shared_proto_msgTypes[12]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use GlobalOptions.ProtoReflect.Descriptor instead.
func (*GlobalOptions) Descriptor() ([]byte, []int) {
	return file_shared_proto_rawDescGZIP(), []int{12}
}

func (x *GlobalOptions) GetLiteralPathspecs() bool {
	if x != nil {
		return x.LiteralPathspecs
	}
	return false
}

// Partition represents storage partition
type Partition struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// id is the identifier of a partition within the storage
	Id            string `protobuf:"bytes,1,opt,name=id,proto3" json:"id,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Partition) Reset() {
	*x = Partition{}
	mi := &file_shared_proto_msgTypes[13]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Partition) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Partition) ProtoMessage() {}

func (x *Partition) ProtoReflect() protoreflect.Message {
	mi := &file_shared_proto_msgTypes[13]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Partition.ProtoReflect.Descriptor instead.
func (*Partition) Descriptor() ([]byte, []int) {
	return file_shared_proto_rawDescGZIP(), []int{13}
}

func (x *Partition) GetId() string {
	if x != nil {
		return x.Id
	}
	return ""
}

var File_shared_proto protoreflect.FileDescriptor

const file_shared_proto_rawDesc = "" +
	"\n" +
	"\fshared.proto\x12\x06gitaly\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\n" +
	"lint.proto\"\xae\x02\n" +
	"\n" +
	"Repository\x12!\n" +
	"\fstorage_name\x18\x02 \x01(\tR\vstorageName\x12#\n" +
	"\rrelative_path\x18\x03 \x01(\tR\frelativePath\x120\n" +
	"\x14git_object_directory\x18\x04 \x01(\tR\x12gitObjectDirectory\x12G\n" +
	" git_alternate_object_directories\x18\x05 \x03(\tR\x1dgitAlternateObjectDirectories\x12#\n" +
	"\rgl_repository\x18\x06 \x01(\tR\fglRepository\x12&\n" +
	"\x0fgl_project_path\x18\b \x01(\tR\rglProjectPathJ\x04\b\x01\x10\x02J\x04\b\a\x10\bR\x04path\"7\n" +
	"\rCommitTrailer\x12\x10\n" +
	"\x03key\x18\x01 \x01(\fR\x03key\x12\x14\n" +
	"\x05value\x18\x02 \x01(\fR\x05value\"q\n" +
	"\x0eCommitStatInfo\x12\x1c\n" +
	"\tadditions\x18\x01 \x01(\x05R\tadditions\x12\x1c\n" +
	"\tdeletions\x18\x02 \x01(\x05R\tdeletions\x12#\n" +
	"\rchanged_files\x18\x03 \x01(\x05R\fchangedFiles\"\xeb\x03\n" +
	"\tGitCommit\x12\x0e\n" +
	"\x02id\x18\x01 \x01(\tR\x02id\x12\x18\n" +
	"\asubject\x18\x02 \x01(\fR\asubject\x12\x12\n" +
	"\x04body\x18\x03 \x01(\fR\x04body\x12,\n" +
	"\x06author\x18\x04 \x01(\v2\x14.gitaly.CommitAuthorR\x06author\x122\n" +
	"\tcommitter\x18\x05 \x01(\v2\x14.gitaly.CommitAuthorR\tcommitter\x12\x1d\n" +
	"\n" +
	"parent_ids\x18\x06 \x03(\tR\tparentIds\x12\x1b\n" +
	"\tbody_size\x18\a \x01(\x03R\bbodySize\x12<\n" +
	"\x0esignature_type\x18\b \x01(\x0e2\x15.gitaly.SignatureTypeR\rsignatureType\x12\x17\n" +
	"\atree_id\x18\t \x01(\tR\x06treeId\x121\n" +
	"\btrailers\x18\n" +
	" \x03(\v2\x15.gitaly.CommitTrailerR\btrailers\x127\n" +
	"\vshort_stats\x18\v \x01(\v2\x16.gitaly.CommitStatInfoR\n" +
	"shortStats\x12#\n" +
	"\rreferenced_by\x18\f \x03(\fR\freferencedBy\x12\x1a\n" +
	"\bencoding\x18\r \x01(\tR\bencoding\"\x84\x01\n" +
	"\fCommitAuthor\x12\x12\n" +
	"\x04name\x18\x01 \x01(\fR\x04name\x12\x14\n" +
	"\x05email\x18\x02 \x01(\fR\x05email\x12.\n" +
	"\x04date\x18\x03 \x01(\v2\x1a.google.protobuf.TimestampR\x04date\x12\x1a\n" +
	"\btimezone\x18\x04 \x01(\fR\btimezone\"\"\n" +
	"\n" +
	"ExitStatus\x12\x14\n" +
	"\x05value\x18\x01 \x01(\x05R\x05value\"T\n" +
	"\x06Branch\x12\x12\n" +
	"\x04name\x18\x01 \x01(\fR\x04name\x126\n" +
	"\rtarget_commit\x18\x02 \x01(\v2\x11.gitaly.GitCommitR\ftargetCommit\"\x8a\x02\n" +
	"\x03Tag\x12\x12\n" +
	"\x04name\x18\x01 \x01(\fR\x04name\x12\x0e\n" +
	"\x02id\x18\x02 \x01(\tR\x02id\x126\n" +
	"\rtarget_commit\x18\x03 \x01(\v2\x11.gitaly.GitCommitR\ftargetCommit\x12\x18\n" +
	"\amessage\x18\x04 \x01(\fR\amessage\x12!\n" +
	"\fmessage_size\x18\x05 \x01(\x03R\vmessageSize\x12,\n" +
	"\x06tagger\x18\x06 \x01(\v2\x14.gitaly.CommitAuthorR\x06tagger\x12<\n" +
	"\x0esignature_type\x18\a \x01(\x0e2\x15.gitaly.SignatureTypeR\rsignatureType\"\x82\x01\n" +
	"\x04User\x12\x13\n" +
	"\x05gl_id\x18\x01 \x01(\tR\x04glId\x12\x12\n" +
	"\x04name\x18\x02 \x01(\fR\x04name\x12\x14\n" +
	"\x05email\x18\x03 \x01(\fR\x05email\x12\x1f\n" +
	"\vgl_username\x18\x04 \x01(\tR\n" +
	"glUsername\x12\x1a\n" +
	"\btimezone\x18\x05 \x01(\tR\btimezone\"F\n" +
	"\n" +
	"ObjectPool\x128\n" +
	"\n" +
	"repository\x18\x01 \x01(\v2\x12.gitaly.RepositoryB\x04\x90\xc6,\x01R\n" +
	"repository\"J\n" +
	"\x13PaginationParameter\x12\x1d\n" +
	"\n" +
	"page_token\x18\x01 \x01(\tR\tpageToken\x12\x14\n" +
	"\x05limit\x18\x02 \x01(\x05R\x05limit\"3\n" +
	"\x10PaginationCursor\x12\x1f\n" +
	"\vnext_cursor\x18\x01 \x01(\tR\n" +
	"nextCursor\"<\n" +
	"\rGlobalOptions\x12+\n" +
	"\x11literal_pathspecs\x18\x01 \x01(\bR\x10literalPathspecs\"\x1b\n" +
	"\tPartition\x12\x0e\n" +
	"\x02id\x18\x01 \x01(\tR\x02id*B\n" +
	"\n" +
	"ObjectType\x12\v\n" +
	"\aUNKNOWN\x10\x00\x12\n" +
	"\n" +
	"\x06COMMIT\x10\x01\x12\b\n" +
	"\x04BLOB\x10\x02\x12\b\n" +
	"\x04TREE\x10\x03\x12\a\n" +
	"\x03TAG\x10\x04*_\n" +
	"\fObjectFormat\x12\x1d\n" +
	"\x19OBJECT_FORMAT_UNSPECIFIED\x10\x00\x12\x16\n" +
	"\x12OBJECT_FORMAT_SHA1\x10\x01\x12\x18\n" +
	"\x14OBJECT_FORMAT_SHA256\x10\x02*5\n" +
	"\rSignatureType\x12\b\n" +
	"\x04NONE\x10\x00\x12\a\n" +
	"\x03PGP\x10\x01\x12\b\n" +
	"\x04X509\x10\x02\x12\a\n" +
	"\x03SSH\x10\x03*.\n" +
	"\rSortDirection\x12\r\n" +
	"\tASCENDING\x10\x00\x12\x0e\n" +
	"\n" +
	"DESCENDING\x10\x01B4Z2gitlab.com/gitlab-org/gitaly/v18/proto/go/gitalypbb\x06proto3"

var (
	file_shared_proto_rawDescOnce sync.Once
	file_shared_proto_rawDescData []byte
)

func file_shared_proto_rawDescGZIP() []byte {
	file_shared_proto_rawDescOnce.Do(func() {
		file_shared_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_shared_proto_rawDesc), len(file_shared_proto_rawDesc)))
	})
	return file_shared_proto_rawDescData
}

var file_shared_proto_enumTypes = make([]protoimpl.EnumInfo, 4)
var file_shared_proto_msgTypes = make([]protoimpl.MessageInfo, 14)
var file_shared_proto_goTypes = []any{
	(ObjectType)(0),               // 0: gitaly.ObjectType
	(ObjectFormat)(0),             // 1: gitaly.ObjectFormat
	(SignatureType)(0),            // 2: gitaly.SignatureType
	(SortDirection)(0),            // 3: gitaly.SortDirection
	(*Repository)(nil),            // 4: gitaly.Repository
	(*CommitTrailer)(nil),         // 5: gitaly.CommitTrailer
	(*CommitStatInfo)(nil),        // 6: gitaly.CommitStatInfo
	(*GitCommit)(nil),             // 7: gitaly.GitCommit
	(*CommitAuthor)(nil),          // 8: gitaly.CommitAuthor
	(*ExitStatus)(nil),            // 9: gitaly.ExitStatus
	(*Branch)(nil),                // 10: gitaly.Branch
	(*Tag)(nil),                   // 11: gitaly.Tag
	(*User)(nil),                  // 12: gitaly.User
	(*ObjectPool)(nil),            // 13: gitaly.ObjectPool
	(*PaginationParameter)(nil),   // 14: gitaly.PaginationParameter
	(*PaginationCursor)(nil),      // 15: gitaly.PaginationCursor
	(*GlobalOptions)(nil),         // 16: gitaly.GlobalOptions
	(*Partition)(nil),             // 17: gitaly.Partition
	(*timestamppb.Timestamp)(nil), // 18: google.protobuf.Timestamp
}
var file_shared_proto_depIdxs = []int32{
	8,  // 0: gitaly.GitCommit.author:type_name -> gitaly.CommitAuthor
	8,  // 1: gitaly.GitCommit.committer:type_name -> gitaly.CommitAuthor
	2,  // 2: gitaly.GitCommit.signature_type:type_name -> gitaly.SignatureType
	5,  // 3: gitaly.GitCommit.trailers:type_name -> gitaly.CommitTrailer
	6,  // 4: gitaly.GitCommit.short_stats:type_name -> gitaly.CommitStatInfo
	18, // 5: gitaly.CommitAuthor.date:type_name -> google.protobuf.Timestamp
	7,  // 6: gitaly.Branch.target_commit:type_name -> gitaly.GitCommit
	7,  // 7: gitaly.Tag.target_commit:type_name -> gitaly.GitCommit
	8,  // 8: gitaly.Tag.tagger:type_name -> gitaly.CommitAuthor
	2,  // 9: gitaly.Tag.signature_type:type_name -> gitaly.SignatureType
	4,  // 10: gitaly.ObjectPool.repository:type_name -> gitaly.Repository
	11, // [11:11] is the sub-list for method output_type
	11, // [11:11] is the sub-list for method input_type
	11, // [11:11] is the sub-list for extension type_name
	11, // [11:11] is the sub-list for extension extendee
	0,  // [0:11] is the sub-list for field type_name
}

func init() { file_shared_proto_init() }
func file_shared_proto_init() {
	if File_shared_proto != nil {
		return
	}
	file_lint_proto_init()
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_shared_proto_rawDesc), len(file_shared_proto_rawDesc)),
			NumEnums:      4,
			NumMessages:   14,
			NumExtensions: 0,
			NumServices:   0,
		},
		GoTypes:           file_shared_proto_goTypes,
		DependencyIndexes: file_shared_proto_depIdxs,
		EnumInfos:         file_shared_proto_enumTypes,
		MessageInfos:      file_shared_proto_msgTypes,
	}.Build()
	File_shared_proto = out.File
	file_shared_proto_goTypes = nil
	file_shared_proto_depIdxs = nil
}
