// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v6.30.2
// source: testproto/error_metadata.proto

package testproto

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

// errorMetadata is a key-value metadata item that may be attached to errors. We only use this
// infrastructure for testing purposes to assert that we add error metadata as expected that would
// otherwise only get logged.
type ErrorMetadata struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// key is the key of the item.
	Key []byte `protobuf:"bytes,1,opt,name=key,proto3" json:"key,omitempty"`
	// value is the value of the item.
	Value         []byte `protobuf:"bytes,2,opt,name=value,proto3" json:"value,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ErrorMetadata) Reset() {
	*x = ErrorMetadata{}
	mi := &file_testproto_error_metadata_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ErrorMetadata) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ErrorMetadata) ProtoMessage() {}

func (x *ErrorMetadata) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_error_metadata_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ErrorMetadata.ProtoReflect.Descriptor instead.
func (*ErrorMetadata) Descriptor() ([]byte, []int) {
	return file_testproto_error_metadata_proto_rawDescGZIP(), []int{0}
}

func (x *ErrorMetadata) GetKey() []byte {
	if x != nil {
		return x.Key
	}
	return nil
}

func (x *ErrorMetadata) GetValue() []byte {
	if x != nil {
		return x.Value
	}
	return nil
}

var File_testproto_error_metadata_proto protoreflect.FileDescriptor

const file_testproto_error_metadata_proto_rawDesc = "" +
	"\n" +
	"\x1etestproto/error_metadata.proto\x12\ttestproto\"7\n" +
	"\rErrorMetadata\x12\x10\n" +
	"\x03key\x18\x01 \x01(\fR\x03key\x12\x14\n" +
	"\x05value\x18\x02 \x01(\fR\x05valueB>Z<gitlab.com/gitlab-org/gitaly/v18/proto/go/gitalypb/testprotob\x06proto3"

var (
	file_testproto_error_metadata_proto_rawDescOnce sync.Once
	file_testproto_error_metadata_proto_rawDescData []byte
)

func file_testproto_error_metadata_proto_rawDescGZIP() []byte {
	file_testproto_error_metadata_proto_rawDescOnce.Do(func() {
		file_testproto_error_metadata_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_testproto_error_metadata_proto_rawDesc), len(file_testproto_error_metadata_proto_rawDesc)))
	})
	return file_testproto_error_metadata_proto_rawDescData
}

var file_testproto_error_metadata_proto_msgTypes = make([]protoimpl.MessageInfo, 1)
var file_testproto_error_metadata_proto_goTypes = []any{
	(*ErrorMetadata)(nil), // 0: testproto.ErrorMetadata
}
var file_testproto_error_metadata_proto_depIdxs = []int32{
	0, // [0:0] is the sub-list for method output_type
	0, // [0:0] is the sub-list for method input_type
	0, // [0:0] is the sub-list for extension type_name
	0, // [0:0] is the sub-list for extension extendee
	0, // [0:0] is the sub-list for field type_name
}

func init() { file_testproto_error_metadata_proto_init() }
func file_testproto_error_metadata_proto_init() {
	if File_testproto_error_metadata_proto != nil {
		return
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_testproto_error_metadata_proto_rawDesc), len(file_testproto_error_metadata_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   1,
			NumExtensions: 0,
			NumServices:   0,
		},
		GoTypes:           file_testproto_error_metadata_proto_goTypes,
		DependencyIndexes: file_testproto_error_metadata_proto_depIdxs,
		MessageInfos:      file_testproto_error_metadata_proto_msgTypes,
	}.Build()
	File_testproto_error_metadata_proto = out.File
	file_testproto_error_metadata_proto_goTypes = nil
	file_testproto_error_metadata_proto_depIdxs = nil
}
