// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v6.30.2
// source: remote.proto

package gitalypb

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

// UpdateRemoteMirrorRequest is a request for the UpdateRemoteMirror RPC.
type UpdateRemoteMirrorRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// repository is the repository whose mirror repository to update.
	Repository *Repository `protobuf:"bytes,1,opt,name=repository,proto3" json:"repository,omitempty"`
	// only_branches_matching contains patterns to match branches against. Only the
	// matched branches are updated in the remote mirror. If no patterns are
	// specified, all branches are updated. The patterns should only contain the
	// branch name without the 'refs/heads/' prefix. "*" can be used as a
	// wildcard to match anything. only_branches_matching can be streamed to the
	// server over multiple messages. Optional.
	OnlyBranchesMatching [][]byte `protobuf:"bytes,3,rep,name=only_branches_matching,json=onlyBranchesMatching,proto3" json:"only_branches_matching,omitempty"` // protolint:disable:this REPEATED_FIELD_NAMES_PLURALIZED
	// ssh_key is the SSH key to use for accessing to the mirror repository.
	// Optional.
	SshKey string `protobuf:"bytes,4,opt,name=ssh_key,json=sshKey,proto3" json:"ssh_key,omitempty"`
	// known_hosts specifies the identities used for strict host key checking.
	// Optional.
	KnownHosts string `protobuf:"bytes,5,opt,name=known_hosts,json=knownHosts,proto3" json:"known_hosts,omitempty"`
	// keep_divergent_refs specifies whether or not to update diverged references
	// in the mirror repository.
	KeepDivergentRefs bool `protobuf:"varint,6,opt,name=keep_divergent_refs,json=keepDivergentRefs,proto3" json:"keep_divergent_refs,omitempty"`
	// remote contains parameters required to connect to the remote repository.
	// This allows Gitaly to use an in-memory remote and does not require any
	// on-disk remote configuration.
	Remote        *UpdateRemoteMirrorRequest_Remote `protobuf:"bytes,7,opt,name=remote,proto3" json:"remote,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *UpdateRemoteMirrorRequest) Reset() {
	*x = UpdateRemoteMirrorRequest{}
	mi := &file_remote_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *UpdateRemoteMirrorRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*UpdateRemoteMirrorRequest) ProtoMessage() {}

func (x *UpdateRemoteMirrorRequest) ProtoReflect() protoreflect.Message {
	mi := &file_remote_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use UpdateRemoteMirrorRequest.ProtoReflect.Descriptor instead.
func (*UpdateRemoteMirrorRequest) Descriptor() ([]byte, []int) {
	return file_remote_proto_rawDescGZIP(), []int{0}
}

func (x *UpdateRemoteMirrorRequest) GetRepository() *Repository {
	if x != nil {
		return x.Repository
	}
	return nil
}

func (x *UpdateRemoteMirrorRequest) GetOnlyBranchesMatching() [][]byte {
	if x != nil {
		return x.OnlyBranchesMatching
	}
	return nil
}

func (x *UpdateRemoteMirrorRequest) GetSshKey() string {
	if x != nil {
		return x.SshKey
	}
	return ""
}

func (x *UpdateRemoteMirrorRequest) GetKnownHosts() string {
	if x != nil {
		return x.KnownHosts
	}
	return ""
}

func (x *UpdateRemoteMirrorRequest) GetKeepDivergentRefs() bool {
	if x != nil {
		return x.KeepDivergentRefs
	}
	return false
}

func (x *UpdateRemoteMirrorRequest) GetRemote() *UpdateRemoteMirrorRequest_Remote {
	if x != nil {
		return x.Remote
	}
	return nil
}

// UpdateRemoteMirrorResponse is a response for the UpdateRemoteMirror RPC.
type UpdateRemoteMirrorResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// divergent_refs contains a list of references that had diverged in the
	// mirror from the source repository.
	DivergentRefs [][]byte `protobuf:"bytes,1,rep,name=divergent_refs,json=divergentRefs,proto3" json:"divergent_refs,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *UpdateRemoteMirrorResponse) Reset() {
	*x = UpdateRemoteMirrorResponse{}
	mi := &file_remote_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *UpdateRemoteMirrorResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*UpdateRemoteMirrorResponse) ProtoMessage() {}

func (x *UpdateRemoteMirrorResponse) ProtoReflect() protoreflect.Message {
	mi := &file_remote_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use UpdateRemoteMirrorResponse.ProtoReflect.Descriptor instead.
func (*UpdateRemoteMirrorResponse) Descriptor() ([]byte, []int) {
	return file_remote_proto_rawDescGZIP(), []int{1}
}

func (x *UpdateRemoteMirrorResponse) GetDivergentRefs() [][]byte {
	if x != nil {
		return x.DivergentRefs
	}
	return nil
}

// FindRemoteRepositoryRequest is a request for the FindRemoteRepository RPC.
type FindRemoteRepositoryRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// remote is the remote repository being checked. Because this RPC is not executed within a
	// repository, the remote parameter must specify a valid Git URL for the targeted repository.
	Remote string `protobuf:"bytes,1,opt,name=remote,proto3" json:"remote,omitempty"`
	// storage_name is used to redirect request to proper storage where it can be handled.
	// As of now it doesn't matter what storage will be used, but it still must be a valid.
	// For more details: https://gitlab.com/gitlab-org/gitaly/-/issues/2442
	StorageName   string `protobuf:"bytes,2,opt,name=storage_name,json=storageName,proto3" json:"storage_name,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *FindRemoteRepositoryRequest) Reset() {
	*x = FindRemoteRepositoryRequest{}
	mi := &file_remote_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *FindRemoteRepositoryRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*FindRemoteRepositoryRequest) ProtoMessage() {}

func (x *FindRemoteRepositoryRequest) ProtoReflect() protoreflect.Message {
	mi := &file_remote_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use FindRemoteRepositoryRequest.ProtoReflect.Descriptor instead.
func (*FindRemoteRepositoryRequest) Descriptor() ([]byte, []int) {
	return file_remote_proto_rawDescGZIP(), []int{2}
}

func (x *FindRemoteRepositoryRequest) GetRemote() string {
	if x != nil {
		return x.Remote
	}
	return ""
}

func (x *FindRemoteRepositoryRequest) GetStorageName() string {
	if x != nil {
		return x.StorageName
	}
	return ""
}

// FindRemoteRepositoryResponse is a response for the FindRemoteRepository RPC. This might throw a
// GRPC Unavailable code to signal the request failure is transient.
type FindRemoteRepositoryResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// exists specifies if the remote repository exists.
	Exists        bool `protobuf:"varint,1,opt,name=exists,proto3" json:"exists,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *FindRemoteRepositoryResponse) Reset() {
	*x = FindRemoteRepositoryResponse{}
	mi := &file_remote_proto_msgTypes[3]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *FindRemoteRepositoryResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*FindRemoteRepositoryResponse) ProtoMessage() {}

func (x *FindRemoteRepositoryResponse) ProtoReflect() protoreflect.Message {
	mi := &file_remote_proto_msgTypes[3]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use FindRemoteRepositoryResponse.ProtoReflect.Descriptor instead.
func (*FindRemoteRepositoryResponse) Descriptor() ([]byte, []int) {
	return file_remote_proto_rawDescGZIP(), []int{3}
}

func (x *FindRemoteRepositoryResponse) GetExists() bool {
	if x != nil {
		return x.Exists
	}
	return false
}

// FindRemoteRootRefRequest represents a request for the FindRemoteRootRef RPC.
type FindRemoteRootRefRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// repository is the repository in which the request shall be executed in. If
	// a remote name is given, then this is the repository in which the remote
	// will be looked up.
	Repository *Repository `protobuf:"bytes,1,opt,name=repository,proto3" json:"repository,omitempty"`
	// remote_url specifies the remote repository URL which should be fetched from.
	RemoteUrl string `protobuf:"bytes,3,opt,name=remote_url,json=remoteUrl,proto3" json:"remote_url,omitempty"`
	// http_authorization_header is the HTTP header which should be added to the
	// request in order to authenticate against the repository.
	HttpAuthorizationHeader string `protobuf:"bytes,4,opt,name=http_authorization_header,json=httpAuthorizationHeader,proto3" json:"http_authorization_header,omitempty"`
	// resolved_address holds the resolved IP address of the remote_url. This is
	// used to avoid DNS rebinding by mapping the url to the resolved address.
	// Only IPv4 dotted decimal ("192.0.2.1"), IPv6 ("2001:db8::68"), or IPv4-mapped
	// IPv6 ("::ffff:192.0.2.1") forms are supported.
	// Works with HTTP/HTTPS/Git/SSH protocols.
	// Optional.
	ResolvedAddress string `protobuf:"bytes,6,opt,name=resolved_address,json=resolvedAddress,proto3" json:"resolved_address,omitempty"`
	unknownFields   protoimpl.UnknownFields
	sizeCache       protoimpl.SizeCache
}

func (x *FindRemoteRootRefRequest) Reset() {
	*x = FindRemoteRootRefRequest{}
	mi := &file_remote_proto_msgTypes[4]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *FindRemoteRootRefRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*FindRemoteRootRefRequest) ProtoMessage() {}

func (x *FindRemoteRootRefRequest) ProtoReflect() protoreflect.Message {
	mi := &file_remote_proto_msgTypes[4]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use FindRemoteRootRefRequest.ProtoReflect.Descriptor instead.
func (*FindRemoteRootRefRequest) Descriptor() ([]byte, []int) {
	return file_remote_proto_rawDescGZIP(), []int{4}
}

func (x *FindRemoteRootRefRequest) GetRepository() *Repository {
	if x != nil {
		return x.Repository
	}
	return nil
}

func (x *FindRemoteRootRefRequest) GetRemoteUrl() string {
	if x != nil {
		return x.RemoteUrl
	}
	return ""
}

func (x *FindRemoteRootRefRequest) GetHttpAuthorizationHeader() string {
	if x != nil {
		return x.HttpAuthorizationHeader
	}
	return ""
}

func (x *FindRemoteRootRefRequest) GetResolvedAddress() string {
	if x != nil {
		return x.ResolvedAddress
	}
	return ""
}

// FindRemoteRootRefResponse represents the response for the FindRemoteRootRef
// request.
type FindRemoteRootRefResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// ref is the name of the remote root reference.
	Ref           string `protobuf:"bytes,1,opt,name=ref,proto3" json:"ref,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *FindRemoteRootRefResponse) Reset() {
	*x = FindRemoteRootRefResponse{}
	mi := &file_remote_proto_msgTypes[5]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *FindRemoteRootRefResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*FindRemoteRootRefResponse) ProtoMessage() {}

func (x *FindRemoteRootRefResponse) ProtoReflect() protoreflect.Message {
	mi := &file_remote_proto_msgTypes[5]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use FindRemoteRootRefResponse.ProtoReflect.Descriptor instead.
func (*FindRemoteRootRefResponse) Descriptor() ([]byte, []int) {
	return file_remote_proto_rawDescGZIP(), []int{5}
}

func (x *FindRemoteRootRefResponse) GetRef() string {
	if x != nil {
		return x.Ref
	}
	return ""
}

// Remote defines a remote repository that diverged references should be pushed to.
type UpdateRemoteMirrorRequest_Remote struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// url is the URL of the remote repository.
	Url string `protobuf:"bytes,1,opt,name=url,proto3" json:"url,omitempty"`
	// http_authorization_header is an optional HTTP header used for
	// authenticating against the remote repository.
	HttpAuthorizationHeader string `protobuf:"bytes,2,opt,name=http_authorization_header,json=httpAuthorizationHeader,proto3" json:"http_authorization_header,omitempty"`
	// resolved_address holds the resolved IP address of the remote_url. This is
	// used to avoid DNS rebinding by mapping the url to the resolved address.
	// Only IPv4 dotted decimal ("192.0.2.1"), IPv6 ("2001:db8::68"), or IPv4-mapped
	// IPv6 ("::ffff:192.0.2.1") forms are supported.
	// Works with HTTP/HTTPS/Git/SSH protocols.
	// Optional.
	ResolvedAddress string `protobuf:"bytes,4,opt,name=resolved_address,json=resolvedAddress,proto3" json:"resolved_address,omitempty"`
	unknownFields   protoimpl.UnknownFields
	sizeCache       protoimpl.SizeCache
}

func (x *UpdateRemoteMirrorRequest_Remote) Reset() {
	*x = UpdateRemoteMirrorRequest_Remote{}
	mi := &file_remote_proto_msgTypes[6]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *UpdateRemoteMirrorRequest_Remote) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*UpdateRemoteMirrorRequest_Remote) ProtoMessage() {}

func (x *UpdateRemoteMirrorRequest_Remote) ProtoReflect() protoreflect.Message {
	mi := &file_remote_proto_msgTypes[6]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use UpdateRemoteMirrorRequest_Remote.ProtoReflect.Descriptor instead.
func (*UpdateRemoteMirrorRequest_Remote) Descriptor() ([]byte, []int) {
	return file_remote_proto_rawDescGZIP(), []int{0, 0}
}

func (x *UpdateRemoteMirrorRequest_Remote) GetUrl() string {
	if x != nil {
		return x.Url
	}
	return ""
}

func (x *UpdateRemoteMirrorRequest_Remote) GetHttpAuthorizationHeader() string {
	if x != nil {
		return x.HttpAuthorizationHeader
	}
	return ""
}

func (x *UpdateRemoteMirrorRequest_Remote) GetResolvedAddress() string {
	if x != nil {
		return x.ResolvedAddress
	}
	return ""
}

var File_remote_proto protoreflect.FileDescriptor

const file_remote_proto_rawDesc = "" +
	"\n" +
	"\fremote.proto\x12\x06gitaly\x1a\n" +
	"lint.proto\x1a\fshared.proto\"\xdc\x03\n" +
	"\x19UpdateRemoteMirrorRequest\x128\n" +
	"\n" +
	"repository\x18\x01 \x01(\v2\x12.gitaly.RepositoryB\x04\x98\xc6,\x01R\n" +
	"repository\x124\n" +
	"\x16only_branches_matching\x18\x03 \x03(\fR\x14onlyBranchesMatching\x12\x17\n" +
	"\assh_key\x18\x04 \x01(\tR\x06sshKey\x12\x1f\n" +
	"\vknown_hosts\x18\x05 \x01(\tR\n" +
	"knownHosts\x12.\n" +
	"\x13keep_divergent_refs\x18\x06 \x01(\bR\x11keepDivergentRefs\x12@\n" +
	"\x06remote\x18\a \x01(\v2(.gitaly.UpdateRemoteMirrorRequest.RemoteR\x06remote\x1a\x92\x01\n" +
	"\x06Remote\x12\x10\n" +
	"\x03url\x18\x01 \x01(\tR\x03url\x12:\n" +
	"\x19http_authorization_header\x18\x02 \x01(\tR\x17httpAuthorizationHeader\x12)\n" +
	"\x10resolved_address\x18\x04 \x01(\tR\x0fresolvedAddressJ\x04\b\x03\x10\x04R\thttp_hostJ\x04\b\x02\x10\x03R\bref_name\"C\n" +
	"\x1aUpdateRemoteMirrorResponse\x12%\n" +
	"\x0edivergent_refs\x18\x01 \x03(\fR\rdivergentRefs\"^\n" +
	"\x1bFindRemoteRepositoryRequest\x12\x16\n" +
	"\x06remote\x18\x01 \x01(\tR\x06remote\x12'\n" +
	"\fstorage_name\x18\x02 \x01(\tB\x04\x88\xc6,\x01R\vstorageName\"6\n" +
	"\x1cFindRemoteRepositoryResponse\x12\x16\n" +
	"\x06exists\x18\x01 \x01(\bR\x06exists\"\xf9\x01\n" +
	"\x18FindRemoteRootRefRequest\x128\n" +
	"\n" +
	"repository\x18\x01 \x01(\v2\x12.gitaly.RepositoryB\x04\x98\xc6,\x01R\n" +
	"repository\x12\x1d\n" +
	"\n" +
	"remote_url\x18\x03 \x01(\tR\tremoteUrl\x12:\n" +
	"\x19http_authorization_header\x18\x04 \x01(\tR\x17httpAuthorizationHeader\x12)\n" +
	"\x10resolved_address\x18\x06 \x01(\tR\x0fresolvedAddressJ\x04\b\x02\x10\x03J\x04\b\x05\x10\x06R\x06remoteR\thttp_host\"-\n" +
	"\x19FindRemoteRootRefResponse\x12\x10\n" +
	"\x03ref\x18\x01 \x01(\tR\x03ref2\xc5\x02\n" +
	"\rRemoteService\x12e\n" +
	"\x12UpdateRemoteMirror\x12!.gitaly.UpdateRemoteMirrorRequest\x1a\".gitaly.UpdateRemoteMirrorResponse\"\x06\xfa\x97(\x02\b\x02(\x01\x12k\n" +
	"\x14FindRemoteRepository\x12#.gitaly.FindRemoteRepositoryRequest\x1a$.gitaly.FindRemoteRepositoryResponse\"\b\xfa\x97(\x04\b\x02\x10\x02\x12`\n" +
	"\x11FindRemoteRootRef\x12 .gitaly.FindRemoteRootRefRequest\x1a!.gitaly.FindRemoteRootRefResponse\"\x06\xfa\x97(\x02\b\x02B4Z2gitlab.com/gitlab-org/gitaly/v18/proto/go/gitalypbb\x06proto3"

var (
	file_remote_proto_rawDescOnce sync.Once
	file_remote_proto_rawDescData []byte
)

func file_remote_proto_rawDescGZIP() []byte {
	file_remote_proto_rawDescOnce.Do(func() {
		file_remote_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_remote_proto_rawDesc), len(file_remote_proto_rawDesc)))
	})
	return file_remote_proto_rawDescData
}

var file_remote_proto_msgTypes = make([]protoimpl.MessageInfo, 7)
var file_remote_proto_goTypes = []any{
	(*UpdateRemoteMirrorRequest)(nil),        // 0: gitaly.UpdateRemoteMirrorRequest
	(*UpdateRemoteMirrorResponse)(nil),       // 1: gitaly.UpdateRemoteMirrorResponse
	(*FindRemoteRepositoryRequest)(nil),      // 2: gitaly.FindRemoteRepositoryRequest
	(*FindRemoteRepositoryResponse)(nil),     // 3: gitaly.FindRemoteRepositoryResponse
	(*FindRemoteRootRefRequest)(nil),         // 4: gitaly.FindRemoteRootRefRequest
	(*FindRemoteRootRefResponse)(nil),        // 5: gitaly.FindRemoteRootRefResponse
	(*UpdateRemoteMirrorRequest_Remote)(nil), // 6: gitaly.UpdateRemoteMirrorRequest.Remote
	(*Repository)(nil),                       // 7: gitaly.Repository
}
var file_remote_proto_depIdxs = []int32{
	7, // 0: gitaly.UpdateRemoteMirrorRequest.repository:type_name -> gitaly.Repository
	6, // 1: gitaly.UpdateRemoteMirrorRequest.remote:type_name -> gitaly.UpdateRemoteMirrorRequest.Remote
	7, // 2: gitaly.FindRemoteRootRefRequest.repository:type_name -> gitaly.Repository
	0, // 3: gitaly.RemoteService.UpdateRemoteMirror:input_type -> gitaly.UpdateRemoteMirrorRequest
	2, // 4: gitaly.RemoteService.FindRemoteRepository:input_type -> gitaly.FindRemoteRepositoryRequest
	4, // 5: gitaly.RemoteService.FindRemoteRootRef:input_type -> gitaly.FindRemoteRootRefRequest
	1, // 6: gitaly.RemoteService.UpdateRemoteMirror:output_type -> gitaly.UpdateRemoteMirrorResponse
	3, // 7: gitaly.RemoteService.FindRemoteRepository:output_type -> gitaly.FindRemoteRepositoryResponse
	5, // 8: gitaly.RemoteService.FindRemoteRootRef:output_type -> gitaly.FindRemoteRootRefResponse
	6, // [6:9] is the sub-list for method output_type
	3, // [3:6] is the sub-list for method input_type
	3, // [3:3] is the sub-list for extension type_name
	3, // [3:3] is the sub-list for extension extendee
	0, // [0:3] is the sub-list for field type_name
}

func init() { file_remote_proto_init() }
func file_remote_proto_init() {
	if File_remote_proto != nil {
		return
	}
	file_lint_proto_init()
	file_shared_proto_init()
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_remote_proto_rawDesc), len(file_remote_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   7,
			NumExtensions: 0,
			NumServices:   1,
		},
		GoTypes:           file_remote_proto_goTypes,
		DependencyIndexes: file_remote_proto_depIdxs,
		MessageInfos:      file_remote_proto_msgTypes,
	}.Build()
	File_remote_proto = out.File
	file_remote_proto_goTypes = nil
	file_remote_proto_depIdxs = nil
}
