// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v6.30.2
// source: backup.proto

package gitalypb

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

// WriteBackupIDRequest is a request for the WriteBackupID RPC.
type WriteBackupIDRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// storage_name is the name of the storage to route this request to.
	StorageName string `protobuf:"bytes,1,opt,name=storage_name,json=storageName,proto3" json:"storage_name,omitempty"`
	// backup_id is the label used to identify backups.
	BackupId      string `protobuf:"bytes,2,opt,name=backup_id,json=backupId,proto3" json:"backup_id,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *WriteBackupIDRequest) Reset() {
	*x = WriteBackupIDRequest{}
	mi := &file_backup_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *WriteBackupIDRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*WriteBackupIDRequest) ProtoMessage() {}

func (x *WriteBackupIDRequest) ProtoReflect() protoreflect.Message {
	mi := &file_backup_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use WriteBackupIDRequest.ProtoReflect.Descriptor instead.
func (*WriteBackupIDRequest) Descriptor() ([]byte, []int) {
	return file_backup_proto_rawDescGZIP(), []int{0}
}

func (x *WriteBackupIDRequest) GetStorageName() string {
	if x != nil {
		return x.StorageName
	}
	return ""
}

func (x *WriteBackupIDRequest) GetBackupId() string {
	if x != nil {
		return x.BackupId
	}
	return ""
}

// WriteBackupIDResponse is a response for the WriteBackupID RPC.
type WriteBackupIDResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *WriteBackupIDResponse) Reset() {
	*x = WriteBackupIDResponse{}
	mi := &file_backup_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *WriteBackupIDResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*WriteBackupIDResponse) ProtoMessage() {}

func (x *WriteBackupIDResponse) ProtoReflect() protoreflect.Message {
	mi := &file_backup_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use WriteBackupIDResponse.ProtoReflect.Descriptor instead.
func (*WriteBackupIDResponse) Descriptor() ([]byte, []int) {
	return file_backup_proto_rawDescGZIP(), []int{1}
}

// ReadLatestBackupIDRequest is a request for the ReadLatestBackupID RPC.
type ReadLatestBackupIDRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// storage_name is the name of the storage to route this request to.
	StorageName   string `protobuf:"bytes,1,opt,name=storage_name,json=storageName,proto3" json:"storage_name,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ReadLatestBackupIDRequest) Reset() {
	*x = ReadLatestBackupIDRequest{}
	mi := &file_backup_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ReadLatestBackupIDRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ReadLatestBackupIDRequest) ProtoMessage() {}

func (x *ReadLatestBackupIDRequest) ProtoReflect() protoreflect.Message {
	mi := &file_backup_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ReadLatestBackupIDRequest.ProtoReflect.Descriptor instead.
func (*ReadLatestBackupIDRequest) Descriptor() ([]byte, []int) {
	return file_backup_proto_rawDescGZIP(), []int{2}
}

func (x *ReadLatestBackupIDRequest) GetStorageName() string {
	if x != nil {
		return x.StorageName
	}
	return ""
}

// ReadLatestBackupIDResponse is a response for the ReadLatestBackupID RPC.
type ReadLatestBackupIDResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// backup_id is the label used to identify backups.
	BackupId      string `protobuf:"bytes,1,opt,name=backup_id,json=backupId,proto3" json:"backup_id,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ReadLatestBackupIDResponse) Reset() {
	*x = ReadLatestBackupIDResponse{}
	mi := &file_backup_proto_msgTypes[3]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ReadLatestBackupIDResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ReadLatestBackupIDResponse) ProtoMessage() {}

func (x *ReadLatestBackupIDResponse) ProtoReflect() protoreflect.Message {
	mi := &file_backup_proto_msgTypes[3]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ReadLatestBackupIDResponse.ProtoReflect.Descriptor instead.
func (*ReadLatestBackupIDResponse) Descriptor() ([]byte, []int) {
	return file_backup_proto_rawDescGZIP(), []int{3}
}

func (x *ReadLatestBackupIDResponse) GetBackupId() string {
	if x != nil {
		return x.BackupId
	}
	return ""
}

var File_backup_proto protoreflect.FileDescriptor

const file_backup_proto_rawDesc = "" +
	"\n" +
	"\fbackup.proto\x12\x06gitaly\x1a\n" +
	"lint.proto\"\\\n" +
	"\x14WriteBackupIDRequest\x12'\n" +
	"\fstorage_name\x18\x01 \x01(\tB\x04\x88\xc6,\x01R\vstorageName\x12\x1b\n" +
	"\tbackup_id\x18\x02 \x01(\tR\bbackupId\"\x17\n" +
	"\x15WriteBackupIDResponse\"D\n" +
	"\x19ReadLatestBackupIDRequest\x12'\n" +
	"\fstorage_name\x18\x01 \x01(\tB\x04\x88\xc6,\x01R\vstorageName\"9\n" +
	"\x1aReadLatestBackupIDResponse\x12\x1b\n" +
	"\tbackup_id\x18\x01 \x01(\tR\bbackupId2\xce\x01\n" +
	"\rBackupService\x12V\n" +
	"\rWriteBackupID\x12\x1c.gitaly.WriteBackupIDRequest\x1a\x1d.gitaly.WriteBackupIDResponse\"\b\xfa\x97(\x04\b\x02\x10\x02\x12e\n" +
	"\x12ReadLatestBackupID\x12!.gitaly.ReadLatestBackupIDRequest\x1a\".gitaly.ReadLatestBackupIDResponse\"\b\xfa\x97(\x04\b\x02\x10\x02B4Z2gitlab.com/gitlab-org/gitaly/v18/proto/go/gitalypbb\x06proto3"

var (
	file_backup_proto_rawDescOnce sync.Once
	file_backup_proto_rawDescData []byte
)

func file_backup_proto_rawDescGZIP() []byte {
	file_backup_proto_rawDescOnce.Do(func() {
		file_backup_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_backup_proto_rawDesc), len(file_backup_proto_rawDesc)))
	})
	return file_backup_proto_rawDescData
}

var file_backup_proto_msgTypes = make([]protoimpl.MessageInfo, 4)
var file_backup_proto_goTypes = []any{
	(*WriteBackupIDRequest)(nil),       // 0: gitaly.WriteBackupIDRequest
	(*WriteBackupIDResponse)(nil),      // 1: gitaly.WriteBackupIDResponse
	(*ReadLatestBackupIDRequest)(nil),  // 2: gitaly.ReadLatestBackupIDRequest
	(*ReadLatestBackupIDResponse)(nil), // 3: gitaly.ReadLatestBackupIDResponse
}
var file_backup_proto_depIdxs = []int32{
	0, // 0: gitaly.BackupService.WriteBackupID:input_type -> gitaly.WriteBackupIDRequest
	2, // 1: gitaly.BackupService.ReadLatestBackupID:input_type -> gitaly.ReadLatestBackupIDRequest
	1, // 2: gitaly.BackupService.WriteBackupID:output_type -> gitaly.WriteBackupIDResponse
	3, // 3: gitaly.BackupService.ReadLatestBackupID:output_type -> gitaly.ReadLatestBackupIDResponse
	2, // [2:4] is the sub-list for method output_type
	0, // [0:2] is the sub-list for method input_type
	0, // [0:0] is the sub-list for extension type_name
	0, // [0:0] is the sub-list for extension extendee
	0, // [0:0] is the sub-list for field type_name
}

func init() { file_backup_proto_init() }
func file_backup_proto_init() {
	if File_backup_proto != nil {
		return
	}
	file_lint_proto_init()
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_backup_proto_rawDesc), len(file_backup_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   4,
			NumExtensions: 0,
			NumServices:   1,
		},
		GoTypes:           file_backup_proto_goTypes,
		DependencyIndexes: file_backup_proto_depIdxs,
		MessageInfos:      file_backup_proto_msgTypes,
	}.Build()
	File_backup_proto = out.File
	file_backup_proto_goTypes = nil
	file_backup_proto_depIdxs = nil
}
