// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v6.30.2
// source: testproto/invalid.proto

package testproto

import (
	gitalypb "gitlab.com/gitlab-org/gitaly/v18/proto/go/gitalypb"
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

type InvalidMethodRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *InvalidMethodRequest) Reset() {
	*x = InvalidMethodRequest{}
	mi := &file_testproto_invalid_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *InvalidMethodRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*InvalidMethodRequest) ProtoMessage() {}

func (x *InvalidMethodRequest) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_invalid_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use InvalidMethodRequest.ProtoReflect.Descriptor instead.
func (*InvalidMethodRequest) Descriptor() ([]byte, []int) {
	return file_testproto_invalid_proto_rawDescGZIP(), []int{0}
}

type InvalidMethodRequestWithRepo struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Destination   *gitalypb.Repository   `protobuf:"bytes,1,opt,name=destination,proto3" json:"destination,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *InvalidMethodRequestWithRepo) Reset() {
	*x = InvalidMethodRequestWithRepo{}
	mi := &file_testproto_invalid_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *InvalidMethodRequestWithRepo) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*InvalidMethodRequestWithRepo) ProtoMessage() {}

func (x *InvalidMethodRequestWithRepo) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_invalid_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use InvalidMethodRequestWithRepo.ProtoReflect.Descriptor instead.
func (*InvalidMethodRequestWithRepo) Descriptor() ([]byte, []int) {
	return file_testproto_invalid_proto_rawDescGZIP(), []int{1}
}

func (x *InvalidMethodRequestWithRepo) GetDestination() *gitalypb.Repository {
	if x != nil {
		return x.Destination
	}
	return nil
}

type InvalidTargetType struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	WrongType     int32                  `protobuf:"varint,1,opt,name=wrong_type,json=wrongType,proto3" json:"wrong_type,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *InvalidTargetType) Reset() {
	*x = InvalidTargetType{}
	mi := &file_testproto_invalid_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *InvalidTargetType) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*InvalidTargetType) ProtoMessage() {}

func (x *InvalidTargetType) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_invalid_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use InvalidTargetType.ProtoReflect.Descriptor instead.
func (*InvalidTargetType) Descriptor() ([]byte, []int) {
	return file_testproto_invalid_proto_rawDescGZIP(), []int{2}
}

func (x *InvalidTargetType) GetWrongType() int32 {
	if x != nil {
		return x.WrongType
	}
	return 0
}

type InvalidMethodResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *InvalidMethodResponse) Reset() {
	*x = InvalidMethodResponse{}
	mi := &file_testproto_invalid_proto_msgTypes[3]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *InvalidMethodResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*InvalidMethodResponse) ProtoMessage() {}

func (x *InvalidMethodResponse) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_invalid_proto_msgTypes[3]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use InvalidMethodResponse.ProtoReflect.Descriptor instead.
func (*InvalidMethodResponse) Descriptor() ([]byte, []int) {
	return file_testproto_invalid_proto_rawDescGZIP(), []int{3}
}

type InvalidNestedRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	InnerMessage  *InvalidTargetType     `protobuf:"bytes,1,opt,name=inner_message,json=innerMessage,proto3" json:"inner_message,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *InvalidNestedRequest) Reset() {
	*x = InvalidNestedRequest{}
	mi := &file_testproto_invalid_proto_msgTypes[4]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *InvalidNestedRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*InvalidNestedRequest) ProtoMessage() {}

func (x *InvalidNestedRequest) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_invalid_proto_msgTypes[4]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use InvalidNestedRequest.ProtoReflect.Descriptor instead.
func (*InvalidNestedRequest) Descriptor() ([]byte, []int) {
	return file_testproto_invalid_proto_rawDescGZIP(), []int{4}
}

func (x *InvalidNestedRequest) GetInnerMessage() *InvalidTargetType {
	if x != nil {
		return x.InnerMessage
	}
	return nil
}

type RequestWithStorage struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	StorageName   string                 `protobuf:"bytes,1,opt,name=storage_name,json=storageName,proto3" json:"storage_name,omitempty"`
	Destination   *gitalypb.Repository   `protobuf:"bytes,2,opt,name=destination,proto3" json:"destination,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *RequestWithStorage) Reset() {
	*x = RequestWithStorage{}
	mi := &file_testproto_invalid_proto_msgTypes[5]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *RequestWithStorage) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*RequestWithStorage) ProtoMessage() {}

func (x *RequestWithStorage) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_invalid_proto_msgTypes[5]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use RequestWithStorage.ProtoReflect.Descriptor instead.
func (*RequestWithStorage) Descriptor() ([]byte, []int) {
	return file_testproto_invalid_proto_rawDescGZIP(), []int{5}
}

func (x *RequestWithStorage) GetStorageName() string {
	if x != nil {
		return x.StorageName
	}
	return ""
}

func (x *RequestWithStorage) GetDestination() *gitalypb.Repository {
	if x != nil {
		return x.Destination
	}
	return nil
}

type RequestWithStorageAndRepo struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	StorageName   string                 `protobuf:"bytes,1,opt,name=storage_name,json=storageName,proto3" json:"storage_name,omitempty"`
	Destination   *gitalypb.Repository   `protobuf:"bytes,2,opt,name=destination,proto3" json:"destination,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *RequestWithStorageAndRepo) Reset() {
	*x = RequestWithStorageAndRepo{}
	mi := &file_testproto_invalid_proto_msgTypes[6]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *RequestWithStorageAndRepo) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*RequestWithStorageAndRepo) ProtoMessage() {}

func (x *RequestWithStorageAndRepo) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_invalid_proto_msgTypes[6]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use RequestWithStorageAndRepo.ProtoReflect.Descriptor instead.
func (*RequestWithStorageAndRepo) Descriptor() ([]byte, []int) {
	return file_testproto_invalid_proto_rawDescGZIP(), []int{6}
}

func (x *RequestWithStorageAndRepo) GetStorageName() string {
	if x != nil {
		return x.StorageName
	}
	return ""
}

func (x *RequestWithStorageAndRepo) GetDestination() *gitalypb.Repository {
	if x != nil {
		return x.Destination
	}
	return nil
}

type RequestWithNestedStorageAndRepo struct {
	state         protoimpl.MessageState     `protogen:"open.v1"`
	InnerMessage  *RequestWithStorageAndRepo `protobuf:"bytes,1,opt,name=inner_message,json=innerMessage,proto3" json:"inner_message,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *RequestWithNestedStorageAndRepo) Reset() {
	*x = RequestWithNestedStorageAndRepo{}
	mi := &file_testproto_invalid_proto_msgTypes[7]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *RequestWithNestedStorageAndRepo) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*RequestWithNestedStorageAndRepo) ProtoMessage() {}

func (x *RequestWithNestedStorageAndRepo) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_invalid_proto_msgTypes[7]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use RequestWithNestedStorageAndRepo.ProtoReflect.Descriptor instead.
func (*RequestWithNestedStorageAndRepo) Descriptor() ([]byte, []int) {
	return file_testproto_invalid_proto_rawDescGZIP(), []int{7}
}

func (x *RequestWithNestedStorageAndRepo) GetInnerMessage() *RequestWithStorageAndRepo {
	if x != nil {
		return x.InnerMessage
	}
	return nil
}

type RequestWithMultipleNestedStorage struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	InnerMessage  *RequestWithStorage    `protobuf:"bytes,1,opt,name=inner_message,json=innerMessage,proto3" json:"inner_message,omitempty"`
	StorageName   string                 `protobuf:"bytes,2,opt,name=storage_name,json=storageName,proto3" json:"storage_name,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *RequestWithMultipleNestedStorage) Reset() {
	*x = RequestWithMultipleNestedStorage{}
	mi := &file_testproto_invalid_proto_msgTypes[8]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *RequestWithMultipleNestedStorage) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*RequestWithMultipleNestedStorage) ProtoMessage() {}

func (x *RequestWithMultipleNestedStorage) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_invalid_proto_msgTypes[8]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use RequestWithMultipleNestedStorage.ProtoReflect.Descriptor instead.
func (*RequestWithMultipleNestedStorage) Descriptor() ([]byte, []int) {
	return file_testproto_invalid_proto_rawDescGZIP(), []int{8}
}

func (x *RequestWithMultipleNestedStorage) GetInnerMessage() *RequestWithStorage {
	if x != nil {
		return x.InnerMessage
	}
	return nil
}

func (x *RequestWithMultipleNestedStorage) GetStorageName() string {
	if x != nil {
		return x.StorageName
	}
	return ""
}

type RequestWithInnerNestedStorage struct {
	state         protoimpl.MessageState                `protogen:"open.v1"`
	Header        *RequestWithInnerNestedStorage_Header `protobuf:"bytes,1,opt,name=header,proto3" json:"header,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *RequestWithInnerNestedStorage) Reset() {
	*x = RequestWithInnerNestedStorage{}
	mi := &file_testproto_invalid_proto_msgTypes[9]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *RequestWithInnerNestedStorage) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*RequestWithInnerNestedStorage) ProtoMessage() {}

func (x *RequestWithInnerNestedStorage) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_invalid_proto_msgTypes[9]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use RequestWithInnerNestedStorage.ProtoReflect.Descriptor instead.
func (*RequestWithInnerNestedStorage) Descriptor() ([]byte, []int) {
	return file_testproto_invalid_proto_rawDescGZIP(), []int{9}
}

func (x *RequestWithInnerNestedStorage) GetHeader() *RequestWithInnerNestedStorage_Header {
	if x != nil {
		return x.Header
	}
	return nil
}

type RequestWithWrongTypeRepository struct {
	state         protoimpl.MessageState                 `protogen:"open.v1"`
	Header        *RequestWithWrongTypeRepository_Header `protobuf:"bytes,1,opt,name=header,proto3" json:"header,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *RequestWithWrongTypeRepository) Reset() {
	*x = RequestWithWrongTypeRepository{}
	mi := &file_testproto_invalid_proto_msgTypes[10]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *RequestWithWrongTypeRepository) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*RequestWithWrongTypeRepository) ProtoMessage() {}

func (x *RequestWithWrongTypeRepository) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_invalid_proto_msgTypes[10]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use RequestWithWrongTypeRepository.ProtoReflect.Descriptor instead.
func (*RequestWithWrongTypeRepository) Descriptor() ([]byte, []int) {
	return file_testproto_invalid_proto_rawDescGZIP(), []int{10}
}

func (x *RequestWithWrongTypeRepository) GetHeader() *RequestWithWrongTypeRepository_Header {
	if x != nil {
		return x.Header
	}
	return nil
}

type RequestWithNestedRepoNotFlagged struct {
	state         protoimpl.MessageState                  `protogen:"open.v1"`
	Header        *RequestWithNestedRepoNotFlagged_Header `protobuf:"bytes,1,opt,name=header,proto3" json:"header,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *RequestWithNestedRepoNotFlagged) Reset() {
	*x = RequestWithNestedRepoNotFlagged{}
	mi := &file_testproto_invalid_proto_msgTypes[11]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *RequestWithNestedRepoNotFlagged) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*RequestWithNestedRepoNotFlagged) ProtoMessage() {}

func (x *RequestWithNestedRepoNotFlagged) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_invalid_proto_msgTypes[11]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use RequestWithNestedRepoNotFlagged.ProtoReflect.Descriptor instead.
func (*RequestWithNestedRepoNotFlagged) Descriptor() ([]byte, []int) {
	return file_testproto_invalid_proto_rawDescGZIP(), []int{11}
}

func (x *RequestWithNestedRepoNotFlagged) GetHeader() *RequestWithNestedRepoNotFlagged_Header {
	if x != nil {
		return x.Header
	}
	return nil
}

type RequestWithPartition struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	StorageName   string                 `protobuf:"bytes,1,opt,name=storage_name,json=storageName,proto3" json:"storage_name,omitempty"`
	PartitionId   uint64                 `protobuf:"varint,2,opt,name=partition_id,json=partitionId,proto3" json:"partition_id,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *RequestWithPartition) Reset() {
	*x = RequestWithPartition{}
	mi := &file_testproto_invalid_proto_msgTypes[12]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *RequestWithPartition) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*RequestWithPartition) ProtoMessage() {}

func (x *RequestWithPartition) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_invalid_proto_msgTypes[12]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use RequestWithPartition.ProtoReflect.Descriptor instead.
func (*RequestWithPartition) Descriptor() ([]byte, []int) {
	return file_testproto_invalid_proto_rawDescGZIP(), []int{12}
}

func (x *RequestWithPartition) GetStorageName() string {
	if x != nil {
		return x.StorageName
	}
	return ""
}

func (x *RequestWithPartition) GetPartitionId() uint64 {
	if x != nil {
		return x.PartitionId
	}
	return 0
}

type RequestWithInnerNestedStorage_Header struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	StorageName   string                 `protobuf:"bytes,1,opt,name=storage_name,json=storageName,proto3" json:"storage_name,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *RequestWithInnerNestedStorage_Header) Reset() {
	*x = RequestWithInnerNestedStorage_Header{}
	mi := &file_testproto_invalid_proto_msgTypes[13]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *RequestWithInnerNestedStorage_Header) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*RequestWithInnerNestedStorage_Header) ProtoMessage() {}

func (x *RequestWithInnerNestedStorage_Header) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_invalid_proto_msgTypes[13]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use RequestWithInnerNestedStorage_Header.ProtoReflect.Descriptor instead.
func (*RequestWithInnerNestedStorage_Header) Descriptor() ([]byte, []int) {
	return file_testproto_invalid_proto_rawDescGZIP(), []int{9, 0}
}

func (x *RequestWithInnerNestedStorage_Header) GetStorageName() string {
	if x != nil {
		return x.StorageName
	}
	return ""
}

type RequestWithWrongTypeRepository_Header struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Repository    *InvalidMethodResponse `protobuf:"bytes,1,opt,name=repository,proto3" json:"repository,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *RequestWithWrongTypeRepository_Header) Reset() {
	*x = RequestWithWrongTypeRepository_Header{}
	mi := &file_testproto_invalid_proto_msgTypes[14]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *RequestWithWrongTypeRepository_Header) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*RequestWithWrongTypeRepository_Header) ProtoMessage() {}

func (x *RequestWithWrongTypeRepository_Header) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_invalid_proto_msgTypes[14]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use RequestWithWrongTypeRepository_Header.ProtoReflect.Descriptor instead.
func (*RequestWithWrongTypeRepository_Header) Descriptor() ([]byte, []int) {
	return file_testproto_invalid_proto_rawDescGZIP(), []int{10, 0}
}

func (x *RequestWithWrongTypeRepository_Header) GetRepository() *InvalidMethodResponse {
	if x != nil {
		return x.Repository
	}
	return nil
}

type RequestWithNestedRepoNotFlagged_Header struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Repository    *gitalypb.Repository   `protobuf:"bytes,1,opt,name=repository,proto3" json:"repository,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *RequestWithNestedRepoNotFlagged_Header) Reset() {
	*x = RequestWithNestedRepoNotFlagged_Header{}
	mi := &file_testproto_invalid_proto_msgTypes[15]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *RequestWithNestedRepoNotFlagged_Header) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*RequestWithNestedRepoNotFlagged_Header) ProtoMessage() {}

func (x *RequestWithNestedRepoNotFlagged_Header) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_invalid_proto_msgTypes[15]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use RequestWithNestedRepoNotFlagged_Header.ProtoReflect.Descriptor instead.
func (*RequestWithNestedRepoNotFlagged_Header) Descriptor() ([]byte, []int) {
	return file_testproto_invalid_proto_rawDescGZIP(), []int{11, 0}
}

func (x *RequestWithNestedRepoNotFlagged_Header) GetRepository() *gitalypb.Repository {
	if x != nil {
		return x.Repository
	}
	return nil
}

var File_testproto_invalid_proto protoreflect.FileDescriptor

const file_testproto_invalid_proto_rawDesc = "" +
	"\n" +
	"\x17testproto/invalid.proto\x12\ttestproto\x1a\n" +
	"lint.proto\x1a\fshared.proto\"\x16\n" +
	"\x14InvalidMethodRequest\"Z\n" +
	"\x1cInvalidMethodRequestWithRepo\x12:\n" +
	"\vdestination\x18\x01 \x01(\v2\x12.gitaly.RepositoryB\x04\x98\xc6,\x01R\vdestination\"8\n" +
	"\x11InvalidTargetType\x12#\n" +
	"\n" +
	"wrong_type\x18\x01 \x01(\x05B\x04\x98\xc6,\x01R\twrongType\"\x17\n" +
	"\x15InvalidMethodResponse\"Y\n" +
	"\x14InvalidNestedRequest\x12A\n" +
	"\rinner_message\x18\x01 \x01(\v2\x1c.testproto.InvalidTargetTypeR\finnerMessage\"s\n" +
	"\x12RequestWithStorage\x12'\n" +
	"\fstorage_name\x18\x01 \x01(\tB\x04\x88\xc6,\x01R\vstorageName\x124\n" +
	"\vdestination\x18\x02 \x01(\v2\x12.gitaly.RepositoryR\vdestination\"\x80\x01\n" +
	"\x19RequestWithStorageAndRepo\x12'\n" +
	"\fstorage_name\x18\x01 \x01(\tB\x04\x88\xc6,\x01R\vstorageName\x12:\n" +
	"\vdestination\x18\x02 \x01(\v2\x12.gitaly.RepositoryB\x04\x98\xc6,\x01R\vdestination\"l\n" +
	"\x1fRequestWithNestedStorageAndRepo\x12I\n" +
	"\rinner_message\x18\x01 \x01(\v2$.testproto.RequestWithStorageAndRepoR\finnerMessage\"\x8f\x01\n" +
	" RequestWithMultipleNestedStorage\x12B\n" +
	"\rinner_message\x18\x01 \x01(\v2\x1d.testproto.RequestWithStorageR\finnerMessage\x12'\n" +
	"\fstorage_name\x18\x02 \x01(\tB\x04\x88\xc6,\x01R\vstorageName\"\x9b\x01\n" +
	"\x1dRequestWithInnerNestedStorage\x12G\n" +
	"\x06header\x18\x01 \x01(\v2/.testproto.RequestWithInnerNestedStorage.HeaderR\x06header\x1a1\n" +
	"\x06Header\x12'\n" +
	"\fstorage_name\x18\x01 \x01(\tB\x04\x88\xc6,\x01R\vstorageName\"\xc2\x01\n" +
	"\x1eRequestWithWrongTypeRepository\x12N\n" +
	"\x06header\x18\x01 \x01(\v20.testproto.RequestWithWrongTypeRepository.HeaderB\x04\x98\xc6,\x01R\x06header\x1aP\n" +
	"\x06Header\x12F\n" +
	"\n" +
	"repository\x18\x01 \x01(\v2 .testproto.InvalidMethodResponseB\x04\x90\xc6,\x01R\n" +
	"repository\"\xb0\x01\n" +
	"\x1fRequestWithNestedRepoNotFlagged\x12O\n" +
	"\x06header\x18\x01 \x01(\v21.testproto.RequestWithNestedRepoNotFlagged.HeaderB\x04\x98\xc6,\x01R\x06header\x1a<\n" +
	"\x06Header\x122\n" +
	"\n" +
	"repository\x18\x01 \x01(\v2\x12.gitaly.RepositoryR\n" +
	"repository\"h\n" +
	"\x14RequestWithPartition\x12'\n" +
	"\fstorage_name\x18\x01 \x01(\tB\x04\x88\xc6,\x01R\vstorageName\x12'\n" +
	"\fpartition_id\x18\x02 \x01(\x04B\x04\xa8\xc6,\x01R\vpartitionId2\x80\x01\n" +
	"\x1cInterceptedWithOperationType\x12Z\n" +
	"\rInvalidMethod\x12\x1f.testproto.InvalidMethodRequest\x1a .testproto.InvalidMethodResponse\"\x06\xfa\x97(\x02\b\x02\x1a\x04\xf0\x97(\x012\x9b\x15\n" +
	"\x0eInvalidService\x12U\n" +
	"\x0eInvalidMethod0\x12\x1f.testproto.InvalidMethodRequest\x1a .testproto.InvalidMethodResponse\"\x00\x12Y\n" +
	"\x0eInvalidMethod1\x12\x1f.testproto.InvalidMethodRequest\x1a .testproto.InvalidMethodResponse\"\x04\xfa\x97(\x00\x12[\n" +
	"\x0eInvalidMethod2\x12\x1f.testproto.InvalidMethodRequest\x1a .testproto.InvalidMethodResponse\"\x06\xfa\x97(\x02\b\x02\x12[\n" +
	"\x0eInvalidMethod4\x12\x1f.testproto.InvalidMethodRequest\x1a .testproto.InvalidMethodResponse\"\x06\xfa\x97(\x02\b\x01\x12e\n" +
	"\x0eInvalidMethod5\x12).testproto.RequestWithWrongTypeRepository\x1a .testproto.InvalidMethodResponse\"\x06\xfa\x97(\x02\b\x01\x12f\n" +
	"\x0eInvalidMethod6\x12*.testproto.RequestWithNestedRepoNotFlagged\x1a .testproto.InvalidMethodResponse\"\x06\xfa\x97(\x02\b\x01\x12X\n" +
	"\x0eInvalidMethod7\x12\x1c.testproto.InvalidTargetType\x1a .testproto.InvalidMethodResponse\"\x06\xfa\x97(\x02\b\x01\x12[\n" +
	"\x0eInvalidMethod8\x12\x1f.testproto.InvalidNestedRequest\x1a .testproto.InvalidMethodResponse\"\x06\xfa\x97(\x02\b\x01\x12e\n" +
	"\x0eInvalidMethod9\x12'.testproto.InvalidMethodRequestWithRepo\x1a .testproto.InvalidMethodResponse\"\b\xfa\x97(\x04\b\x01\x10\x02\x12a\n" +
	"\x0fInvalidMethod10\x12$.testproto.RequestWithStorageAndRepo\x1a .testproto.InvalidMethodResponse\"\x06\xfa\x97(\x02\b\x02\x12g\n" +
	"\x0fInvalidMethod11\x12*.testproto.RequestWithNestedStorageAndRepo\x1a .testproto.InvalidMethodResponse\"\x06\xfa\x97(\x02\b\x01\x12[\n" +
	"\x0fInvalidMethod13\x12\x1c.testproto.InvalidTargetType\x1a .testproto.InvalidMethodResponse\"\b\xfa\x97(\x04\b\x01\x10\x02\x12j\n" +
	"\x0fInvalidMethod14\x12+.testproto.RequestWithMultipleNestedStorage\x1a .testproto.InvalidMethodResponse\"\b\xfa\x97(\x04\b\x01\x10\x02\x12^\n" +
	"\x0fInvalidMethod15\x12\x1f.testproto.InvalidMethodRequest\x1a .testproto.InvalidMethodResponse\"\b\xfa\x97(\x04\b\x02\x10\x03\x12f\n" +
	"\x0fInvalidMethod16\x12'.testproto.InvalidMethodRequestWithRepo\x1a .testproto.InvalidMethodResponse\"\b\xfa\x97(\x04\b\x02\x10\x03\x12\\\n" +
	"\x0fInvalidMethod17\x12\x1d.testproto.RequestWithStorage\x1a .testproto.InvalidMethodResponse\"\b\xfa\x97(\x04\b\x02\x10\x03\x12\\\n" +
	"\x0fInvalidMethod18\x12\x1f.testproto.RequestWithPartition\x1a .testproto.InvalidMethodResponse\"\x06\xfa\x97(\x02\b\x02\x12^\n" +
	"\x0fInvalidMethod19\x12\x1f.testproto.RequestWithPartition\x1a .testproto.InvalidMethodResponse\"\b\xfa\x97(\x04\b\x02\x10\x02\x12m\n" +
	" MaintenanceWithMissingRepository\x12\x1f.testproto.InvalidMethodRequest\x1a .testproto.InvalidMethodResponse\"\x06\xfa\x97(\x02\b\x03\x12z\n" +
	"\"MaintenanceWithUnflaggedRepository\x12*.testproto.RequestWithNestedRepoNotFlagged\x1a .testproto.InvalidMethodResponse\"\x06\xfa\x97(\x02\b\x03\x12\x7f\n" +
	"(MaintenanceWithWrongNestedRepositoryType\x12).testproto.RequestWithWrongTypeRepository\x1a .testproto.InvalidMethodResponse\"\x06\xfa\x97(\x02\b\x03\x12j\n" +
	" MaintenanceWithInvalidTargetType\x12\x1c.testproto.InvalidTargetType\x1a .testproto.InvalidMethodResponse\"\x06\xfa\x97(\x02\b\x03\x12p\n" +
	"#MaintenanceWithInvalidNestedRequest\x12\x1f.testproto.InvalidNestedRequest\x1a .testproto.InvalidMethodResponse\"\x06\xfa\x97(\x02\b\x03\x12u\n" +
	"#MaintenanceWithStorageAndRepository\x12$.testproto.RequestWithStorageAndRepo\x1a .testproto.InvalidMethodResponse\"\x06\xfa\x97(\x02\b\x03\x12\x81\x01\n" +
	")MaintenanceWithNestedStorageAndRepository\x12*.testproto.RequestWithNestedStorageAndRepo\x1a .testproto.InvalidMethodResponse\"\x06\xfa\x97(\x02\b\x03\x12r\n" +
	"\x1bMaintenanceWithStorageScope\x12'.testproto.InvalidMethodRequestWithRepo\x1a .testproto.InvalidMethodResponse\"\b\xfa\x97(\x04\b\x03\x10\x02B>Z<gitlab.com/gitlab-org/gitaly/v18/proto/go/gitalypb/testprotob\x06proto3"

var (
	file_testproto_invalid_proto_rawDescOnce sync.Once
	file_testproto_invalid_proto_rawDescData []byte
)

func file_testproto_invalid_proto_rawDescGZIP() []byte {
	file_testproto_invalid_proto_rawDescOnce.Do(func() {
		file_testproto_invalid_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_testproto_invalid_proto_rawDesc), len(file_testproto_invalid_proto_rawDesc)))
	})
	return file_testproto_invalid_proto_rawDescData
}

var file_testproto_invalid_proto_msgTypes = make([]protoimpl.MessageInfo, 16)
var file_testproto_invalid_proto_goTypes = []any{
	(*InvalidMethodRequest)(nil),                   // 0: testproto.InvalidMethodRequest
	(*InvalidMethodRequestWithRepo)(nil),           // 1: testproto.InvalidMethodRequestWithRepo
	(*InvalidTargetType)(nil),                      // 2: testproto.InvalidTargetType
	(*InvalidMethodResponse)(nil),                  // 3: testproto.InvalidMethodResponse
	(*InvalidNestedRequest)(nil),                   // 4: testproto.InvalidNestedRequest
	(*RequestWithStorage)(nil),                     // 5: testproto.RequestWithStorage
	(*RequestWithStorageAndRepo)(nil),              // 6: testproto.RequestWithStorageAndRepo
	(*RequestWithNestedStorageAndRepo)(nil),        // 7: testproto.RequestWithNestedStorageAndRepo
	(*RequestWithMultipleNestedStorage)(nil),       // 8: testproto.RequestWithMultipleNestedStorage
	(*RequestWithInnerNestedStorage)(nil),          // 9: testproto.RequestWithInnerNestedStorage
	(*RequestWithWrongTypeRepository)(nil),         // 10: testproto.RequestWithWrongTypeRepository
	(*RequestWithNestedRepoNotFlagged)(nil),        // 11: testproto.RequestWithNestedRepoNotFlagged
	(*RequestWithPartition)(nil),                   // 12: testproto.RequestWithPartition
	(*RequestWithInnerNestedStorage_Header)(nil),   // 13: testproto.RequestWithInnerNestedStorage.Header
	(*RequestWithWrongTypeRepository_Header)(nil),  // 14: testproto.RequestWithWrongTypeRepository.Header
	(*RequestWithNestedRepoNotFlagged_Header)(nil), // 15: testproto.RequestWithNestedRepoNotFlagged.Header
	(*gitalypb.Repository)(nil),                    // 16: gitaly.Repository
}
var file_testproto_invalid_proto_depIdxs = []int32{
	16, // 0: testproto.InvalidMethodRequestWithRepo.destination:type_name -> gitaly.Repository
	2,  // 1: testproto.InvalidNestedRequest.inner_message:type_name -> testproto.InvalidTargetType
	16, // 2: testproto.RequestWithStorage.destination:type_name -> gitaly.Repository
	16, // 3: testproto.RequestWithStorageAndRepo.destination:type_name -> gitaly.Repository
	6,  // 4: testproto.RequestWithNestedStorageAndRepo.inner_message:type_name -> testproto.RequestWithStorageAndRepo
	5,  // 5: testproto.RequestWithMultipleNestedStorage.inner_message:type_name -> testproto.RequestWithStorage
	13, // 6: testproto.RequestWithInnerNestedStorage.header:type_name -> testproto.RequestWithInnerNestedStorage.Header
	14, // 7: testproto.RequestWithWrongTypeRepository.header:type_name -> testproto.RequestWithWrongTypeRepository.Header
	15, // 8: testproto.RequestWithNestedRepoNotFlagged.header:type_name -> testproto.RequestWithNestedRepoNotFlagged.Header
	3,  // 9: testproto.RequestWithWrongTypeRepository.Header.repository:type_name -> testproto.InvalidMethodResponse
	16, // 10: testproto.RequestWithNestedRepoNotFlagged.Header.repository:type_name -> gitaly.Repository
	0,  // 11: testproto.InterceptedWithOperationType.InvalidMethod:input_type -> testproto.InvalidMethodRequest
	0,  // 12: testproto.InvalidService.InvalidMethod0:input_type -> testproto.InvalidMethodRequest
	0,  // 13: testproto.InvalidService.InvalidMethod1:input_type -> testproto.InvalidMethodRequest
	0,  // 14: testproto.InvalidService.InvalidMethod2:input_type -> testproto.InvalidMethodRequest
	0,  // 15: testproto.InvalidService.InvalidMethod4:input_type -> testproto.InvalidMethodRequest
	10, // 16: testproto.InvalidService.InvalidMethod5:input_type -> testproto.RequestWithWrongTypeRepository
	11, // 17: testproto.InvalidService.InvalidMethod6:input_type -> testproto.RequestWithNestedRepoNotFlagged
	2,  // 18: testproto.InvalidService.InvalidMethod7:input_type -> testproto.InvalidTargetType
	4,  // 19: testproto.InvalidService.InvalidMethod8:input_type -> testproto.InvalidNestedRequest
	1,  // 20: testproto.InvalidService.InvalidMethod9:input_type -> testproto.InvalidMethodRequestWithRepo
	6,  // 21: testproto.InvalidService.InvalidMethod10:input_type -> testproto.RequestWithStorageAndRepo
	7,  // 22: testproto.InvalidService.InvalidMethod11:input_type -> testproto.RequestWithNestedStorageAndRepo
	2,  // 23: testproto.InvalidService.InvalidMethod13:input_type -> testproto.InvalidTargetType
	8,  // 24: testproto.InvalidService.InvalidMethod14:input_type -> testproto.RequestWithMultipleNestedStorage
	0,  // 25: testproto.InvalidService.InvalidMethod15:input_type -> testproto.InvalidMethodRequest
	1,  // 26: testproto.InvalidService.InvalidMethod16:input_type -> testproto.InvalidMethodRequestWithRepo
	5,  // 27: testproto.InvalidService.InvalidMethod17:input_type -> testproto.RequestWithStorage
	12, // 28: testproto.InvalidService.InvalidMethod18:input_type -> testproto.RequestWithPartition
	12, // 29: testproto.InvalidService.InvalidMethod19:input_type -> testproto.RequestWithPartition
	0,  // 30: testproto.InvalidService.MaintenanceWithMissingRepository:input_type -> testproto.InvalidMethodRequest
	11, // 31: testproto.InvalidService.MaintenanceWithUnflaggedRepository:input_type -> testproto.RequestWithNestedRepoNotFlagged
	10, // 32: testproto.InvalidService.MaintenanceWithWrongNestedRepositoryType:input_type -> testproto.RequestWithWrongTypeRepository
	2,  // 33: testproto.InvalidService.MaintenanceWithInvalidTargetType:input_type -> testproto.InvalidTargetType
	4,  // 34: testproto.InvalidService.MaintenanceWithInvalidNestedRequest:input_type -> testproto.InvalidNestedRequest
	6,  // 35: testproto.InvalidService.MaintenanceWithStorageAndRepository:input_type -> testproto.RequestWithStorageAndRepo
	7,  // 36: testproto.InvalidService.MaintenanceWithNestedStorageAndRepository:input_type -> testproto.RequestWithNestedStorageAndRepo
	1,  // 37: testproto.InvalidService.MaintenanceWithStorageScope:input_type -> testproto.InvalidMethodRequestWithRepo
	3,  // 38: testproto.InterceptedWithOperationType.InvalidMethod:output_type -> testproto.InvalidMethodResponse
	3,  // 39: testproto.InvalidService.InvalidMethod0:output_type -> testproto.InvalidMethodResponse
	3,  // 40: testproto.InvalidService.InvalidMethod1:output_type -> testproto.InvalidMethodResponse
	3,  // 41: testproto.InvalidService.InvalidMethod2:output_type -> testproto.InvalidMethodResponse
	3,  // 42: testproto.InvalidService.InvalidMethod4:output_type -> testproto.InvalidMethodResponse
	3,  // 43: testproto.InvalidService.InvalidMethod5:output_type -> testproto.InvalidMethodResponse
	3,  // 44: testproto.InvalidService.InvalidMethod6:output_type -> testproto.InvalidMethodResponse
	3,  // 45: testproto.InvalidService.InvalidMethod7:output_type -> testproto.InvalidMethodResponse
	3,  // 46: testproto.InvalidService.InvalidMethod8:output_type -> testproto.InvalidMethodResponse
	3,  // 47: testproto.InvalidService.InvalidMethod9:output_type -> testproto.InvalidMethodResponse
	3,  // 48: testproto.InvalidService.InvalidMethod10:output_type -> testproto.InvalidMethodResponse
	3,  // 49: testproto.InvalidService.InvalidMethod11:output_type -> testproto.InvalidMethodResponse
	3,  // 50: testproto.InvalidService.InvalidMethod13:output_type -> testproto.InvalidMethodResponse
	3,  // 51: testproto.InvalidService.InvalidMethod14:output_type -> testproto.InvalidMethodResponse
	3,  // 52: testproto.InvalidService.InvalidMethod15:output_type -> testproto.InvalidMethodResponse
	3,  // 53: testproto.InvalidService.InvalidMethod16:output_type -> testproto.InvalidMethodResponse
	3,  // 54: testproto.InvalidService.InvalidMethod17:output_type -> testproto.InvalidMethodResponse
	3,  // 55: testproto.InvalidService.InvalidMethod18:output_type -> testproto.InvalidMethodResponse
	3,  // 56: testproto.InvalidService.InvalidMethod19:output_type -> testproto.InvalidMethodResponse
	3,  // 57: testproto.InvalidService.MaintenanceWithMissingRepository:output_type -> testproto.InvalidMethodResponse
	3,  // 58: testproto.InvalidService.MaintenanceWithUnflaggedRepository:output_type -> testproto.InvalidMethodResponse
	3,  // 59: testproto.InvalidService.MaintenanceWithWrongNestedRepositoryType:output_type -> testproto.InvalidMethodResponse
	3,  // 60: testproto.InvalidService.MaintenanceWithInvalidTargetType:output_type -> testproto.InvalidMethodResponse
	3,  // 61: testproto.InvalidService.MaintenanceWithInvalidNestedRequest:output_type -> testproto.InvalidMethodResponse
	3,  // 62: testproto.InvalidService.MaintenanceWithStorageAndRepository:output_type -> testproto.InvalidMethodResponse
	3,  // 63: testproto.InvalidService.MaintenanceWithNestedStorageAndRepository:output_type -> testproto.InvalidMethodResponse
	3,  // 64: testproto.InvalidService.MaintenanceWithStorageScope:output_type -> testproto.InvalidMethodResponse
	38, // [38:65] is the sub-list for method output_type
	11, // [11:38] is the sub-list for method input_type
	11, // [11:11] is the sub-list for extension type_name
	11, // [11:11] is the sub-list for extension extendee
	0,  // [0:11] is the sub-list for field type_name
}

func init() { file_testproto_invalid_proto_init() }
func file_testproto_invalid_proto_init() {
	if File_testproto_invalid_proto != nil {
		return
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_testproto_invalid_proto_rawDesc), len(file_testproto_invalid_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   16,
			NumExtensions: 0,
			NumServices:   2,
		},
		GoTypes:           file_testproto_invalid_proto_goTypes,
		DependencyIndexes: file_testproto_invalid_proto_depIdxs,
		MessageInfos:      file_testproto_invalid_proto_msgTypes,
	}.Build()
	File_testproto_invalid_proto = out.File
	file_testproto_invalid_proto_goTypes = nil
	file_testproto_invalid_proto_depIdxs = nil
}
