// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v6.30.2
// source: partition.proto

package gitalypb

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

// BackupPartitionRequest is a request for the BackupPartition RPC.
type BackupPartitionRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// storage_name is the name of the storage containing the partition to be backed up.
	StorageName string `protobuf:"bytes,1,opt,name=storage_name,json=storageName,proto3" json:"storage_name,omitempty"`
	// partition_id is the identifier of the specific partition within the storage to
	// be backed up, which may contain one or more repositories and associated data.
	PartitionId   string `protobuf:"bytes,2,opt,name=partition_id,json=partitionId,proto3" json:"partition_id,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *BackupPartitionRequest) Reset() {
	*x = BackupPartitionRequest{}
	mi := &file_partition_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *BackupPartitionRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*BackupPartitionRequest) ProtoMessage() {}

func (x *BackupPartitionRequest) ProtoReflect() protoreflect.Message {
	mi := &file_partition_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use BackupPartitionRequest.ProtoReflect.Descriptor instead.
func (*BackupPartitionRequest) Descriptor() ([]byte, []int) {
	return file_partition_proto_rawDescGZIP(), []int{0}
}

func (x *BackupPartitionRequest) GetStorageName() string {
	if x != nil {
		return x.StorageName
	}
	return ""
}

func (x *BackupPartitionRequest) GetPartitionId() string {
	if x != nil {
		return x.PartitionId
	}
	return ""
}

// BackupPartitionResponse is a response for the BackupPartition RPC.
type BackupPartitionResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *BackupPartitionResponse) Reset() {
	*x = BackupPartitionResponse{}
	mi := &file_partition_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *BackupPartitionResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*BackupPartitionResponse) ProtoMessage() {}

func (x *BackupPartitionResponse) ProtoReflect() protoreflect.Message {
	mi := &file_partition_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use BackupPartitionResponse.ProtoReflect.Descriptor instead.
func (*BackupPartitionResponse) Descriptor() ([]byte, []int) {
	return file_partition_proto_rawDescGZIP(), []int{1}
}

// ListPartitionsRequest is a request for the ListPartitions RPC.
type ListPartitionsRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// storage_name is the name of the storage in which partitions will be searched for.
	StorageName string `protobuf:"bytes,1,opt,name=storage_name,json=storageName,proto3" json:"storage_name,omitempty"`
	// pagination_params controls paging. Refer to PaginationParameter documentation for
	// further info.
	PaginationParams *PaginationParameter `protobuf:"bytes,2,opt,name=pagination_params,json=paginationParams,proto3" json:"pagination_params,omitempty"`
	unknownFields    protoimpl.UnknownFields
	sizeCache        protoimpl.SizeCache
}

func (x *ListPartitionsRequest) Reset() {
	*x = ListPartitionsRequest{}
	mi := &file_partition_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ListPartitionsRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ListPartitionsRequest) ProtoMessage() {}

func (x *ListPartitionsRequest) ProtoReflect() protoreflect.Message {
	mi := &file_partition_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ListPartitionsRequest.ProtoReflect.Descriptor instead.
func (*ListPartitionsRequest) Descriptor() ([]byte, []int) {
	return file_partition_proto_rawDescGZIP(), []int{2}
}

func (x *ListPartitionsRequest) GetStorageName() string {
	if x != nil {
		return x.StorageName
	}
	return ""
}

func (x *ListPartitionsRequest) GetPaginationParams() *PaginationParameter {
	if x != nil {
		return x.PaginationParams
	}
	return nil
}

// ListPartitionsResponse is a response for the ListPartitions RPC.
type ListPartitionsResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// partitions is the list of partitions found.
	Partitions []*Partition `protobuf:"bytes,1,rep,name=partitions,proto3" json:"partitions,omitempty"`
	// pagination_cursor contains the page token to fetch the next page of results. Refer to PaginationCursor
	// documentation for further info.
	PaginationCursor *PaginationCursor `protobuf:"bytes,2,opt,name=pagination_cursor,json=paginationCursor,proto3" json:"pagination_cursor,omitempty"`
	unknownFields    protoimpl.UnknownFields
	sizeCache        protoimpl.SizeCache
}

func (x *ListPartitionsResponse) Reset() {
	*x = ListPartitionsResponse{}
	mi := &file_partition_proto_msgTypes[3]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ListPartitionsResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ListPartitionsResponse) ProtoMessage() {}

func (x *ListPartitionsResponse) ProtoReflect() protoreflect.Message {
	mi := &file_partition_proto_msgTypes[3]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ListPartitionsResponse.ProtoReflect.Descriptor instead.
func (*ListPartitionsResponse) Descriptor() ([]byte, []int) {
	return file_partition_proto_rawDescGZIP(), []int{3}
}

func (x *ListPartitionsResponse) GetPartitions() []*Partition {
	if x != nil {
		return x.Partitions
	}
	return nil
}

func (x *ListPartitionsResponse) GetPaginationCursor() *PaginationCursor {
	if x != nil {
		return x.PaginationCursor
	}
	return nil
}

// KVPair represents single key-value pair.
type KVPair struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// key is the key of the key-value pair.
	Key []byte `protobuf:"bytes,1,opt,name=key,proto3" json:"key,omitempty"`
	// value is the value of the key-value pair.
	Value         []byte `protobuf:"bytes,2,opt,name=value,proto3" json:"value,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *KVPair) Reset() {
	*x = KVPair{}
	mi := &file_partition_proto_msgTypes[4]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *KVPair) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*KVPair) ProtoMessage() {}

func (x *KVPair) ProtoReflect() protoreflect.Message {
	mi := &file_partition_proto_msgTypes[4]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use KVPair.ProtoReflect.Descriptor instead.
func (*KVPair) Descriptor() ([]byte, []int) {
	return file_partition_proto_rawDescGZIP(), []int{4}
}

func (x *KVPair) GetKey() []byte {
	if x != nil {
		return x.Key
	}
	return nil
}

func (x *KVPair) GetValue() []byte {
	if x != nil {
		return x.Value
	}
	return nil
}

var File_partition_proto protoreflect.FileDescriptor

const file_partition_proto_rawDesc = "" +
	"\n" +
	"\x0fpartition.proto\x12\x06gitaly\x1a\n" +
	"lint.proto\x1a\fshared.proto\"j\n" +
	"\x16BackupPartitionRequest\x12'\n" +
	"\fstorage_name\x18\x01 \x01(\tB\x04\x88\xc6,\x01R\vstorageName\x12'\n" +
	"\fpartition_id\x18\x02 \x01(\tB\x04\xa8\xc6,\x01R\vpartitionId\"\x19\n" +
	"\x17BackupPartitionResponse\"\x8a\x01\n" +
	"\x15ListPartitionsRequest\x12'\n" +
	"\fstorage_name\x18\x01 \x01(\tB\x04\x88\xc6,\x01R\vstorageName\x12H\n" +
	"\x11pagination_params\x18\x02 \x01(\v2\x1b.gitaly.PaginationParameterR\x10paginationParams\"\x92\x01\n" +
	"\x16ListPartitionsResponse\x121\n" +
	"\n" +
	"partitions\x18\x01 \x03(\v2\x11.gitaly.PartitionR\n" +
	"partitions\x12E\n" +
	"\x11pagination_cursor\x18\x02 \x01(\v2\x18.gitaly.PaginationCursorR\x10paginationCursor\"0\n" +
	"\x06KVPair\x12\x10\n" +
	"\x03key\x18\x01 \x01(\fR\x03key\x12\x14\n" +
	"\x05value\x18\x02 \x01(\fR\x05value2\xcb\x01\n" +
	"\x10PartitionService\x12\\\n" +
	"\x0fBackupPartition\x12\x1e.gitaly.BackupPartitionRequest\x1a\x1f.gitaly.BackupPartitionResponse\"\b\xfa\x97(\x04\b\x02\x10\x03\x12Y\n" +
	"\x0eListPartitions\x12\x1d.gitaly.ListPartitionsRequest\x1a\x1e.gitaly.ListPartitionsResponse\"\b\xfa\x97(\x04\b\x02\x10\x02B4Z2gitlab.com/gitlab-org/gitaly/v18/proto/go/gitalypbb\x06proto3"

var (
	file_partition_proto_rawDescOnce sync.Once
	file_partition_proto_rawDescData []byte
)

func file_partition_proto_rawDescGZIP() []byte {
	file_partition_proto_rawDescOnce.Do(func() {
		file_partition_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_partition_proto_rawDesc), len(file_partition_proto_rawDesc)))
	})
	return file_partition_proto_rawDescData
}

var file_partition_proto_msgTypes = make([]protoimpl.MessageInfo, 5)
var file_partition_proto_goTypes = []any{
	(*BackupPartitionRequest)(nil),  // 0: gitaly.BackupPartitionRequest
	(*BackupPartitionResponse)(nil), // 1: gitaly.BackupPartitionResponse
	(*ListPartitionsRequest)(nil),   // 2: gitaly.ListPartitionsRequest
	(*ListPartitionsResponse)(nil),  // 3: gitaly.ListPartitionsResponse
	(*KVPair)(nil),                  // 4: gitaly.KVPair
	(*PaginationParameter)(nil),     // 5: gitaly.PaginationParameter
	(*Partition)(nil),               // 6: gitaly.Partition
	(*PaginationCursor)(nil),        // 7: gitaly.PaginationCursor
}
var file_partition_proto_depIdxs = []int32{
	5, // 0: gitaly.ListPartitionsRequest.pagination_params:type_name -> gitaly.PaginationParameter
	6, // 1: gitaly.ListPartitionsResponse.partitions:type_name -> gitaly.Partition
	7, // 2: gitaly.ListPartitionsResponse.pagination_cursor:type_name -> gitaly.PaginationCursor
	0, // 3: gitaly.PartitionService.BackupPartition:input_type -> gitaly.BackupPartitionRequest
	2, // 4: gitaly.PartitionService.ListPartitions:input_type -> gitaly.ListPartitionsRequest
	1, // 5: gitaly.PartitionService.BackupPartition:output_type -> gitaly.BackupPartitionResponse
	3, // 6: gitaly.PartitionService.ListPartitions:output_type -> gitaly.ListPartitionsResponse
	5, // [5:7] is the sub-list for method output_type
	3, // [3:5] is the sub-list for method input_type
	3, // [3:3] is the sub-list for extension type_name
	3, // [3:3] is the sub-list for extension extendee
	0, // [0:3] is the sub-list for field type_name
}

func init() { file_partition_proto_init() }
func file_partition_proto_init() {
	if File_partition_proto != nil {
		return
	}
	file_lint_proto_init()
	file_shared_proto_init()
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_partition_proto_rawDesc), len(file_partition_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   5,
			NumExtensions: 0,
			NumServices:   1,
		},
		GoTypes:           file_partition_proto_goTypes,
		DependencyIndexes: file_partition_proto_depIdxs,
		MessageInfos:      file_partition_proto_msgTypes,
	}.Build()
	File_partition_proto = out.File
	file_partition_proto_goTypes = nil
	file_partition_proto_depIdxs = nil
}
