// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v6.30.2
// source: analysis.proto

package gitalypb

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

// CheckBlobsGeneratedRequest is a request for the CheckBlobsGenerated RPC. The client may send
// multiple requests through the stream to check multiple sets of files. The first request must have
// the repository field set. Every request, including the first, must contain a blob set with at
// least a single entry.
type CheckBlobsGeneratedRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// repository is the repository that generated files are being checked for.
	Repository *Repository `protobuf:"bytes,1,opt,name=repository,proto3" json:"repository,omitempty"`
	// blobs is a set of blobs that the generated file check is being performed on.
	Blobs         []*CheckBlobsGeneratedRequest_Blob `protobuf:"bytes,2,rep,name=blobs,proto3" json:"blobs,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *CheckBlobsGeneratedRequest) Reset() {
	*x = CheckBlobsGeneratedRequest{}
	mi := &file_analysis_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *CheckBlobsGeneratedRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*CheckBlobsGeneratedRequest) ProtoMessage() {}

func (x *CheckBlobsGeneratedRequest) ProtoReflect() protoreflect.Message {
	mi := &file_analysis_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use CheckBlobsGeneratedRequest.ProtoReflect.Descriptor instead.
func (*CheckBlobsGeneratedRequest) Descriptor() ([]byte, []int) {
	return file_analysis_proto_rawDescGZIP(), []int{0}
}

func (x *CheckBlobsGeneratedRequest) GetRepository() *Repository {
	if x != nil {
		return x.Repository
	}
	return nil
}

func (x *CheckBlobsGeneratedRequest) GetBlobs() []*CheckBlobsGeneratedRequest_Blob {
	if x != nil {
		return x.Blobs
	}
	return nil
}

// CheckBlobsGeneratedResponse is a response for the CheckBlobsGenerated RPC. If the client sends
// multiple requests, the server responds with a separate response message for each request.
type CheckBlobsGeneratedResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// blobs is a set of blobs that the generated file check has been performed on.
	Blobs         []*CheckBlobsGeneratedResponse_Blob `protobuf:"bytes,1,rep,name=blobs,proto3" json:"blobs,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *CheckBlobsGeneratedResponse) Reset() {
	*x = CheckBlobsGeneratedResponse{}
	mi := &file_analysis_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *CheckBlobsGeneratedResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*CheckBlobsGeneratedResponse) ProtoMessage() {}

func (x *CheckBlobsGeneratedResponse) ProtoReflect() protoreflect.Message {
	mi := &file_analysis_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use CheckBlobsGeneratedResponse.ProtoReflect.Descriptor instead.
func (*CheckBlobsGeneratedResponse) Descriptor() ([]byte, []int) {
	return file_analysis_proto_rawDescGZIP(), []int{1}
}

func (x *CheckBlobsGeneratedResponse) GetBlobs() []*CheckBlobsGeneratedResponse_Blob {
	if x != nil {
		return x.Blobs
	}
	return nil
}

// Blob defines the required information to determine if a blob is generated.
type CheckBlobsGeneratedRequest_Blob struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// revision is a Git revision that resolves to a blob.
	Revision []byte `protobuf:"bytes,1,opt,name=revision,proto3" json:"revision,omitempty"`
	// path is the file path of the blob and is used to gain insight as to whether the blob is
	// generated.
	Path          []byte `protobuf:"bytes,2,opt,name=path,proto3" json:"path,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *CheckBlobsGeneratedRequest_Blob) Reset() {
	*x = CheckBlobsGeneratedRequest_Blob{}
	mi := &file_analysis_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *CheckBlobsGeneratedRequest_Blob) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*CheckBlobsGeneratedRequest_Blob) ProtoMessage() {}

func (x *CheckBlobsGeneratedRequest_Blob) ProtoReflect() protoreflect.Message {
	mi := &file_analysis_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use CheckBlobsGeneratedRequest_Blob.ProtoReflect.Descriptor instead.
func (*CheckBlobsGeneratedRequest_Blob) Descriptor() ([]byte, []int) {
	return file_analysis_proto_rawDescGZIP(), []int{0, 0}
}

func (x *CheckBlobsGeneratedRequest_Blob) GetRevision() []byte {
	if x != nil {
		return x.Revision
	}
	return nil
}

func (x *CheckBlobsGeneratedRequest_Blob) GetPath() []byte {
	if x != nil {
		return x.Path
	}
	return nil
}

// Blob defines the status of the generated blob check for a revision.
type CheckBlobsGeneratedResponse_Blob struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// revision is the Git revision of the checked blob provided in the request.
	Revision []byte `protobuf:"bytes,1,opt,name=revision,proto3" json:"revision,omitempty"`
	// generated is the result of the file generation check for a particular blob.
	Generated     bool `protobuf:"varint,2,opt,name=generated,proto3" json:"generated,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *CheckBlobsGeneratedResponse_Blob) Reset() {
	*x = CheckBlobsGeneratedResponse_Blob{}
	mi := &file_analysis_proto_msgTypes[3]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *CheckBlobsGeneratedResponse_Blob) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*CheckBlobsGeneratedResponse_Blob) ProtoMessage() {}

func (x *CheckBlobsGeneratedResponse_Blob) ProtoReflect() protoreflect.Message {
	mi := &file_analysis_proto_msgTypes[3]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use CheckBlobsGeneratedResponse_Blob.ProtoReflect.Descriptor instead.
func (*CheckBlobsGeneratedResponse_Blob) Descriptor() ([]byte, []int) {
	return file_analysis_proto_rawDescGZIP(), []int{1, 0}
}

func (x *CheckBlobsGeneratedResponse_Blob) GetRevision() []byte {
	if x != nil {
		return x.Revision
	}
	return nil
}

func (x *CheckBlobsGeneratedResponse_Blob) GetGenerated() bool {
	if x != nil {
		return x.Generated
	}
	return false
}

var File_analysis_proto protoreflect.FileDescriptor

const file_analysis_proto_rawDesc = "" +
	"\n" +
	"\x0eanalysis.proto\x12\x06gitaly\x1a\n" +
	"lint.proto\x1a\fshared.proto\"\xcd\x01\n" +
	"\x1aCheckBlobsGeneratedRequest\x128\n" +
	"\n" +
	"repository\x18\x01 \x01(\v2\x12.gitaly.RepositoryB\x04\x98\xc6,\x01R\n" +
	"repository\x12=\n" +
	"\x05blobs\x18\x02 \x03(\v2'.gitaly.CheckBlobsGeneratedRequest.BlobR\x05blobs\x1a6\n" +
	"\x04Blob\x12\x1a\n" +
	"\brevision\x18\x01 \x01(\fR\brevision\x12\x12\n" +
	"\x04path\x18\x02 \x01(\fR\x04path\"\x9f\x01\n" +
	"\x1bCheckBlobsGeneratedResponse\x12>\n" +
	"\x05blobs\x18\x01 \x03(\v2(.gitaly.CheckBlobsGeneratedResponse.BlobR\x05blobs\x1a@\n" +
	"\x04Blob\x12\x1a\n" +
	"\brevision\x18\x01 \x01(\fR\brevision\x12\x1c\n" +
	"\tgenerated\x18\x02 \x01(\bR\tgenerated2}\n" +
	"\x0fAnalysisService\x12j\n" +
	"\x13CheckBlobsGenerated\x12\".gitaly.CheckBlobsGeneratedRequest\x1a#.gitaly.CheckBlobsGeneratedResponse\"\x06\xfa\x97(\x02\b\x02(\x010\x01B4Z2gitlab.com/gitlab-org/gitaly/v18/proto/go/gitalypbb\x06proto3"

var (
	file_analysis_proto_rawDescOnce sync.Once
	file_analysis_proto_rawDescData []byte
)

func file_analysis_proto_rawDescGZIP() []byte {
	file_analysis_proto_rawDescOnce.Do(func() {
		file_analysis_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_analysis_proto_rawDesc), len(file_analysis_proto_rawDesc)))
	})
	return file_analysis_proto_rawDescData
}

var file_analysis_proto_msgTypes = make([]protoimpl.MessageInfo, 4)
var file_analysis_proto_goTypes = []any{
	(*CheckBlobsGeneratedRequest)(nil),       // 0: gitaly.CheckBlobsGeneratedRequest
	(*CheckBlobsGeneratedResponse)(nil),      // 1: gitaly.CheckBlobsGeneratedResponse
	(*CheckBlobsGeneratedRequest_Blob)(nil),  // 2: gitaly.CheckBlobsGeneratedRequest.Blob
	(*CheckBlobsGeneratedResponse_Blob)(nil), // 3: gitaly.CheckBlobsGeneratedResponse.Blob
	(*Repository)(nil),                       // 4: gitaly.Repository
}
var file_analysis_proto_depIdxs = []int32{
	4, // 0: gitaly.CheckBlobsGeneratedRequest.repository:type_name -> gitaly.Repository
	2, // 1: gitaly.CheckBlobsGeneratedRequest.blobs:type_name -> gitaly.CheckBlobsGeneratedRequest.Blob
	3, // 2: gitaly.CheckBlobsGeneratedResponse.blobs:type_name -> gitaly.CheckBlobsGeneratedResponse.Blob
	0, // 3: gitaly.AnalysisService.CheckBlobsGenerated:input_type -> gitaly.CheckBlobsGeneratedRequest
	1, // 4: gitaly.AnalysisService.CheckBlobsGenerated:output_type -> gitaly.CheckBlobsGeneratedResponse
	4, // [4:5] is the sub-list for method output_type
	3, // [3:4] is the sub-list for method input_type
	3, // [3:3] is the sub-list for extension type_name
	3, // [3:3] is the sub-list for extension extendee
	0, // [0:3] is the sub-list for field type_name
}

func init() { file_analysis_proto_init() }
func file_analysis_proto_init() {
	if File_analysis_proto != nil {
		return
	}
	file_lint_proto_init()
	file_shared_proto_init()
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_analysis_proto_rawDesc), len(file_analysis_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   4,
			NumExtensions: 0,
			NumServices:   1,
		},
		GoTypes:           file_analysis_proto_goTypes,
		DependencyIndexes: file_analysis_proto_depIdxs,
		MessageInfos:      file_analysis_proto_msgTypes,
	}.Build()
	File_analysis_proto = out.File
	file_analysis_proto_goTypes = nil
	file_analysis_proto_depIdxs = nil
}
