// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v6.30.2
// source: packfile.proto

package gitalypb

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

// PackfileNegotiationStatistics represents the git packfile negotiation.
type PackfileNegotiationStatistics struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// payload_size is the total size of all pktlines' payload in bytes. This excludes the pktline prefix.
	PayloadSize int64 `protobuf:"varint,1,opt,name=payload_size,json=payloadSize,proto3" json:"payload_size,omitempty"`
	// packets is the total number of packets.
	Packets int64 `protobuf:"varint,2,opt,name=packets,proto3" json:"packets,omitempty"`
	// caps is the capabilities announced by the client.
	Caps []string `protobuf:"bytes,3,rep,name=caps,proto3" json:"caps,omitempty"`
	// wants is the number of objects the client announced it wants.
	Wants int64 `protobuf:"varint,4,opt,name=wants,proto3" json:"wants,omitempty"`
	// haves is the number of objects the client announced it has.
	Haves int64 `protobuf:"varint,5,opt,name=haves,proto3" json:"haves,omitempty"`
	// shallows is the number of shallow boundaries announced by the client.
	Shallows int64 `protobuf:"varint,6,opt,name=shallows,proto3" json:"shallows,omitempty"`
	// deepen is one of "deepen <depth>", "deepen-since <timestamp>", "deepen-not <ref>".
	// [deepen <depth>|deepen-since <timestamp>|deepen-not <ref>]
	Deepen string `protobuf:"bytes,7,opt,name=deepen,proto3" json:"deepen,omitempty"`
	// filter is specified by the client.
	Filter        string `protobuf:"bytes,8,opt,name=filter,proto3" json:"filter,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *PackfileNegotiationStatistics) Reset() {
	*x = PackfileNegotiationStatistics{}
	mi := &file_packfile_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *PackfileNegotiationStatistics) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*PackfileNegotiationStatistics) ProtoMessage() {}

func (x *PackfileNegotiationStatistics) ProtoReflect() protoreflect.Message {
	mi := &file_packfile_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use PackfileNegotiationStatistics.ProtoReflect.Descriptor instead.
func (*PackfileNegotiationStatistics) Descriptor() ([]byte, []int) {
	return file_packfile_proto_rawDescGZIP(), []int{0}
}

func (x *PackfileNegotiationStatistics) GetPayloadSize() int64 {
	if x != nil {
		return x.PayloadSize
	}
	return 0
}

func (x *PackfileNegotiationStatistics) GetPackets() int64 {
	if x != nil {
		return x.Packets
	}
	return 0
}

func (x *PackfileNegotiationStatistics) GetCaps() []string {
	if x != nil {
		return x.Caps
	}
	return nil
}

func (x *PackfileNegotiationStatistics) GetWants() int64 {
	if x != nil {
		return x.Wants
	}
	return 0
}

func (x *PackfileNegotiationStatistics) GetHaves() int64 {
	if x != nil {
		return x.Haves
	}
	return 0
}

func (x *PackfileNegotiationStatistics) GetShallows() int64 {
	if x != nil {
		return x.Shallows
	}
	return 0
}

func (x *PackfileNegotiationStatistics) GetDeepen() string {
	if x != nil {
		return x.Deepen
	}
	return ""
}

func (x *PackfileNegotiationStatistics) GetFilter() string {
	if x != nil {
		return x.Filter
	}
	return ""
}

var File_packfile_proto protoreflect.FileDescriptor

const file_packfile_proto_rawDesc = "" +
	"\n" +
	"\x0epackfile.proto\x12\x06gitaly\"\xe8\x01\n" +
	"\x1dPackfileNegotiationStatistics\x12!\n" +
	"\fpayload_size\x18\x01 \x01(\x03R\vpayloadSize\x12\x18\n" +
	"\apackets\x18\x02 \x01(\x03R\apackets\x12\x12\n" +
	"\x04caps\x18\x03 \x03(\tR\x04caps\x12\x14\n" +
	"\x05wants\x18\x04 \x01(\x03R\x05wants\x12\x14\n" +
	"\x05haves\x18\x05 \x01(\x03R\x05haves\x12\x1a\n" +
	"\bshallows\x18\x06 \x01(\x03R\bshallows\x12\x16\n" +
	"\x06deepen\x18\a \x01(\tR\x06deepen\x12\x16\n" +
	"\x06filter\x18\b \x01(\tR\x06filterB4Z2gitlab.com/gitlab-org/gitaly/v18/proto/go/gitalypbb\x06proto3"

var (
	file_packfile_proto_rawDescOnce sync.Once
	file_packfile_proto_rawDescData []byte
)

func file_packfile_proto_rawDescGZIP() []byte {
	file_packfile_proto_rawDescOnce.Do(func() {
		file_packfile_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_packfile_proto_rawDesc), len(file_packfile_proto_rawDesc)))
	})
	return file_packfile_proto_rawDescData
}

var file_packfile_proto_msgTypes = make([]protoimpl.MessageInfo, 1)
var file_packfile_proto_goTypes = []any{
	(*PackfileNegotiationStatistics)(nil), // 0: gitaly.PackfileNegotiationStatistics
}
var file_packfile_proto_depIdxs = []int32{
	0, // [0:0] is the sub-list for method output_type
	0, // [0:0] is the sub-list for method input_type
	0, // [0:0] is the sub-list for extension type_name
	0, // [0:0] is the sub-list for extension extendee
	0, // [0:0] is the sub-list for field type_name
}

func init() { file_packfile_proto_init() }
func file_packfile_proto_init() {
	if File_packfile_proto != nil {
		return
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_packfile_proto_rawDesc), len(file_packfile_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   1,
			NumExtensions: 0,
			NumServices:   0,
		},
		GoTypes:           file_packfile_proto_goTypes,
		DependencyIndexes: file_packfile_proto_depIdxs,
		MessageInfos:      file_packfile_proto_msgTypes,
	}.Build()
	File_packfile_proto = out.File
	file_packfile_proto_goTypes = nil
	file_packfile_proto_depIdxs = nil
}
