// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v6.30.2
// source: smarthttp.proto

package gitalypb

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

// InfoRefsRequest is a request for the InfoRefsUploadPack and InfoRefsUploadPack rpcs.
type InfoRefsRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// repository is the repository on which to operate.
	Repository *Repository `protobuf:"bytes,1,opt,name=repository,proto3" json:"repository,omitempty"`
	// git_config_options are parameters to use with git -c (key=value pairs).
	GitConfigOptions []string `protobuf:"bytes,2,rep,name=git_config_options,json=gitConfigOptions,proto3" json:"git_config_options,omitempty"`
	// git_protocol is the git protocol version.
	GitProtocol   string `protobuf:"bytes,3,opt,name=git_protocol,json=gitProtocol,proto3" json:"git_protocol,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *InfoRefsRequest) Reset() {
	*x = InfoRefsRequest{}
	mi := &file_smarthttp_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *InfoRefsRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*InfoRefsRequest) ProtoMessage() {}

func (x *InfoRefsRequest) ProtoReflect() protoreflect.Message {
	mi := &file_smarthttp_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use InfoRefsRequest.ProtoReflect.Descriptor instead.
func (*InfoRefsRequest) Descriptor() ([]byte, []int) {
	return file_smarthttp_proto_rawDescGZIP(), []int{0}
}

func (x *InfoRefsRequest) GetRepository() *Repository {
	if x != nil {
		return x.Repository
	}
	return nil
}

func (x *InfoRefsRequest) GetGitConfigOptions() []string {
	if x != nil {
		return x.GitConfigOptions
	}
	return nil
}

func (x *InfoRefsRequest) GetGitProtocol() string {
	if x != nil {
		return x.GitProtocol
	}
	return ""
}

// InfoRefsResponse is the response of InfoRefsUploadPack and InfoRefsUploadPack rpcs.
// It is used to provide the client with the servers advertised refs.
type InfoRefsResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// data is the raw data copied from the stdout of git-upload-pack(1) or
	// git-receive-pack(1) when used with the `--advertise-refs` flag.
	Data          []byte `protobuf:"bytes,1,opt,name=data,proto3" json:"data,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *InfoRefsResponse) Reset() {
	*x = InfoRefsResponse{}
	mi := &file_smarthttp_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *InfoRefsResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*InfoRefsResponse) ProtoMessage() {}

func (x *InfoRefsResponse) ProtoReflect() protoreflect.Message {
	mi := &file_smarthttp_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use InfoRefsResponse.ProtoReflect.Descriptor instead.
func (*InfoRefsResponse) Descriptor() ([]byte, []int) {
	return file_smarthttp_proto_rawDescGZIP(), []int{1}
}

func (x *InfoRefsResponse) GetData() []byte {
	if x != nil {
		return x.Data
	}
	return nil
}

// PostUploadPackWithSidechannelRequest is the request for the PostUploadPackWithSidechannel rpc.
type PostUploadPackWithSidechannelRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// repository is the repository on which to operate.
	Repository *Repository `protobuf:"bytes,1,opt,name=repository,proto3" json:"repository,omitempty"`
	// git_config_options are parameters to use with git -c (key=value pairs).
	GitConfigOptions []string `protobuf:"bytes,2,rep,name=git_config_options,json=gitConfigOptions,proto3" json:"git_config_options,omitempty"`
	// git_protocol is the git protocol version.
	GitProtocol   string `protobuf:"bytes,3,opt,name=git_protocol,json=gitProtocol,proto3" json:"git_protocol,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *PostUploadPackWithSidechannelRequest) Reset() {
	*x = PostUploadPackWithSidechannelRequest{}
	mi := &file_smarthttp_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *PostUploadPackWithSidechannelRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*PostUploadPackWithSidechannelRequest) ProtoMessage() {}

func (x *PostUploadPackWithSidechannelRequest) ProtoReflect() protoreflect.Message {
	mi := &file_smarthttp_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use PostUploadPackWithSidechannelRequest.ProtoReflect.Descriptor instead.
func (*PostUploadPackWithSidechannelRequest) Descriptor() ([]byte, []int) {
	return file_smarthttp_proto_rawDescGZIP(), []int{2}
}

func (x *PostUploadPackWithSidechannelRequest) GetRepository() *Repository {
	if x != nil {
		return x.Repository
	}
	return nil
}

func (x *PostUploadPackWithSidechannelRequest) GetGitConfigOptions() []string {
	if x != nil {
		return x.GitConfigOptions
	}
	return nil
}

func (x *PostUploadPackWithSidechannelRequest) GetGitProtocol() string {
	if x != nil {
		return x.GitProtocol
	}
	return ""
}

// PostUploadPackWithSidechannelResponse is the response for the PostUploadPackWithSidechannel rpc.
// This is an empty response since the raw data is transferred to the client via the sidechannel
// exclusively.
type PostUploadPackWithSidechannelResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// packfile_negotiation_statistics is the packfile negotiation statistics.
	PackfileNegotiationStatistics *PackfileNegotiationStatistics `protobuf:"bytes,1,opt,name=packfile_negotiation_statistics,json=packfileNegotiationStatistics,proto3" json:"packfile_negotiation_statistics,omitempty"`
	unknownFields                 protoimpl.UnknownFields
	sizeCache                     protoimpl.SizeCache
}

func (x *PostUploadPackWithSidechannelResponse) Reset() {
	*x = PostUploadPackWithSidechannelResponse{}
	mi := &file_smarthttp_proto_msgTypes[3]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *PostUploadPackWithSidechannelResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*PostUploadPackWithSidechannelResponse) ProtoMessage() {}

func (x *PostUploadPackWithSidechannelResponse) ProtoReflect() protoreflect.Message {
	mi := &file_smarthttp_proto_msgTypes[3]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use PostUploadPackWithSidechannelResponse.ProtoReflect.Descriptor instead.
func (*PostUploadPackWithSidechannelResponse) Descriptor() ([]byte, []int) {
	return file_smarthttp_proto_rawDescGZIP(), []int{3}
}

func (x *PostUploadPackWithSidechannelResponse) GetPackfileNegotiationStatistics() *PackfileNegotiationStatistics {
	if x != nil {
		return x.PackfileNegotiationStatistics
	}
	return nil
}

// PostReceivePackRequest is the request for the PostReceivePack rpc. It is a stream used to
// transfer the raw data from the client to the servers stdin of git-receive-pack(1) process.
type PostReceivePackRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// repository is the repository on which to operate.
	// It should only be present in the first message of the stream.
	Repository *Repository `protobuf:"bytes,1,opt,name=repository,proto3" json:"repository,omitempty"`
	// data is the raw data to be copied to stdin of 'git receive-pack'.
	Data []byte `protobuf:"bytes,2,opt,name=data,proto3" json:"data,omitempty"`
	// gl_id is the GitLab ID of the user. This is used by Git {pre,post}-receive hooks.
	// It should only be present in the first message of the stream.
	GlId string `protobuf:"bytes,3,opt,name=gl_id,json=glId,proto3" json:"gl_id,omitempty"`
	// gl_repository refers to the GitLab repository. This is used by Git {pre,post}-receive hooks.
	// It should only be present in the first message of the stream.
	GlRepository string `protobuf:"bytes,4,opt,name=gl_repository,json=glRepository,proto3" json:"gl_repository,omitempty"`
	// gl_username is the GitLab Username of the user. This is used by Git {pre,post}-receive hooks.
	// It should only be present in the first message of the stream.
	GlUsername string `protobuf:"bytes,5,opt,name=gl_username,json=glUsername,proto3" json:"gl_username,omitempty"`
	// git_protocol is the git protocol version.
	GitProtocol string `protobuf:"bytes,6,opt,name=git_protocol,json=gitProtocol,proto3" json:"git_protocol,omitempty"`
	// git_config_options are parameters to use with git -c (key=value pairs).
	GitConfigOptions []string `protobuf:"bytes,7,rep,name=git_config_options,json=gitConfigOptions,proto3" json:"git_config_options,omitempty"`
	unknownFields    protoimpl.UnknownFields
	sizeCache        protoimpl.SizeCache
}

func (x *PostReceivePackRequest) Reset() {
	*x = PostReceivePackRequest{}
	mi := &file_smarthttp_proto_msgTypes[4]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *PostReceivePackRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*PostReceivePackRequest) ProtoMessage() {}

func (x *PostReceivePackRequest) ProtoReflect() protoreflect.Message {
	mi := &file_smarthttp_proto_msgTypes[4]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use PostReceivePackRequest.ProtoReflect.Descriptor instead.
func (*PostReceivePackRequest) Descriptor() ([]byte, []int) {
	return file_smarthttp_proto_rawDescGZIP(), []int{4}
}

func (x *PostReceivePackRequest) GetRepository() *Repository {
	if x != nil {
		return x.Repository
	}
	return nil
}

func (x *PostReceivePackRequest) GetData() []byte {
	if x != nil {
		return x.Data
	}
	return nil
}

func (x *PostReceivePackRequest) GetGlId() string {
	if x != nil {
		return x.GlId
	}
	return ""
}

func (x *PostReceivePackRequest) GetGlRepository() string {
	if x != nil {
		return x.GlRepository
	}
	return ""
}

func (x *PostReceivePackRequest) GetGlUsername() string {
	if x != nil {
		return x.GlUsername
	}
	return ""
}

func (x *PostReceivePackRequest) GetGitProtocol() string {
	if x != nil {
		return x.GitProtocol
	}
	return ""
}

func (x *PostReceivePackRequest) GetGitConfigOptions() []string {
	if x != nil {
		return x.GitConfigOptions
	}
	return nil
}

// PostReceivePackResponse is the response for the PostReceivePack rpc. It is a stream used to
// transfer the raw data from the stdout of git-receive-pack(1) from the server to the client.
type PostReceivePackResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// data is the raw data from the stdout of 'git receive-pack'.
	Data          []byte `protobuf:"bytes,1,opt,name=data,proto3" json:"data,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *PostReceivePackResponse) Reset() {
	*x = PostReceivePackResponse{}
	mi := &file_smarthttp_proto_msgTypes[5]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *PostReceivePackResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*PostReceivePackResponse) ProtoMessage() {}

func (x *PostReceivePackResponse) ProtoReflect() protoreflect.Message {
	mi := &file_smarthttp_proto_msgTypes[5]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use PostReceivePackResponse.ProtoReflect.Descriptor instead.
func (*PostReceivePackResponse) Descriptor() ([]byte, []int) {
	return file_smarthttp_proto_rawDescGZIP(), []int{5}
}

func (x *PostReceivePackResponse) GetData() []byte {
	if x != nil {
		return x.Data
	}
	return nil
}

var File_smarthttp_proto protoreflect.FileDescriptor

const file_smarthttp_proto_rawDesc = "" +
	"\n" +
	"\x0fsmarthttp.proto\x12\x06gitaly\x1a\n" +
	"lint.proto\x1a\x0epackfile.proto\x1a\fshared.proto\"\x9c\x01\n" +
	"\x0fInfoRefsRequest\x128\n" +
	"\n" +
	"repository\x18\x01 \x01(\v2\x12.gitaly.RepositoryB\x04\x98\xc6,\x01R\n" +
	"repository\x12,\n" +
	"\x12git_config_options\x18\x02 \x03(\tR\x10gitConfigOptions\x12!\n" +
	"\fgit_protocol\x18\x03 \x01(\tR\vgitProtocol\"&\n" +
	"\x10InfoRefsResponse\x12\x12\n" +
	"\x04data\x18\x01 \x01(\fR\x04data\"\xb1\x01\n" +
	"$PostUploadPackWithSidechannelRequest\x128\n" +
	"\n" +
	"repository\x18\x01 \x01(\v2\x12.gitaly.RepositoryB\x04\x98\xc6,\x01R\n" +
	"repository\x12,\n" +
	"\x12git_config_options\x18\x02 \x03(\tR\x10gitConfigOptions\x12!\n" +
	"\fgit_protocol\x18\x03 \x01(\tR\vgitProtocol\"\x96\x01\n" +
	"%PostUploadPackWithSidechannelResponse\x12m\n" +
	"\x1fpackfile_negotiation_statistics\x18\x01 \x01(\v2%.gitaly.PackfileNegotiationStatisticsR\x1dpackfileNegotiationStatistics\"\x92\x02\n" +
	"\x16PostReceivePackRequest\x128\n" +
	"\n" +
	"repository\x18\x01 \x01(\v2\x12.gitaly.RepositoryB\x04\x98\xc6,\x01R\n" +
	"repository\x12\x12\n" +
	"\x04data\x18\x02 \x01(\fR\x04data\x12\x13\n" +
	"\x05gl_id\x18\x03 \x01(\tR\x04glId\x12#\n" +
	"\rgl_repository\x18\x04 \x01(\tR\fglRepository\x12\x1f\n" +
	"\vgl_username\x18\x05 \x01(\tR\n" +
	"glUsername\x12!\n" +
	"\fgit_protocol\x18\x06 \x01(\tR\vgitProtocol\x12,\n" +
	"\x12git_config_options\x18\a \x03(\tR\x10gitConfigOptions\"-\n" +
	"\x17PostReceivePackResponse\x12\x12\n" +
	"\x04data\x18\x01 \x01(\fR\x04data2\xa0\x03\n" +
	"\x10SmartHTTPService\x12Q\n" +
	"\x12InfoRefsUploadPack\x12\x17.gitaly.InfoRefsRequest\x1a\x18.gitaly.InfoRefsResponse\"\x06\xfa\x97(\x02\b\x020\x01\x12R\n" +
	"\x13InfoRefsReceivePack\x12\x17.gitaly.InfoRefsRequest\x1a\x18.gitaly.InfoRefsResponse\"\x06\xfa\x97(\x02\b\x020\x01\x12\x84\x01\n" +
	"\x1dPostUploadPackWithSidechannel\x12,.gitaly.PostUploadPackWithSidechannelRequest\x1a-.gitaly.PostUploadPackWithSidechannelResponse\"\x06\xfa\x97(\x02\b\x02\x12^\n" +
	"\x0fPostReceivePack\x12\x1e.gitaly.PostReceivePackRequest\x1a\x1f.gitaly.PostReceivePackResponse\"\x06\xfa\x97(\x02\b\x01(\x010\x01B4Z2gitlab.com/gitlab-org/gitaly/v18/proto/go/gitalypbb\x06proto3"

var (
	file_smarthttp_proto_rawDescOnce sync.Once
	file_smarthttp_proto_rawDescData []byte
)

func file_smarthttp_proto_rawDescGZIP() []byte {
	file_smarthttp_proto_rawDescOnce.Do(func() {
		file_smarthttp_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_smarthttp_proto_rawDesc), len(file_smarthttp_proto_rawDesc)))
	})
	return file_smarthttp_proto_rawDescData
}

var file_smarthttp_proto_msgTypes = make([]protoimpl.MessageInfo, 6)
var file_smarthttp_proto_goTypes = []any{
	(*InfoRefsRequest)(nil),                       // 0: gitaly.InfoRefsRequest
	(*InfoRefsResponse)(nil),                      // 1: gitaly.InfoRefsResponse
	(*PostUploadPackWithSidechannelRequest)(nil),  // 2: gitaly.PostUploadPackWithSidechannelRequest
	(*PostUploadPackWithSidechannelResponse)(nil), // 3: gitaly.PostUploadPackWithSidechannelResponse
	(*PostReceivePackRequest)(nil),                // 4: gitaly.PostReceivePackRequest
	(*PostReceivePackResponse)(nil),               // 5: gitaly.PostReceivePackResponse
	(*Repository)(nil),                            // 6: gitaly.Repository
	(*PackfileNegotiationStatistics)(nil),         // 7: gitaly.PackfileNegotiationStatistics
}
var file_smarthttp_proto_depIdxs = []int32{
	6, // 0: gitaly.InfoRefsRequest.repository:type_name -> gitaly.Repository
	6, // 1: gitaly.PostUploadPackWithSidechannelRequest.repository:type_name -> gitaly.Repository
	7, // 2: gitaly.PostUploadPackWithSidechannelResponse.packfile_negotiation_statistics:type_name -> gitaly.PackfileNegotiationStatistics
	6, // 3: gitaly.PostReceivePackRequest.repository:type_name -> gitaly.Repository
	0, // 4: gitaly.SmartHTTPService.InfoRefsUploadPack:input_type -> gitaly.InfoRefsRequest
	0, // 5: gitaly.SmartHTTPService.InfoRefsReceivePack:input_type -> gitaly.InfoRefsRequest
	2, // 6: gitaly.SmartHTTPService.PostUploadPackWithSidechannel:input_type -> gitaly.PostUploadPackWithSidechannelRequest
	4, // 7: gitaly.SmartHTTPService.PostReceivePack:input_type -> gitaly.PostReceivePackRequest
	1, // 8: gitaly.SmartHTTPService.InfoRefsUploadPack:output_type -> gitaly.InfoRefsResponse
	1, // 9: gitaly.SmartHTTPService.InfoRefsReceivePack:output_type -> gitaly.InfoRefsResponse
	3, // 10: gitaly.SmartHTTPService.PostUploadPackWithSidechannel:output_type -> gitaly.PostUploadPackWithSidechannelResponse
	5, // 11: gitaly.SmartHTTPService.PostReceivePack:output_type -> gitaly.PostReceivePackResponse
	8, // [8:12] is the sub-list for method output_type
	4, // [4:8] is the sub-list for method input_type
	4, // [4:4] is the sub-list for extension type_name
	4, // [4:4] is the sub-list for extension extendee
	0, // [0:4] is the sub-list for field type_name
}

func init() { file_smarthttp_proto_init() }
func file_smarthttp_proto_init() {
	if File_smarthttp_proto != nil {
		return
	}
	file_lint_proto_init()
	file_packfile_proto_init()
	file_shared_proto_init()
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_smarthttp_proto_rawDesc), len(file_smarthttp_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   6,
			NumExtensions: 0,
			NumServices:   1,
		},
		GoTypes:           file_smarthttp_proto_goTypes,
		DependencyIndexes: file_smarthttp_proto_depIdxs,
		MessageInfos:      file_smarthttp_proto_msgTypes,
	}.Build()
	File_smarthttp_proto = out.File
	file_smarthttp_proto_goTypes = nil
	file_smarthttp_proto_depIdxs = nil
}
