// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v6.30.2
// source: ssh.proto

package gitalypb

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

// SSHUploadPackWithSidechannelRequest is a request for the SSHUploadPackWithSidechannel RPC.
type SSHUploadPackWithSidechannelRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// repository is the repository where git-upload-pack(1) is spawned.
	Repository *Repository `protobuf:"bytes,1,opt,name=repository,proto3" json:"repository,omitempty"`
	// git_config_options are parameters to use with git -c (key=value pairs).
	GitConfigOptions []string `protobuf:"bytes,2,rep,name=git_config_options,json=gitConfigOptions,proto3" json:"git_config_options,omitempty"`
	// git_protocol is the git protocol version.
	GitProtocol   string `protobuf:"bytes,3,opt,name=git_protocol,json=gitProtocol,proto3" json:"git_protocol,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *SSHUploadPackWithSidechannelRequest) Reset() {
	*x = SSHUploadPackWithSidechannelRequest{}
	mi := &file_ssh_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *SSHUploadPackWithSidechannelRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*SSHUploadPackWithSidechannelRequest) ProtoMessage() {}

func (x *SSHUploadPackWithSidechannelRequest) ProtoReflect() protoreflect.Message {
	mi := &file_ssh_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use SSHUploadPackWithSidechannelRequest.ProtoReflect.Descriptor instead.
func (*SSHUploadPackWithSidechannelRequest) Descriptor() ([]byte, []int) {
	return file_ssh_proto_rawDescGZIP(), []int{0}
}

func (x *SSHUploadPackWithSidechannelRequest) GetRepository() *Repository {
	if x != nil {
		return x.Repository
	}
	return nil
}

func (x *SSHUploadPackWithSidechannelRequest) GetGitConfigOptions() []string {
	if x != nil {
		return x.GitConfigOptions
	}
	return nil
}

func (x *SSHUploadPackWithSidechannelRequest) GetGitProtocol() string {
	if x != nil {
		return x.GitProtocol
	}
	return ""
}

// SSHUploadPackWithSidechannelResponse is a response for the SSHUploadPackWithSidechannel RPC.
type SSHUploadPackWithSidechannelResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// packfile_negotiation_statistics is the packfile negotiation statistics.
	PackfileNegotiationStatistics *PackfileNegotiationStatistics `protobuf:"bytes,1,opt,name=packfile_negotiation_statistics,json=packfileNegotiationStatistics,proto3" json:"packfile_negotiation_statistics,omitempty"`
	unknownFields                 protoimpl.UnknownFields
	sizeCache                     protoimpl.SizeCache
}

func (x *SSHUploadPackWithSidechannelResponse) Reset() {
	*x = SSHUploadPackWithSidechannelResponse{}
	mi := &file_ssh_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *SSHUploadPackWithSidechannelResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*SSHUploadPackWithSidechannelResponse) ProtoMessage() {}

func (x *SSHUploadPackWithSidechannelResponse) ProtoReflect() protoreflect.Message {
	mi := &file_ssh_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use SSHUploadPackWithSidechannelResponse.ProtoReflect.Descriptor instead.
func (*SSHUploadPackWithSidechannelResponse) Descriptor() ([]byte, []int) {
	return file_ssh_proto_rawDescGZIP(), []int{1}
}

func (x *SSHUploadPackWithSidechannelResponse) GetPackfileNegotiationStatistics() *PackfileNegotiationStatistics {
	if x != nil {
		return x.PackfileNegotiationStatistics
	}
	return nil
}

// SSHReceivePackRequest is a request for the SSHReceivePack RPC. All fields other than stdin must
// be set in the first request message. Subsequent requests in the stream must only contain the
// stdin field.
type SSHReceivePackRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// repository is the repository where git-receive-pack(1) is spawned.
	Repository *Repository `protobuf:"bytes,1,opt,name=repository,proto3" json:"repository,omitempty"`
	// stdin is a chunk of raw data to be copied to git-receive-pack(1) standard input
	Stdin []byte `protobuf:"bytes,2,opt,name=stdin,proto3" json:"stdin,omitempty"`
	// gl_id is the GitLab ID of the user. This is used by Git {pre,post}-receive hooks.
	GlId string `protobuf:"bytes,3,opt,name=gl_id,json=glId,proto3" json:"gl_id,omitempty"`
	// gl_repository refers to the GitLab repository. This is used by Git {pre,post}-receive hooks.
	GlRepository string `protobuf:"bytes,4,opt,name=gl_repository,json=glRepository,proto3" json:"gl_repository,omitempty"`
	// gl_username is the GitLab Username of the user. This is used by Git {pre,post}-receive hooks.
	GlUsername string `protobuf:"bytes,5,opt,name=gl_username,json=glUsername,proto3" json:"gl_username,omitempty"`
	// git_protocol is the git protocol version.
	GitProtocol string `protobuf:"bytes,6,opt,name=git_protocol,json=gitProtocol,proto3" json:"git_protocol,omitempty"`
	// git_config_options are parameters to use with git -c (key=value pairs).
	GitConfigOptions []string `protobuf:"bytes,7,rep,name=git_config_options,json=gitConfigOptions,proto3" json:"git_config_options,omitempty"`
	unknownFields    protoimpl.UnknownFields
	sizeCache        protoimpl.SizeCache
}

func (x *SSHReceivePackRequest) Reset() {
	*x = SSHReceivePackRequest{}
	mi := &file_ssh_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *SSHReceivePackRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*SSHReceivePackRequest) ProtoMessage() {}

func (x *SSHReceivePackRequest) ProtoReflect() protoreflect.Message {
	mi := &file_ssh_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use SSHReceivePackRequest.ProtoReflect.Descriptor instead.
func (*SSHReceivePackRequest) Descriptor() ([]byte, []int) {
	return file_ssh_proto_rawDescGZIP(), []int{2}
}

func (x *SSHReceivePackRequest) GetRepository() *Repository {
	if x != nil {
		return x.Repository
	}
	return nil
}

func (x *SSHReceivePackRequest) GetStdin() []byte {
	if x != nil {
		return x.Stdin
	}
	return nil
}

func (x *SSHReceivePackRequest) GetGlId() string {
	if x != nil {
		return x.GlId
	}
	return ""
}

func (x *SSHReceivePackRequest) GetGlRepository() string {
	if x != nil {
		return x.GlRepository
	}
	return ""
}

func (x *SSHReceivePackRequest) GetGlUsername() string {
	if x != nil {
		return x.GlUsername
	}
	return ""
}

func (x *SSHReceivePackRequest) GetGitProtocol() string {
	if x != nil {
		return x.GitProtocol
	}
	return ""
}

func (x *SSHReceivePackRequest) GetGitConfigOptions() []string {
	if x != nil {
		return x.GitConfigOptions
	}
	return nil
}

// SSHReceivePackResponse is a response for the SSHReceivePack RPC. Responses are stream back to
// clients in chunks containing the stdout and stderr from git-receive-pack(1).
type SSHReceivePackResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// stdout is a chunk of raw data from git-receive-pack(1) standard output.
	Stdout []byte `protobuf:"bytes,1,opt,name=stdout,proto3" json:"stdout,omitempty"`
	// stderr is a chunk of raw data from git-receive-pack(1) standard error.
	Stderr []byte `protobuf:"bytes,2,opt,name=stderr,proto3" json:"stderr,omitempty"`
	// exit_status is the exit status when the command has finished. This field
	// may be nil. This is intentional.
	ExitStatus    *ExitStatus `protobuf:"bytes,3,opt,name=exit_status,json=exitStatus,proto3" json:"exit_status,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *SSHReceivePackResponse) Reset() {
	*x = SSHReceivePackResponse{}
	mi := &file_ssh_proto_msgTypes[3]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *SSHReceivePackResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*SSHReceivePackResponse) ProtoMessage() {}

func (x *SSHReceivePackResponse) ProtoReflect() protoreflect.Message {
	mi := &file_ssh_proto_msgTypes[3]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use SSHReceivePackResponse.ProtoReflect.Descriptor instead.
func (*SSHReceivePackResponse) Descriptor() ([]byte, []int) {
	return file_ssh_proto_rawDescGZIP(), []int{3}
}

func (x *SSHReceivePackResponse) GetStdout() []byte {
	if x != nil {
		return x.Stdout
	}
	return nil
}

func (x *SSHReceivePackResponse) GetStderr() []byte {
	if x != nil {
		return x.Stderr
	}
	return nil
}

func (x *SSHReceivePackResponse) GetExitStatus() *ExitStatus {
	if x != nil {
		return x.ExitStatus
	}
	return nil
}

// SSHUploadArchiveRequest is a request for the SSHUploadArchive RPC. The first request of the
// stream must only contain repository. All subsequent requests must only contain stdin data.
type SSHUploadArchiveRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// repository is the repository where git-upload-archive(1) is spawned.
	Repository *Repository `protobuf:"bytes,1,opt,name=repository,proto3" json:"repository,omitempty"`
	// stdin is a chunk of raw data to be copied to git-upload-archive(1) standard input.
	Stdin         []byte `protobuf:"bytes,2,opt,name=stdin,proto3" json:"stdin,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *SSHUploadArchiveRequest) Reset() {
	*x = SSHUploadArchiveRequest{}
	mi := &file_ssh_proto_msgTypes[4]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *SSHUploadArchiveRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*SSHUploadArchiveRequest) ProtoMessage() {}

func (x *SSHUploadArchiveRequest) ProtoReflect() protoreflect.Message {
	mi := &file_ssh_proto_msgTypes[4]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use SSHUploadArchiveRequest.ProtoReflect.Descriptor instead.
func (*SSHUploadArchiveRequest) Descriptor() ([]byte, []int) {
	return file_ssh_proto_rawDescGZIP(), []int{4}
}

func (x *SSHUploadArchiveRequest) GetRepository() *Repository {
	if x != nil {
		return x.Repository
	}
	return nil
}

func (x *SSHUploadArchiveRequest) GetStdin() []byte {
	if x != nil {
		return x.Stdin
	}
	return nil
}

// SSHUploadArchiveResponse is a response for the SSHUploadArchive RPC. Responses are stream back to
// clients in chunks containing the stdout and stderr from git-upload-archive(1).
type SSHUploadArchiveResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// stdout is a chunk of raw data from git-upload-archive(1) standard output.
	Stdout []byte `protobuf:"bytes,1,opt,name=stdout,proto3" json:"stdout,omitempty"`
	// stderr is a chunk of raw data from git-upload-archive(1) standard error.
	Stderr []byte `protobuf:"bytes,2,opt,name=stderr,proto3" json:"stderr,omitempty"`
	// exit_status is the exit status when the command has finished. This field
	// may be nil. This is intentional.
	ExitStatus    *ExitStatus `protobuf:"bytes,3,opt,name=exit_status,json=exitStatus,proto3" json:"exit_status,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *SSHUploadArchiveResponse) Reset() {
	*x = SSHUploadArchiveResponse{}
	mi := &file_ssh_proto_msgTypes[5]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *SSHUploadArchiveResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*SSHUploadArchiveResponse) ProtoMessage() {}

func (x *SSHUploadArchiveResponse) ProtoReflect() protoreflect.Message {
	mi := &file_ssh_proto_msgTypes[5]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use SSHUploadArchiveResponse.ProtoReflect.Descriptor instead.
func (*SSHUploadArchiveResponse) Descriptor() ([]byte, []int) {
	return file_ssh_proto_rawDescGZIP(), []int{5}
}

func (x *SSHUploadArchiveResponse) GetStdout() []byte {
	if x != nil {
		return x.Stdout
	}
	return nil
}

func (x *SSHUploadArchiveResponse) GetStderr() []byte {
	if x != nil {
		return x.Stderr
	}
	return nil
}

func (x *SSHUploadArchiveResponse) GetExitStatus() *ExitStatus {
	if x != nil {
		return x.ExitStatus
	}
	return nil
}

var File_ssh_proto protoreflect.FileDescriptor

const file_ssh_proto_rawDesc = "" +
	"\n" +
	"\tssh.proto\x12\x06gitaly\x1a\n" +
	"lint.proto\x1a\x0epackfile.proto\x1a\fshared.proto\"\xb0\x01\n" +
	"#SSHUploadPackWithSidechannelRequest\x128\n" +
	"\n" +
	"repository\x18\x01 \x01(\v2\x12.gitaly.RepositoryB\x04\x98\xc6,\x01R\n" +
	"repository\x12,\n" +
	"\x12git_config_options\x18\x02 \x03(\tR\x10gitConfigOptions\x12!\n" +
	"\fgit_protocol\x18\x03 \x01(\tR\vgitProtocol\"\x95\x01\n" +
	"$SSHUploadPackWithSidechannelResponse\x12m\n" +
	"\x1fpackfile_negotiation_statistics\x18\x01 \x01(\v2%.gitaly.PackfileNegotiationStatisticsR\x1dpackfileNegotiationStatistics\"\x93\x02\n" +
	"\x15SSHReceivePackRequest\x128\n" +
	"\n" +
	"repository\x18\x01 \x01(\v2\x12.gitaly.RepositoryB\x04\x98\xc6,\x01R\n" +
	"repository\x12\x14\n" +
	"\x05stdin\x18\x02 \x01(\fR\x05stdin\x12\x13\n" +
	"\x05gl_id\x18\x03 \x01(\tR\x04glId\x12#\n" +
	"\rgl_repository\x18\x04 \x01(\tR\fglRepository\x12\x1f\n" +
	"\vgl_username\x18\x05 \x01(\tR\n" +
	"glUsername\x12!\n" +
	"\fgit_protocol\x18\x06 \x01(\tR\vgitProtocol\x12,\n" +
	"\x12git_config_options\x18\a \x03(\tR\x10gitConfigOptions\"}\n" +
	"\x16SSHReceivePackResponse\x12\x16\n" +
	"\x06stdout\x18\x01 \x01(\fR\x06stdout\x12\x16\n" +
	"\x06stderr\x18\x02 \x01(\fR\x06stderr\x123\n" +
	"\vexit_status\x18\x03 \x01(\v2\x12.gitaly.ExitStatusR\n" +
	"exitStatus\"i\n" +
	"\x17SSHUploadArchiveRequest\x128\n" +
	"\n" +
	"repository\x18\x01 \x01(\v2\x12.gitaly.RepositoryB\x04\x98\xc6,\x01R\n" +
	"repository\x12\x14\n" +
	"\x05stdin\x18\x02 \x01(\fR\x05stdin\"\x7f\n" +
	"\x18SSHUploadArchiveResponse\x12\x16\n" +
	"\x06stdout\x18\x01 \x01(\fR\x06stdout\x12\x16\n" +
	"\x06stderr\x18\x02 \x01(\fR\x06stderr\x123\n" +
	"\vexit_status\x18\x03 \x01(\v2\x12.gitaly.ExitStatusR\n" +
	"exitStatus2\xd0\x02\n" +
	"\n" +
	"SSHService\x12\x81\x01\n" +
	"\x1cSSHUploadPackWithSidechannel\x12+.gitaly.SSHUploadPackWithSidechannelRequest\x1a,.gitaly.SSHUploadPackWithSidechannelResponse\"\x06\xfa\x97(\x02\b\x02\x12[\n" +
	"\x0eSSHReceivePack\x12\x1d.gitaly.SSHReceivePackRequest\x1a\x1e.gitaly.SSHReceivePackResponse\"\x06\xfa\x97(\x02\b\x01(\x010\x01\x12a\n" +
	"\x10SSHUploadArchive\x12\x1f.gitaly.SSHUploadArchiveRequest\x1a .gitaly.SSHUploadArchiveResponse\"\x06\xfa\x97(\x02\b\x02(\x010\x01B4Z2gitlab.com/gitlab-org/gitaly/v18/proto/go/gitalypbb\x06proto3"

var (
	file_ssh_proto_rawDescOnce sync.Once
	file_ssh_proto_rawDescData []byte
)

func file_ssh_proto_rawDescGZIP() []byte {
	file_ssh_proto_rawDescOnce.Do(func() {
		file_ssh_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_ssh_proto_rawDesc), len(file_ssh_proto_rawDesc)))
	})
	return file_ssh_proto_rawDescData
}

var file_ssh_proto_msgTypes = make([]protoimpl.MessageInfo, 6)
var file_ssh_proto_goTypes = []any{
	(*SSHUploadPackWithSidechannelRequest)(nil),  // 0: gitaly.SSHUploadPackWithSidechannelRequest
	(*SSHUploadPackWithSidechannelResponse)(nil), // 1: gitaly.SSHUploadPackWithSidechannelResponse
	(*SSHReceivePackRequest)(nil),                // 2: gitaly.SSHReceivePackRequest
	(*SSHReceivePackResponse)(nil),               // 3: gitaly.SSHReceivePackResponse
	(*SSHUploadArchiveRequest)(nil),              // 4: gitaly.SSHUploadArchiveRequest
	(*SSHUploadArchiveResponse)(nil),             // 5: gitaly.SSHUploadArchiveResponse
	(*Repository)(nil),                           // 6: gitaly.Repository
	(*PackfileNegotiationStatistics)(nil),        // 7: gitaly.PackfileNegotiationStatistics
	(*ExitStatus)(nil),                           // 8: gitaly.ExitStatus
}
var file_ssh_proto_depIdxs = []int32{
	6, // 0: gitaly.SSHUploadPackWithSidechannelRequest.repository:type_name -> gitaly.Repository
	7, // 1: gitaly.SSHUploadPackWithSidechannelResponse.packfile_negotiation_statistics:type_name -> gitaly.PackfileNegotiationStatistics
	6, // 2: gitaly.SSHReceivePackRequest.repository:type_name -> gitaly.Repository
	8, // 3: gitaly.SSHReceivePackResponse.exit_status:type_name -> gitaly.ExitStatus
	6, // 4: gitaly.SSHUploadArchiveRequest.repository:type_name -> gitaly.Repository
	8, // 5: gitaly.SSHUploadArchiveResponse.exit_status:type_name -> gitaly.ExitStatus
	0, // 6: gitaly.SSHService.SSHUploadPackWithSidechannel:input_type -> gitaly.SSHUploadPackWithSidechannelRequest
	2, // 7: gitaly.SSHService.SSHReceivePack:input_type -> gitaly.SSHReceivePackRequest
	4, // 8: gitaly.SSHService.SSHUploadArchive:input_type -> gitaly.SSHUploadArchiveRequest
	1, // 9: gitaly.SSHService.SSHUploadPackWithSidechannel:output_type -> gitaly.SSHUploadPackWithSidechannelResponse
	3, // 10: gitaly.SSHService.SSHReceivePack:output_type -> gitaly.SSHReceivePackResponse
	5, // 11: gitaly.SSHService.SSHUploadArchive:output_type -> gitaly.SSHUploadArchiveResponse
	9, // [9:12] is the sub-list for method output_type
	6, // [6:9] is the sub-list for method input_type
	6, // [6:6] is the sub-list for extension type_name
	6, // [6:6] is the sub-list for extension extendee
	0, // [0:6] is the sub-list for field type_name
}

func init() { file_ssh_proto_init() }
func file_ssh_proto_init() {
	if File_ssh_proto != nil {
		return
	}
	file_lint_proto_init()
	file_packfile_proto_init()
	file_shared_proto_init()
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_ssh_proto_rawDesc), len(file_ssh_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   6,
			NumExtensions: 0,
			NumServices:   1,
		},
		GoTypes:           file_ssh_proto_goTypes,
		DependencyIndexes: file_ssh_proto_depIdxs,
		MessageInfos:      file_ssh_proto_msgTypes,
	}.Build()
	File_ssh_proto = out.File
	file_ssh_proto_goTypes = nil
	file_ssh_proto_depIdxs = nil
}
