// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v6.30.2
// source: conflicts.proto

package gitalypb

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	timestamppb "google.golang.org/protobuf/types/known/timestamppb"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

// ListConflictFilesRequest is the request for the ListConflictFilesRequest rpc.
type ListConflictFilesRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// repository is the repository for which we want to list the conflicted files.
	Repository *Repository `protobuf:"bytes,1,opt,name=repository,proto3" json:"repository,omitempty"`
	// our_commit_oid is the tree-ish OID to merge into.
	OurCommitOid string `protobuf:"bytes,2,opt,name=our_commit_oid,json=ourCommitOid,proto3" json:"our_commit_oid,omitempty"`
	// their_commit_oid is the tree-ish OID to merge from.
	TheirCommitOid string `protobuf:"bytes,3,opt,name=their_commit_oid,json=theirCommitOid,proto3" json:"their_commit_oid,omitempty"`
	// allow_tree_conflicts will not cause the request to fail in case there are
	// tree conflicts. If set to true, then responses may contain conflict files
	// where some of the paths are unset.
	AllowTreeConflicts bool `protobuf:"varint,4,opt,name=allow_tree_conflicts,json=allowTreeConflicts,proto3" json:"allow_tree_conflicts,omitempty"`
	// skip_content will skip the parsing and streaming of conflicted file's content.
	// This can be useful when we only want to know if there is a conflict and which files
	// are conflicted but don't care about the contents of the conflicted files.
	SkipContent   bool `protobuf:"varint,5,opt,name=skip_content,json=skipContent,proto3" json:"skip_content,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ListConflictFilesRequest) Reset() {
	*x = ListConflictFilesRequest{}
	mi := &file_conflicts_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ListConflictFilesRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ListConflictFilesRequest) ProtoMessage() {}

func (x *ListConflictFilesRequest) ProtoReflect() protoreflect.Message {
	mi := &file_conflicts_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ListConflictFilesRequest.ProtoReflect.Descriptor instead.
func (*ListConflictFilesRequest) Descriptor() ([]byte, []int) {
	return file_conflicts_proto_rawDescGZIP(), []int{0}
}

func (x *ListConflictFilesRequest) GetRepository() *Repository {
	if x != nil {
		return x.Repository
	}
	return nil
}

func (x *ListConflictFilesRequest) GetOurCommitOid() string {
	if x != nil {
		return x.OurCommitOid
	}
	return ""
}

func (x *ListConflictFilesRequest) GetTheirCommitOid() string {
	if x != nil {
		return x.TheirCommitOid
	}
	return ""
}

func (x *ListConflictFilesRequest) GetAllowTreeConflicts() bool {
	if x != nil {
		return x.AllowTreeConflicts
	}
	return false
}

func (x *ListConflictFilesRequest) GetSkipContent() bool {
	if x != nil {
		return x.SkipContent
	}
	return false
}

// ConflictFileHeader contains parsed conflicted file information from the `git-merge-tree(1)`
// output for an individual file.
type ConflictFileHeader struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// commit_oid is the tree-ish revision being merged into.
	CommitOid string `protobuf:"bytes,2,opt,name=commit_oid,json=commitOid,proto3" json:"commit_oid,omitempty"`
	// their_path is the path of the conflicted file being merged from.
	TheirPath []byte `protobuf:"bytes,3,opt,name=their_path,json=theirPath,proto3" json:"their_path,omitempty"`
	// our_path is the path of the conflicted file being merged into.
	OurPath []byte `protobuf:"bytes,4,opt,name=our_path,json=ourPath,proto3" json:"our_path,omitempty"`
	// our_mode is the mode of the conflicted file being merged into.
	OurMode int32 `protobuf:"varint,5,opt,name=our_mode,json=ourMode,proto3" json:"our_mode,omitempty"`
	// ancestor_path is path of the conflicted file for the merge base.
	AncestorPath  []byte `protobuf:"bytes,6,opt,name=ancestor_path,json=ancestorPath,proto3" json:"ancestor_path,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ConflictFileHeader) Reset() {
	*x = ConflictFileHeader{}
	mi := &file_conflicts_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ConflictFileHeader) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ConflictFileHeader) ProtoMessage() {}

func (x *ConflictFileHeader) ProtoReflect() protoreflect.Message {
	mi := &file_conflicts_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ConflictFileHeader.ProtoReflect.Descriptor instead.
func (*ConflictFileHeader) Descriptor() ([]byte, []int) {
	return file_conflicts_proto_rawDescGZIP(), []int{1}
}

func (x *ConflictFileHeader) GetCommitOid() string {
	if x != nil {
		return x.CommitOid
	}
	return ""
}

func (x *ConflictFileHeader) GetTheirPath() []byte {
	if x != nil {
		return x.TheirPath
	}
	return nil
}

func (x *ConflictFileHeader) GetOurPath() []byte {
	if x != nil {
		return x.OurPath
	}
	return nil
}

func (x *ConflictFileHeader) GetOurMode() int32 {
	if x != nil {
		return x.OurMode
	}
	return 0
}

func (x *ConflictFileHeader) GetAncestorPath() []byte {
	if x != nil {
		return x.AncestorPath
	}
	return nil
}

// ConflictFile contains conflicted file information.
type ConflictFile struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// Types that are valid to be assigned to ConflictFilePayload:
	//
	//	*ConflictFile_Header
	//	*ConflictFile_Content
	ConflictFilePayload isConflictFile_ConflictFilePayload `protobuf_oneof:"conflict_file_payload"`
	unknownFields       protoimpl.UnknownFields
	sizeCache           protoimpl.SizeCache
}

func (x *ConflictFile) Reset() {
	*x = ConflictFile{}
	mi := &file_conflicts_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ConflictFile) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ConflictFile) ProtoMessage() {}

func (x *ConflictFile) ProtoReflect() protoreflect.Message {
	mi := &file_conflicts_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ConflictFile.ProtoReflect.Descriptor instead.
func (*ConflictFile) Descriptor() ([]byte, []int) {
	return file_conflicts_proto_rawDescGZIP(), []int{2}
}

func (x *ConflictFile) GetConflictFilePayload() isConflictFile_ConflictFilePayload {
	if x != nil {
		return x.ConflictFilePayload
	}
	return nil
}

func (x *ConflictFile) GetHeader() *ConflictFileHeader {
	if x != nil {
		if x, ok := x.ConflictFilePayload.(*ConflictFile_Header); ok {
			return x.Header
		}
	}
	return nil
}

func (x *ConflictFile) GetContent() []byte {
	if x != nil {
		if x, ok := x.ConflictFilePayload.(*ConflictFile_Content); ok {
			return x.Content
		}
	}
	return nil
}

type isConflictFile_ConflictFilePayload interface {
	isConflictFile_ConflictFilePayload()
}

type ConflictFile_Header struct {
	// header contains the header information about a conflicted file.
	Header *ConflictFileHeader `protobuf:"bytes,1,opt,name=header,proto3,oneof"`
}

type ConflictFile_Content struct {
	// content is the content data from the conflicted file. The file data may be broken into chunks
	// and sent in the stream following a file header. if skip_content is set in the request,
	// conflicted file content is not streamed.
	Content []byte `protobuf:"bytes,2,opt,name=content,proto3,oneof"`
}

func (*ConflictFile_Header) isConflictFile_ConflictFilePayload() {}

func (*ConflictFile_Content) isConflictFile_ConflictFilePayload() {}

// ListConflictFilesResponse is the response for the ListConflicts RPC.
type ListConflictFilesResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// files is a list of conflicted file information being sent in the stream. This information
	// contains conflicted file headers and optionally the conflicted file content.
	Files         []*ConflictFile `protobuf:"bytes,1,rep,name=files,proto3" json:"files,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ListConflictFilesResponse) Reset() {
	*x = ListConflictFilesResponse{}
	mi := &file_conflicts_proto_msgTypes[3]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ListConflictFilesResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ListConflictFilesResponse) ProtoMessage() {}

func (x *ListConflictFilesResponse) ProtoReflect() protoreflect.Message {
	mi := &file_conflicts_proto_msgTypes[3]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ListConflictFilesResponse.ProtoReflect.Descriptor instead.
func (*ListConflictFilesResponse) Descriptor() ([]byte, []int) {
	return file_conflicts_proto_rawDescGZIP(), []int{3}
}

func (x *ListConflictFilesResponse) GetFiles() []*ConflictFile {
	if x != nil {
		return x.Files
	}
	return nil
}

// ResolveConflictsRequestHeader is the first message that must be sent for
// each ResolveConflicts call.
type ResolveConflictsRequestHeader struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// repository is the repository in which conflicts shall be resolved and
	// where SourceBranch shall be updated with the resolved conflict.
	Repository *Repository `protobuf:"bytes,1,opt,name=repository,proto3" json:"repository,omitempty"`
	// our_commit_oid is the OID of the commit representing the local commit.
	OurCommitOid string `protobuf:"bytes,2,opt,name=our_commit_oid,json=ourCommitOid,proto3" json:"our_commit_oid,omitempty"`
	// target_repository is the repository from which TheirCommitOid shall be
	// retrieved.
	TargetRepository *Repository `protobuf:"bytes,3,opt,name=target_repository,json=targetRepository,proto3" json:"target_repository,omitempty"`
	// their_commit_oid is the OID of the commit representing the remote commit
	// which is to be merged into the local commit.
	TheirCommitOid string `protobuf:"bytes,4,opt,name=their_commit_oid,json=theirCommitOid,proto3" json:"their_commit_oid,omitempty"`
	// source_branch is the branch on which the new commit shall be created.
	SourceBranch []byte `protobuf:"bytes,5,opt,name=source_branch,json=sourceBranch,proto3" json:"source_branch,omitempty"`
	// target_branch identifies the branch which will be fetched from
	// target_repository in case TheirCommitOid does not exist in Repository.
	TargetBranch []byte `protobuf:"bytes,6,opt,name=target_branch,json=targetBranch,proto3" json:"target_branch,omitempty"`
	// commit_message is the message of the newly created merge commit.
	CommitMessage []byte `protobuf:"bytes,7,opt,name=commit_message,json=commitMessage,proto3" json:"commit_message,omitempty"`
	// user is the user used as author and committer of the newly created merge
	// commit.
	User *User `protobuf:"bytes,8,opt,name=user,proto3" json:"user,omitempty"`
	// timestamp is the optional timestamp to use for the commit as committer
	// date. If it's not set, the current time will be used.
	Timestamp     *timestamppb.Timestamp `protobuf:"bytes,9,opt,name=timestamp,proto3" json:"timestamp,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ResolveConflictsRequestHeader) Reset() {
	*x = ResolveConflictsRequestHeader{}
	mi := &file_conflicts_proto_msgTypes[4]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ResolveConflictsRequestHeader) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ResolveConflictsRequestHeader) ProtoMessage() {}

func (x *ResolveConflictsRequestHeader) ProtoReflect() protoreflect.Message {
	mi := &file_conflicts_proto_msgTypes[4]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ResolveConflictsRequestHeader.ProtoReflect.Descriptor instead.
func (*ResolveConflictsRequestHeader) Descriptor() ([]byte, []int) {
	return file_conflicts_proto_rawDescGZIP(), []int{4}
}

func (x *ResolveConflictsRequestHeader) GetRepository() *Repository {
	if x != nil {
		return x.Repository
	}
	return nil
}

func (x *ResolveConflictsRequestHeader) GetOurCommitOid() string {
	if x != nil {
		return x.OurCommitOid
	}
	return ""
}

func (x *ResolveConflictsRequestHeader) GetTargetRepository() *Repository {
	if x != nil {
		return x.TargetRepository
	}
	return nil
}

func (x *ResolveConflictsRequestHeader) GetTheirCommitOid() string {
	if x != nil {
		return x.TheirCommitOid
	}
	return ""
}

func (x *ResolveConflictsRequestHeader) GetSourceBranch() []byte {
	if x != nil {
		return x.SourceBranch
	}
	return nil
}

func (x *ResolveConflictsRequestHeader) GetTargetBranch() []byte {
	if x != nil {
		return x.TargetBranch
	}
	return nil
}

func (x *ResolveConflictsRequestHeader) GetCommitMessage() []byte {
	if x != nil {
		return x.CommitMessage
	}
	return nil
}

func (x *ResolveConflictsRequestHeader) GetUser() *User {
	if x != nil {
		return x.User
	}
	return nil
}

func (x *ResolveConflictsRequestHeader) GetTimestamp() *timestamppb.Timestamp {
	if x != nil {
		return x.Timestamp
	}
	return nil
}

// ResolveConflictsRequest is a request for the ResolveConflicts RPC.
type ResolveConflictsRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// RequestPayload is the payload part of the request. The first message sent
	// must always be a ResolveConflictsRequestHeader, whereas all remaining
	// requests must be FilesJson requests.
	//
	// Types that are valid to be assigned to ResolveConflictsRequestPayload:
	//
	//	*ResolveConflictsRequest_Header
	//	*ResolveConflictsRequest_FilesJson
	ResolveConflictsRequestPayload isResolveConflictsRequest_ResolveConflictsRequestPayload `protobuf_oneof:"resolve_conflicts_request_payload"`
	unknownFields                  protoimpl.UnknownFields
	sizeCache                      protoimpl.SizeCache
}

func (x *ResolveConflictsRequest) Reset() {
	*x = ResolveConflictsRequest{}
	mi := &file_conflicts_proto_msgTypes[5]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ResolveConflictsRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ResolveConflictsRequest) ProtoMessage() {}

func (x *ResolveConflictsRequest) ProtoReflect() protoreflect.Message {
	mi := &file_conflicts_proto_msgTypes[5]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ResolveConflictsRequest.ProtoReflect.Descriptor instead.
func (*ResolveConflictsRequest) Descriptor() ([]byte, []int) {
	return file_conflicts_proto_rawDescGZIP(), []int{5}
}

func (x *ResolveConflictsRequest) GetResolveConflictsRequestPayload() isResolveConflictsRequest_ResolveConflictsRequestPayload {
	if x != nil {
		return x.ResolveConflictsRequestPayload
	}
	return nil
}

func (x *ResolveConflictsRequest) GetHeader() *ResolveConflictsRequestHeader {
	if x != nil {
		if x, ok := x.ResolveConflictsRequestPayload.(*ResolveConflictsRequest_Header); ok {
			return x.Header
		}
	}
	return nil
}

func (x *ResolveConflictsRequest) GetFilesJson() []byte {
	if x != nil {
		if x, ok := x.ResolveConflictsRequestPayload.(*ResolveConflictsRequest_FilesJson); ok {
			return x.FilesJson
		}
	}
	return nil
}

type isResolveConflictsRequest_ResolveConflictsRequestPayload interface {
	isResolveConflictsRequest_ResolveConflictsRequestPayload()
}

type ResolveConflictsRequest_Header struct {
	// header is the initial message specifying parameters of the RPC call.
	Header *ResolveConflictsRequestHeader `protobuf:"bytes,1,opt,name=header,proto3,oneof"`
}

type ResolveConflictsRequest_FilesJson struct {
	// files_json is a JSON-encoded list of conflicts resolutions.
	FilesJson []byte `protobuf:"bytes,2,opt,name=files_json,json=filesJson,proto3,oneof"`
}

func (*ResolveConflictsRequest_Header) isResolveConflictsRequest_ResolveConflictsRequestPayload() {}

func (*ResolveConflictsRequest_FilesJson) isResolveConflictsRequest_ResolveConflictsRequestPayload() {
}

// ResolveConflictsResponse is a response of the ResolveConflicts RPC. Conflict
// resolution may have failed even if the RPC has returned OK. The user must
// check ResolutionError to verify whether the merge commit was correctly
// computed or not.
type ResolveConflictsResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// resolution_error contains a description of why conflict resolution has
	// failed.
	ResolutionError string `protobuf:"bytes,1,opt,name=resolution_error,json=resolutionError,proto3" json:"resolution_error,omitempty"`
	unknownFields   protoimpl.UnknownFields
	sizeCache       protoimpl.SizeCache
}

func (x *ResolveConflictsResponse) Reset() {
	*x = ResolveConflictsResponse{}
	mi := &file_conflicts_proto_msgTypes[6]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ResolveConflictsResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ResolveConflictsResponse) ProtoMessage() {}

func (x *ResolveConflictsResponse) ProtoReflect() protoreflect.Message {
	mi := &file_conflicts_proto_msgTypes[6]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ResolveConflictsResponse.ProtoReflect.Descriptor instead.
func (*ResolveConflictsResponse) Descriptor() ([]byte, []int) {
	return file_conflicts_proto_rawDescGZIP(), []int{6}
}

func (x *ResolveConflictsResponse) GetResolutionError() string {
	if x != nil {
		return x.ResolutionError
	}
	return ""
}

var File_conflicts_proto protoreflect.FileDescriptor

const file_conflicts_proto_rawDesc = "" +
	"\n" +
	"\x0fconflicts.proto\x12\x06gitaly\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\n" +
	"lint.proto\x1a\fshared.proto\"\xf9\x01\n" +
	"\x18ListConflictFilesRequest\x128\n" +
	"\n" +
	"repository\x18\x01 \x01(\v2\x12.gitaly.RepositoryB\x04\x98\xc6,\x01R\n" +
	"repository\x12$\n" +
	"\x0eour_commit_oid\x18\x02 \x01(\tR\fourCommitOid\x12(\n" +
	"\x10their_commit_oid\x18\x03 \x01(\tR\x0etheirCommitOid\x120\n" +
	"\x14allow_tree_conflicts\x18\x04 \x01(\bR\x12allowTreeConflicts\x12!\n" +
	"\fskip_content\x18\x05 \x01(\bR\vskipContent\"\xb3\x01\n" +
	"\x12ConflictFileHeader\x12\x1d\n" +
	"\n" +
	"commit_oid\x18\x02 \x01(\tR\tcommitOid\x12\x1d\n" +
	"\n" +
	"their_path\x18\x03 \x01(\fR\ttheirPath\x12\x19\n" +
	"\bour_path\x18\x04 \x01(\fR\aourPath\x12\x19\n" +
	"\bour_mode\x18\x05 \x01(\x05R\aourMode\x12#\n" +
	"\rancestor_path\x18\x06 \x01(\fR\fancestorPathJ\x04\b\x01\x10\x02\"y\n" +
	"\fConflictFile\x124\n" +
	"\x06header\x18\x01 \x01(\v2\x1a.gitaly.ConflictFileHeaderH\x00R\x06header\x12\x1a\n" +
	"\acontent\x18\x02 \x01(\fH\x00R\acontentB\x17\n" +
	"\x15conflict_file_payload\"G\n" +
	"\x19ListConflictFilesResponse\x12*\n" +
	"\x05files\x18\x01 \x03(\v2\x14.gitaly.ConflictFileR\x05files\"\xb7\x03\n" +
	"\x1dResolveConflictsRequestHeader\x128\n" +
	"\n" +
	"repository\x18\x01 \x01(\v2\x12.gitaly.RepositoryB\x04\x98\xc6,\x01R\n" +
	"repository\x12$\n" +
	"\x0eour_commit_oid\x18\x02 \x01(\tR\fourCommitOid\x12?\n" +
	"\x11target_repository\x18\x03 \x01(\v2\x12.gitaly.RepositoryR\x10targetRepository\x12(\n" +
	"\x10their_commit_oid\x18\x04 \x01(\tR\x0etheirCommitOid\x12#\n" +
	"\rsource_branch\x18\x05 \x01(\fR\fsourceBranch\x12#\n" +
	"\rtarget_branch\x18\x06 \x01(\fR\ftargetBranch\x12%\n" +
	"\x0ecommit_message\x18\a \x01(\fR\rcommitMessage\x12 \n" +
	"\x04user\x18\b \x01(\v2\f.gitaly.UserR\x04user\x128\n" +
	"\ttimestamp\x18\t \x01(\v2\x1a.google.protobuf.TimestampR\ttimestamp\"\xa0\x01\n" +
	"\x17ResolveConflictsRequest\x12?\n" +
	"\x06header\x18\x01 \x01(\v2%.gitaly.ResolveConflictsRequestHeaderH\x00R\x06header\x12\x1f\n" +
	"\n" +
	"files_json\x18\x02 \x01(\fH\x00R\tfilesJsonB#\n" +
	"!resolve_conflicts_request_payload\"E\n" +
	"\x18ResolveConflictsResponse\x12)\n" +
	"\x10resolution_error\x18\x01 \x01(\tR\x0fresolutionError2\xd7\x01\n" +
	"\x10ConflictsService\x12b\n" +
	"\x11ListConflictFiles\x12 .gitaly.ListConflictFilesRequest\x1a!.gitaly.ListConflictFilesResponse\"\x06\xfa\x97(\x02\b\x020\x01\x12_\n" +
	"\x10ResolveConflicts\x12\x1f.gitaly.ResolveConflictsRequest\x1a .gitaly.ResolveConflictsResponse\"\x06\xfa\x97(\x02\b\x01(\x01B4Z2gitlab.com/gitlab-org/gitaly/v18/proto/go/gitalypbb\x06proto3"

var (
	file_conflicts_proto_rawDescOnce sync.Once
	file_conflicts_proto_rawDescData []byte
)

func file_conflicts_proto_rawDescGZIP() []byte {
	file_conflicts_proto_rawDescOnce.Do(func() {
		file_conflicts_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_conflicts_proto_rawDesc), len(file_conflicts_proto_rawDesc)))
	})
	return file_conflicts_proto_rawDescData
}

var file_conflicts_proto_msgTypes = make([]protoimpl.MessageInfo, 7)
var file_conflicts_proto_goTypes = []any{
	(*ListConflictFilesRequest)(nil),      // 0: gitaly.ListConflictFilesRequest
	(*ConflictFileHeader)(nil),            // 1: gitaly.ConflictFileHeader
	(*ConflictFile)(nil),                  // 2: gitaly.ConflictFile
	(*ListConflictFilesResponse)(nil),     // 3: gitaly.ListConflictFilesResponse
	(*ResolveConflictsRequestHeader)(nil), // 4: gitaly.ResolveConflictsRequestHeader
	(*ResolveConflictsRequest)(nil),       // 5: gitaly.ResolveConflictsRequest
	(*ResolveConflictsResponse)(nil),      // 6: gitaly.ResolveConflictsResponse
	(*Repository)(nil),                    // 7: gitaly.Repository
	(*User)(nil),                          // 8: gitaly.User
	(*timestamppb.Timestamp)(nil),         // 9: google.protobuf.Timestamp
}
var file_conflicts_proto_depIdxs = []int32{
	7,  // 0: gitaly.ListConflictFilesRequest.repository:type_name -> gitaly.Repository
	1,  // 1: gitaly.ConflictFile.header:type_name -> gitaly.ConflictFileHeader
	2,  // 2: gitaly.ListConflictFilesResponse.files:type_name -> gitaly.ConflictFile
	7,  // 3: gitaly.ResolveConflictsRequestHeader.repository:type_name -> gitaly.Repository
	7,  // 4: gitaly.ResolveConflictsRequestHeader.target_repository:type_name -> gitaly.Repository
	8,  // 5: gitaly.ResolveConflictsRequestHeader.user:type_name -> gitaly.User
	9,  // 6: gitaly.ResolveConflictsRequestHeader.timestamp:type_name -> google.protobuf.Timestamp
	4,  // 7: gitaly.ResolveConflictsRequest.header:type_name -> gitaly.ResolveConflictsRequestHeader
	0,  // 8: gitaly.ConflictsService.ListConflictFiles:input_type -> gitaly.ListConflictFilesRequest
	5,  // 9: gitaly.ConflictsService.ResolveConflicts:input_type -> gitaly.ResolveConflictsRequest
	3,  // 10: gitaly.ConflictsService.ListConflictFiles:output_type -> gitaly.ListConflictFilesResponse
	6,  // 11: gitaly.ConflictsService.ResolveConflicts:output_type -> gitaly.ResolveConflictsResponse
	10, // [10:12] is the sub-list for method output_type
	8,  // [8:10] is the sub-list for method input_type
	8,  // [8:8] is the sub-list for extension type_name
	8,  // [8:8] is the sub-list for extension extendee
	0,  // [0:8] is the sub-list for field type_name
}

func init() { file_conflicts_proto_init() }
func file_conflicts_proto_init() {
	if File_conflicts_proto != nil {
		return
	}
	file_lint_proto_init()
	file_shared_proto_init()
	file_conflicts_proto_msgTypes[2].OneofWrappers = []any{
		(*ConflictFile_Header)(nil),
		(*ConflictFile_Content)(nil),
	}
	file_conflicts_proto_msgTypes[5].OneofWrappers = []any{
		(*ResolveConflictsRequest_Header)(nil),
		(*ResolveConflictsRequest_FilesJson)(nil),
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_conflicts_proto_rawDesc), len(file_conflicts_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   7,
			NumExtensions: 0,
			NumServices:   1,
		},
		GoTypes:           file_conflicts_proto_goTypes,
		DependencyIndexes: file_conflicts_proto_depIdxs,
		MessageInfos:      file_conflicts_proto_msgTypes,
	}.Build()
	File_conflicts_proto = out.File
	file_conflicts_proto_goTypes = nil
	file_conflicts_proto_depIdxs = nil
}
