// Code generated by protoc-gen-go-grpc. DO NOT EDIT.
// versions:
// - protoc-gen-go-grpc v1.6.1
// - protoc             v6.30.2
// source: backup.proto

package gitalypb

import (
	context "context"
	grpc "google.golang.org/grpc"
	codes "google.golang.org/grpc/codes"
	status "google.golang.org/grpc/status"
)

// This is a compile-time assertion to ensure that this generated file
// is compatible with the grpc package it is being compiled against.
// Requires gRPC-Go v1.64.0 or later.
const _ = grpc.SupportPackageIsVersion9

const (
	BackupService_WriteBackupID_FullMethodName      = "/gitaly.BackupService/WriteBackupID"
	BackupService_ReadLatestBackupID_FullMethodName = "/gitaly.BackupService/ReadLatestBackupID"
)

// BackupServiceClient is the client API for BackupService service.
//
// For semantics around ctx use and closing/ending streaming RPCs, please refer to https://pkg.go.dev/google.golang.org/grpc/?tab=doc#ClientConn.NewStream.
//
// BackupService provides RPCs for managing backup run metadata.
type BackupServiceClient interface {
	// WriteBackupID writes a backup ID marker to object-storage so that the
	// latest completed backup run can be discovered later.
	WriteBackupID(ctx context.Context, in *WriteBackupIDRequest, opts ...grpc.CallOption) (*WriteBackupIDResponse, error)
	// ReadLatestBackupID returns the ID of the most recently completed backup
	// run by iterating the backup_ids/ prefix in object-storage.
	ReadLatestBackupID(ctx context.Context, in *ReadLatestBackupIDRequest, opts ...grpc.CallOption) (*ReadLatestBackupIDResponse, error)
}

type backupServiceClient struct {
	cc grpc.ClientConnInterface
}

func NewBackupServiceClient(cc grpc.ClientConnInterface) BackupServiceClient {
	return &backupServiceClient{cc}
}

func (c *backupServiceClient) WriteBackupID(ctx context.Context, in *WriteBackupIDRequest, opts ...grpc.CallOption) (*WriteBackupIDResponse, error) {
	cOpts := append([]grpc.CallOption{grpc.StaticMethod()}, opts...)
	out := new(WriteBackupIDResponse)
	err := c.cc.Invoke(ctx, BackupService_WriteBackupID_FullMethodName, in, out, cOpts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

func (c *backupServiceClient) ReadLatestBackupID(ctx context.Context, in *ReadLatestBackupIDRequest, opts ...grpc.CallOption) (*ReadLatestBackupIDResponse, error) {
	cOpts := append([]grpc.CallOption{grpc.StaticMethod()}, opts...)
	out := new(ReadLatestBackupIDResponse)
	err := c.cc.Invoke(ctx, BackupService_ReadLatestBackupID_FullMethodName, in, out, cOpts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

// BackupServiceServer is the server API for BackupService service.
// All implementations must embed UnimplementedBackupServiceServer
// for forward compatibility.
//
// BackupService provides RPCs for managing backup run metadata.
type BackupServiceServer interface {
	// WriteBackupID writes a backup ID marker to object-storage so that the
	// latest completed backup run can be discovered later.
	WriteBackupID(context.Context, *WriteBackupIDRequest) (*WriteBackupIDResponse, error)
	// ReadLatestBackupID returns the ID of the most recently completed backup
	// run by iterating the backup_ids/ prefix in object-storage.
	ReadLatestBackupID(context.Context, *ReadLatestBackupIDRequest) (*ReadLatestBackupIDResponse, error)
	mustEmbedUnimplementedBackupServiceServer()
}

// UnimplementedBackupServiceServer must be embedded to have
// forward compatible implementations.
//
// NOTE: this should be embedded by value instead of pointer to avoid a nil
// pointer dereference when methods are called.
type UnimplementedBackupServiceServer struct{}

func (UnimplementedBackupServiceServer) WriteBackupID(context.Context, *WriteBackupIDRequest) (*WriteBackupIDResponse, error) {
	return nil, status.Error(codes.Unimplemented, "method WriteBackupID not implemented")
}
func (UnimplementedBackupServiceServer) ReadLatestBackupID(context.Context, *ReadLatestBackupIDRequest) (*ReadLatestBackupIDResponse, error) {
	return nil, status.Error(codes.Unimplemented, "method ReadLatestBackupID not implemented")
}
func (UnimplementedBackupServiceServer) mustEmbedUnimplementedBackupServiceServer() {}
func (UnimplementedBackupServiceServer) testEmbeddedByValue()                       {}

// UnsafeBackupServiceServer may be embedded to opt out of forward compatibility for this service.
// Use of this interface is not recommended, as added methods to BackupServiceServer will
// result in compilation errors.
type UnsafeBackupServiceServer interface {
	mustEmbedUnimplementedBackupServiceServer()
}

func RegisterBackupServiceServer(s grpc.ServiceRegistrar, srv BackupServiceServer) {
	// If the following call panics, it indicates UnimplementedBackupServiceServer was
	// embedded by pointer and is nil.  This will cause panics if an
	// unimplemented method is ever invoked, so we test this at initialization
	// time to prevent it from happening at runtime later due to I/O.
	if t, ok := srv.(interface{ testEmbeddedByValue() }); ok {
		t.testEmbeddedByValue()
	}
	s.RegisterService(&BackupService_ServiceDesc, srv)
}

func _BackupService_WriteBackupID_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(WriteBackupIDRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(BackupServiceServer).WriteBackupID(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: BackupService_WriteBackupID_FullMethodName,
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(BackupServiceServer).WriteBackupID(ctx, req.(*WriteBackupIDRequest))
	}
	return interceptor(ctx, in, info, handler)
}

func _BackupService_ReadLatestBackupID_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(ReadLatestBackupIDRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(BackupServiceServer).ReadLatestBackupID(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: BackupService_ReadLatestBackupID_FullMethodName,
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(BackupServiceServer).ReadLatestBackupID(ctx, req.(*ReadLatestBackupIDRequest))
	}
	return interceptor(ctx, in, info, handler)
}

// BackupService_ServiceDesc is the grpc.ServiceDesc for BackupService service.
// It's only intended for direct use with grpc.RegisterService,
// and not to be introspected or modified (even as a copy)
var BackupService_ServiceDesc = grpc.ServiceDesc{
	ServiceName: "gitaly.BackupService",
	HandlerType: (*BackupServiceServer)(nil),
	Methods: []grpc.MethodDesc{
		{
			MethodName: "WriteBackupID",
			Handler:    _BackupService_WriteBackupID_Handler,
		},
		{
			MethodName: "ReadLatestBackupID",
			Handler:    _BackupService_ReadLatestBackupID_Handler,
		},
	},
	Streams:  []grpc.StreamDesc{},
	Metadata: "backup.proto",
}
