// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v6.30.2
// source: errors.proto

package gitalypb

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	durationpb "google.golang.org/protobuf/types/known/durationpb"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

// ErrorType is the type of error encountered on the index operation.
type IndexError_ErrorType int32

const (
	// ERROR_TYPE_UNSPECIFIED is the default error type and should never be set.
	IndexError_ERROR_TYPE_UNSPECIFIED IndexError_ErrorType = 0
	// ERROR_TYPE_EMPTY_PATH indicates an empty path was provided by the caller.
	IndexError_ERROR_TYPE_EMPTY_PATH IndexError_ErrorType = 1
	// ERROR_TYPE_INVALID_PATH indicates a path either contained '.git', or was
	// incorrectly formatted, e.g. invalid://file/path.
	IndexError_ERROR_TYPE_INVALID_PATH IndexError_ErrorType = 2
	// ERROR_TYPE_DIRECTORY_EXISTS indicates the directory being created already exists.
	IndexError_ERROR_TYPE_DIRECTORY_EXISTS IndexError_ErrorType = 3
	// ERROR_TYPE_DIRECTORY_TRAVERSAL indicates the path contains a directory traversal
	// sequence.
	IndexError_ERROR_TYPE_DIRECTORY_TRAVERSAL IndexError_ErrorType = 4
	// ERROR_TYPE_FILE_EXISTS indicates the file being created already exists.
	IndexError_ERROR_TYPE_FILE_EXISTS IndexError_ErrorType = 5
	// ERROR_TYPE_FILE_NOT_FOUND indicates the file being updated does not exist.
	IndexError_ERROR_TYPE_FILE_NOT_FOUND IndexError_ErrorType = 6
)

// Enum value maps for IndexError_ErrorType.
var (
	IndexError_ErrorType_name = map[int32]string{
		0: "ERROR_TYPE_UNSPECIFIED",
		1: "ERROR_TYPE_EMPTY_PATH",
		2: "ERROR_TYPE_INVALID_PATH",
		3: "ERROR_TYPE_DIRECTORY_EXISTS",
		4: "ERROR_TYPE_DIRECTORY_TRAVERSAL",
		5: "ERROR_TYPE_FILE_EXISTS",
		6: "ERROR_TYPE_FILE_NOT_FOUND",
	}
	IndexError_ErrorType_value = map[string]int32{
		"ERROR_TYPE_UNSPECIFIED":         0,
		"ERROR_TYPE_EMPTY_PATH":          1,
		"ERROR_TYPE_INVALID_PATH":        2,
		"ERROR_TYPE_DIRECTORY_EXISTS":    3,
		"ERROR_TYPE_DIRECTORY_TRAVERSAL": 4,
		"ERROR_TYPE_FILE_EXISTS":         5,
		"ERROR_TYPE_FILE_NOT_FOUND":      6,
	}
)

func (x IndexError_ErrorType) Enum() *IndexError_ErrorType {
	p := new(IndexError_ErrorType)
	*p = x
	return p
}

func (x IndexError_ErrorType) String() string {
	return protoimpl.X.EnumStringOf(x.Descriptor(), protoreflect.EnumNumber(x))
}

func (IndexError_ErrorType) Descriptor() protoreflect.EnumDescriptor {
	return file_errors_proto_enumTypes[0].Descriptor()
}

func (IndexError_ErrorType) Type() protoreflect.EnumType {
	return &file_errors_proto_enumTypes[0]
}

func (x IndexError_ErrorType) Number() protoreflect.EnumNumber {
	return protoreflect.EnumNumber(x)
}

// Deprecated: Use IndexError_ErrorType.Descriptor instead.
func (IndexError_ErrorType) EnumDescriptor() ([]byte, []int) {
	return file_errors_proto_rawDescGZIP(), []int{1, 0}
}

// HookType is the type of the hook that has been running. Please consult githooks(5) for more
// information about the specific types.
type CustomHookError_HookType int32

const (
	// HOOK_TYPE_UNSPECIFIED is the default hook type and should never be set.
	CustomHookError_HOOK_TYPE_UNSPECIFIED CustomHookError_HookType = 0
	// HOOK_TYPE_PRERECEIVE is executed after all changes have been written into a temporary staging
	// directory, but before any references in the repository have been updated. It is executed with
	// all references that are about to be updated at once. If this hook exits, then no references
	// will have been updated in the repository and staged objects will have been discarded.
	CustomHookError_HOOK_TYPE_PRERECEIVE CustomHookError_HookType = 1
	// HOOK_TYPE_UPDATE is executed after the pre-receive hook. It is executed per reference that is
	// about to be updated and can be used to reject only a subset of reference updates. If this
	// hook error is raised then a subset of references may have already been updated.
	CustomHookError_HOOK_TYPE_UPDATE CustomHookError_HookType = 2
	// HOOK_TYPE_POSTRECEIVE is executed after objects have been migrated into the repository and
	// after references have been updated. An error in this hook will not impact the changes
	// anymore as everything has already been persisted.
	CustomHookError_HOOK_TYPE_POSTRECEIVE CustomHookError_HookType = 3
)

// Enum value maps for CustomHookError_HookType.
var (
	CustomHookError_HookType_name = map[int32]string{
		0: "HOOK_TYPE_UNSPECIFIED",
		1: "HOOK_TYPE_PRERECEIVE",
		2: "HOOK_TYPE_UPDATE",
		3: "HOOK_TYPE_POSTRECEIVE",
	}
	CustomHookError_HookType_value = map[string]int32{
		"HOOK_TYPE_UNSPECIFIED": 0,
		"HOOK_TYPE_PRERECEIVE":  1,
		"HOOK_TYPE_UPDATE":      2,
		"HOOK_TYPE_POSTRECEIVE": 3,
	}
)

func (x CustomHookError_HookType) Enum() *CustomHookError_HookType {
	p := new(CustomHookError_HookType)
	*p = x
	return p
}

func (x CustomHookError_HookType) String() string {
	return protoimpl.X.EnumStringOf(x.Descriptor(), protoreflect.EnumNumber(x))
}

func (CustomHookError_HookType) Descriptor() protoreflect.EnumDescriptor {
	return file_errors_proto_enumTypes[1].Descriptor()
}

func (CustomHookError_HookType) Type() protoreflect.EnumType {
	return &file_errors_proto_enumTypes[1]
}

func (x CustomHookError_HookType) Number() protoreflect.EnumNumber {
	return protoreflect.EnumNumber(x)
}

// Deprecated: Use CustomHookError_HookType.Descriptor instead.
func (CustomHookError_HookType) EnumDescriptor() ([]byte, []int) {
	return file_errors_proto_rawDescGZIP(), []int{13, 0}
}

// ErrorType is the type of error encountered.
type PathError_ErrorType int32

const (
	// ERROR_TYPE_UNSPECIFIED is the default error type and should never be set.
	PathError_ERROR_TYPE_UNSPECIFIED PathError_ErrorType = 0
	// ERROR_TYPE_EMPTY_PATH is the error type when the provided path is empty.
	PathError_ERROR_TYPE_EMPTY_PATH PathError_ErrorType = 1
	// ERROR_TYPE_RELATIVE_PATH_ESCAPES_REPOSITORY is the error type when there are
	// traversing components found in the path and it either escapes the repository or is not
	// supported by the RPC.
	PathError_ERROR_TYPE_RELATIVE_PATH_ESCAPES_REPOSITORY PathError_ErrorType = 2
	// ERROR_TYPE_ABSOLUTE_PATH is the error type when an absolute path is provided
	// while a relative path was expected.
	PathError_ERROR_TYPE_ABSOLUTE_PATH PathError_ErrorType = 3
	// ERROR_TYPE_LONG_PATH is the error type when the path is too long.
	PathError_ERROR_TYPE_LONG_PATH PathError_ErrorType = 4
	// ERROR_TYPE_INVALID_PATH is the error type when the path given is invalid
	// and its contents can't be checked
	PathError_ERROR_TYPE_INVALID_PATH PathError_ErrorType = 5
	// ERROR_TYPE_PATH_EXISTS is the error type when an update to the path
	// would result to the same value
	PathError_ERROR_TYPE_PATH_EXISTS PathError_ErrorType = 6
)

// Enum value maps for PathError_ErrorType.
var (
	PathError_ErrorType_name = map[int32]string{
		0: "ERROR_TYPE_UNSPECIFIED",
		1: "ERROR_TYPE_EMPTY_PATH",
		2: "ERROR_TYPE_RELATIVE_PATH_ESCAPES_REPOSITORY",
		3: "ERROR_TYPE_ABSOLUTE_PATH",
		4: "ERROR_TYPE_LONG_PATH",
		5: "ERROR_TYPE_INVALID_PATH",
		6: "ERROR_TYPE_PATH_EXISTS",
	}
	PathError_ErrorType_value = map[string]int32{
		"ERROR_TYPE_UNSPECIFIED":                      0,
		"ERROR_TYPE_EMPTY_PATH":                       1,
		"ERROR_TYPE_RELATIVE_PATH_ESCAPES_REPOSITORY": 2,
		"ERROR_TYPE_ABSOLUTE_PATH":                    3,
		"ERROR_TYPE_LONG_PATH":                        4,
		"ERROR_TYPE_INVALID_PATH":                     5,
		"ERROR_TYPE_PATH_EXISTS":                      6,
	}
)

func (x PathError_ErrorType) Enum() *PathError_ErrorType {
	p := new(PathError_ErrorType)
	*p = x
	return p
}

func (x PathError_ErrorType) String() string {
	return protoimpl.X.EnumStringOf(x.Descriptor(), protoreflect.EnumNumber(x))
}

func (PathError_ErrorType) Descriptor() protoreflect.EnumDescriptor {
	return file_errors_proto_enumTypes[2].Descriptor()
}

func (PathError_ErrorType) Type() protoreflect.EnumType {
	return &file_errors_proto_enumTypes[2]
}

func (x PathError_ErrorType) Number() protoreflect.EnumNumber {
	return protoreflect.EnumNumber(x)
}

// Deprecated: Use PathError_ErrorType.Descriptor instead.
func (PathError_ErrorType) EnumDescriptor() ([]byte, []int) {
	return file_errors_proto_rawDescGZIP(), []int{14, 0}
}

// AccessCheckError is an error returned by GitLab's `/internal/allowed`
// endpoint.
type AccessCheckError struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// error_message is the error message as returned by the endpoint.
	ErrorMessage string `protobuf:"bytes,1,opt,name=error_message,json=errorMessage,proto3" json:"error_message,omitempty"`
	// protocol is the protocol used.
	Protocol string `protobuf:"bytes,2,opt,name=protocol,proto3" json:"protocol,omitempty"`
	// user_id is the user ID as which changes had been pushed.
	UserId string `protobuf:"bytes,3,opt,name=user_id,json=userId,proto3" json:"user_id,omitempty"`
	// changes is the set of changes which have failed the access check.
	Changes       []byte `protobuf:"bytes,4,opt,name=changes,proto3" json:"changes,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *AccessCheckError) Reset() {
	*x = AccessCheckError{}
	mi := &file_errors_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *AccessCheckError) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*AccessCheckError) ProtoMessage() {}

func (x *AccessCheckError) ProtoReflect() protoreflect.Message {
	mi := &file_errors_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use AccessCheckError.ProtoReflect.Descriptor instead.
func (*AccessCheckError) Descriptor() ([]byte, []int) {
	return file_errors_proto_rawDescGZIP(), []int{0}
}

func (x *AccessCheckError) GetErrorMessage() string {
	if x != nil {
		return x.ErrorMessage
	}
	return ""
}

func (x *AccessCheckError) GetProtocol() string {
	if x != nil {
		return x.Protocol
	}
	return ""
}

func (x *AccessCheckError) GetUserId() string {
	if x != nil {
		return x.UserId
	}
	return ""
}

func (x *AccessCheckError) GetChanges() []byte {
	if x != nil {
		return x.Changes
	}
	return nil
}

// IndexError is an error returned when an operation fails to due a conflict with
// the repository index.
type IndexError struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// path is the file or directory path that triggered the error.
	Path []byte `protobuf:"bytes,1,opt,name=path,proto3" json:"path,omitempty"`
	// error_type is the type of index error that occurred.
	ErrorType     IndexError_ErrorType `protobuf:"varint,2,opt,name=error_type,json=errorType,proto3,enum=gitaly.IndexError_ErrorType" json:"error_type,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *IndexError) Reset() {
	*x = IndexError{}
	mi := &file_errors_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *IndexError) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*IndexError) ProtoMessage() {}

func (x *IndexError) ProtoReflect() protoreflect.Message {
	mi := &file_errors_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use IndexError.ProtoReflect.Descriptor instead.
func (*IndexError) Descriptor() ([]byte, []int) {
	return file_errors_proto_rawDescGZIP(), []int{1}
}

func (x *IndexError) GetPath() []byte {
	if x != nil {
		return x.Path
	}
	return nil
}

func (x *IndexError) GetErrorType() IndexError_ErrorType {
	if x != nil {
		return x.ErrorType
	}
	return IndexError_ERROR_TYPE_UNSPECIFIED
}

// InvalidRefFormatError is an error returned when refs have an invalid format.
type InvalidRefFormatError struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// refs are the offending refs with invalid formats.
	Refs          [][]byte `protobuf:"bytes,2,rep,name=refs,proto3" json:"refs,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *InvalidRefFormatError) Reset() {
	*x = InvalidRefFormatError{}
	mi := &file_errors_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *InvalidRefFormatError) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*InvalidRefFormatError) ProtoMessage() {}

func (x *InvalidRefFormatError) ProtoReflect() protoreflect.Message {
	mi := &file_errors_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use InvalidRefFormatError.ProtoReflect.Descriptor instead.
func (*InvalidRefFormatError) Descriptor() ([]byte, []int) {
	return file_errors_proto_rawDescGZIP(), []int{2}
}

func (x *InvalidRefFormatError) GetRefs() [][]byte {
	if x != nil {
		return x.Refs
	}
	return nil
}

// NotAncestorError is an error returned when parent_revision is not an ancestor
// of the child_revision.
type NotAncestorError struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// parent_revision is the revision checked against ChildRevision for whether it
	// is an ancestor of ChildRevision
	ParentRevision []byte `protobuf:"bytes,1,opt,name=parent_revision,json=parentRevision,proto3" json:"parent_revision,omitempty"`
	// child_revision is the revision checked against ParentRevision for whether
	// it is a descendent of ChildRevision.
	ChildRevision []byte `protobuf:"bytes,2,opt,name=child_revision,json=childRevision,proto3" json:"child_revision,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *NotAncestorError) Reset() {
	*x = NotAncestorError{}
	mi := &file_errors_proto_msgTypes[3]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *NotAncestorError) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*NotAncestorError) ProtoMessage() {}

func (x *NotAncestorError) ProtoReflect() protoreflect.Message {
	mi := &file_errors_proto_msgTypes[3]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use NotAncestorError.ProtoReflect.Descriptor instead.
func (*NotAncestorError) Descriptor() ([]byte, []int) {
	return file_errors_proto_rawDescGZIP(), []int{3}
}

func (x *NotAncestorError) GetParentRevision() []byte {
	if x != nil {
		return x.ParentRevision
	}
	return nil
}

func (x *NotAncestorError) GetChildRevision() []byte {
	if x != nil {
		return x.ChildRevision
	}
	return nil
}

// ChangesAlreadyAppliedError is an error returned when the operation would
// have resulted in no changes because these changes have already been applied.
type ChangesAlreadyAppliedError struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ChangesAlreadyAppliedError) Reset() {
	*x = ChangesAlreadyAppliedError{}
	mi := &file_errors_proto_msgTypes[4]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ChangesAlreadyAppliedError) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ChangesAlreadyAppliedError) ProtoMessage() {}

func (x *ChangesAlreadyAppliedError) ProtoReflect() protoreflect.Message {
	mi := &file_errors_proto_msgTypes[4]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ChangesAlreadyAppliedError.ProtoReflect.Descriptor instead.
func (*ChangesAlreadyAppliedError) Descriptor() ([]byte, []int) {
	return file_errors_proto_rawDescGZIP(), []int{4}
}

// MergeConflictError is an error returned in the case when merging two commits
// fails due to a merge conflict.
type MergeConflictError struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// conflicting_files is the set of files which have been conflicting. If this
	// field is empty, then there has still been a merge conflict, but it wasn't
	// able to determine which files have been conflicting.
	ConflictingFiles [][]byte `protobuf:"bytes,1,rep,name=conflicting_files,json=conflictingFiles,proto3" json:"conflicting_files,omitempty"`
	// conflicting_commit_ids is the set of commit IDs that caused the conflict. In the general case,
	// this should be set to two commit IDs.
	ConflictingCommitIds []string `protobuf:"bytes,2,rep,name=conflicting_commit_ids,json=conflictingCommitIds,proto3" json:"conflicting_commit_ids,omitempty"`
	unknownFields        protoimpl.UnknownFields
	sizeCache            protoimpl.SizeCache
}

func (x *MergeConflictError) Reset() {
	*x = MergeConflictError{}
	mi := &file_errors_proto_msgTypes[5]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *MergeConflictError) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*MergeConflictError) ProtoMessage() {}

func (x *MergeConflictError) ProtoReflect() protoreflect.Message {
	mi := &file_errors_proto_msgTypes[5]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use MergeConflictError.ProtoReflect.Descriptor instead.
func (*MergeConflictError) Descriptor() ([]byte, []int) {
	return file_errors_proto_rawDescGZIP(), []int{5}
}

func (x *MergeConflictError) GetConflictingFiles() [][]byte {
	if x != nil {
		return x.ConflictingFiles
	}
	return nil
}

func (x *MergeConflictError) GetConflictingCommitIds() []string {
	if x != nil {
		return x.ConflictingCommitIds
	}
	return nil
}

// ReferencesLockedError is an error returned when an ref update fails because
// the references have already been locked by another process.
type ReferencesLockedError struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// refs are the references that could not be locked.
	Refs          [][]byte `protobuf:"bytes,1,rep,name=refs,proto3" json:"refs,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ReferencesLockedError) Reset() {
	*x = ReferencesLockedError{}
	mi := &file_errors_proto_msgTypes[6]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ReferencesLockedError) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ReferencesLockedError) ProtoMessage() {}

func (x *ReferencesLockedError) ProtoReflect() protoreflect.Message {
	mi := &file_errors_proto_msgTypes[6]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ReferencesLockedError.ProtoReflect.Descriptor instead.
func (*ReferencesLockedError) Descriptor() ([]byte, []int) {
	return file_errors_proto_rawDescGZIP(), []int{6}
}

func (x *ReferencesLockedError) GetRefs() [][]byte {
	if x != nil {
		return x.Refs
	}
	return nil
}

// ReferenceExistsError is an error returned when a reference that ought not to exist does exist
// already.
type ReferenceExistsError struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// reference_name is the name of the reference that exists already.
	ReferenceName []byte `protobuf:"bytes,1,opt,name=reference_name,json=referenceName,proto3" json:"reference_name,omitempty"`
	// oid is the object ID of the reference that preexists already.
	Oid           string `protobuf:"bytes,2,opt,name=oid,proto3" json:"oid,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ReferenceExistsError) Reset() {
	*x = ReferenceExistsError{}
	mi := &file_errors_proto_msgTypes[7]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ReferenceExistsError) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ReferenceExistsError) ProtoMessage() {}

func (x *ReferenceExistsError) ProtoReflect() protoreflect.Message {
	mi := &file_errors_proto_msgTypes[7]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ReferenceExistsError.ProtoReflect.Descriptor instead.
func (*ReferenceExistsError) Descriptor() ([]byte, []int) {
	return file_errors_proto_rawDescGZIP(), []int{7}
}

func (x *ReferenceExistsError) GetReferenceName() []byte {
	if x != nil {
		return x.ReferenceName
	}
	return nil
}

func (x *ReferenceExistsError) GetOid() string {
	if x != nil {
		return x.Oid
	}
	return ""
}

// ReferenceNotFoundError is an error returned when a reference that ought to exist does not exist.
type ReferenceNotFoundError struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// reference_name is the name of the reference that does not exist.
	ReferenceName []byte `protobuf:"bytes,1,opt,name=reference_name,json=referenceName,proto3" json:"reference_name,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ReferenceNotFoundError) Reset() {
	*x = ReferenceNotFoundError{}
	mi := &file_errors_proto_msgTypes[8]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ReferenceNotFoundError) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ReferenceNotFoundError) ProtoMessage() {}

func (x *ReferenceNotFoundError) ProtoReflect() protoreflect.Message {
	mi := &file_errors_proto_msgTypes[8]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ReferenceNotFoundError.ProtoReflect.Descriptor instead.
func (*ReferenceNotFoundError) Descriptor() ([]byte, []int) {
	return file_errors_proto_rawDescGZIP(), []int{8}
}

func (x *ReferenceNotFoundError) GetReferenceName() []byte {
	if x != nil {
		return x.ReferenceName
	}
	return nil
}

// ReferenceStateMismatchError is an error returned when updating a reference fails because its pointing to a different
// object ID than expected.
type ReferenceStateMismatchError struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// reference_name is the name of the reference that was failed to update.
	ReferenceName []byte `protobuf:"bytes,1,opt,name=reference_name,json=referenceName,proto3" json:"reference_name,omitempty"`
	// expected_object_id is the object ID that the reference was expected to point to but didn't.
	ExpectedObjectId []byte `protobuf:"bytes,2,opt,name=expected_object_id,json=expectedObjectId,proto3" json:"expected_object_id,omitempty"`
	// actual_object_id is the object ID that the reference actually pointed to at the time when it was tried to be
	// updated.
	ActualObjectId []byte `protobuf:"bytes,3,opt,name=actual_object_id,json=actualObjectId,proto3" json:"actual_object_id,omitempty"`
	unknownFields  protoimpl.UnknownFields
	sizeCache      protoimpl.SizeCache
}

func (x *ReferenceStateMismatchError) Reset() {
	*x = ReferenceStateMismatchError{}
	mi := &file_errors_proto_msgTypes[9]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ReferenceStateMismatchError) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ReferenceStateMismatchError) ProtoMessage() {}

func (x *ReferenceStateMismatchError) ProtoReflect() protoreflect.Message {
	mi := &file_errors_proto_msgTypes[9]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ReferenceStateMismatchError.ProtoReflect.Descriptor instead.
func (*ReferenceStateMismatchError) Descriptor() ([]byte, []int) {
	return file_errors_proto_rawDescGZIP(), []int{9}
}

func (x *ReferenceStateMismatchError) GetReferenceName() []byte {
	if x != nil {
		return x.ReferenceName
	}
	return nil
}

func (x *ReferenceStateMismatchError) GetExpectedObjectId() []byte {
	if x != nil {
		return x.ExpectedObjectId
	}
	return nil
}

func (x *ReferenceStateMismatchError) GetActualObjectId() []byte {
	if x != nil {
		return x.ActualObjectId
	}
	return nil
}

// ReferenceUpdateError is an error returned when updating a reference has
// failed.
type ReferenceUpdateError struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// reference_name is the name of the reference that failed to be updated.
	ReferenceName []byte `protobuf:"bytes,1,opt,name=reference_name,json=referenceName,proto3" json:"reference_name,omitempty"`
	// old_oid is the object ID the reference should have pointed to before the update.
	OldOid string `protobuf:"bytes,2,opt,name=old_oid,json=oldOid,proto3" json:"old_oid,omitempty"`
	// new_oid is the object ID the reference should have pointed to after the update.
	NewOid        string `protobuf:"bytes,3,opt,name=new_oid,json=newOid,proto3" json:"new_oid,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ReferenceUpdateError) Reset() {
	*x = ReferenceUpdateError{}
	mi := &file_errors_proto_msgTypes[10]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ReferenceUpdateError) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ReferenceUpdateError) ProtoMessage() {}

func (x *ReferenceUpdateError) ProtoReflect() protoreflect.Message {
	mi := &file_errors_proto_msgTypes[10]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ReferenceUpdateError.ProtoReflect.Descriptor instead.
func (*ReferenceUpdateError) Descriptor() ([]byte, []int) {
	return file_errors_proto_rawDescGZIP(), []int{10}
}

func (x *ReferenceUpdateError) GetReferenceName() []byte {
	if x != nil {
		return x.ReferenceName
	}
	return nil
}

func (x *ReferenceUpdateError) GetOldOid() string {
	if x != nil {
		return x.OldOid
	}
	return ""
}

func (x *ReferenceUpdateError) GetNewOid() string {
	if x != nil {
		return x.NewOid
	}
	return ""
}

// ResolveRevisionError is an error returned when resolving a specific revision
// has failed.
type ResolveRevisionError struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// revision is the name of the revision that was tried to be resolved.
	Revision      []byte `protobuf:"bytes,1,opt,name=revision,proto3" json:"revision,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ResolveRevisionError) Reset() {
	*x = ResolveRevisionError{}
	mi := &file_errors_proto_msgTypes[11]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ResolveRevisionError) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ResolveRevisionError) ProtoMessage() {}

func (x *ResolveRevisionError) ProtoReflect() protoreflect.Message {
	mi := &file_errors_proto_msgTypes[11]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ResolveRevisionError.ProtoReflect.Descriptor instead.
func (*ResolveRevisionError) Descriptor() ([]byte, []int) {
	return file_errors_proto_rawDescGZIP(), []int{11}
}

func (x *ResolveRevisionError) GetRevision() []byte {
	if x != nil {
		return x.Revision
	}
	return nil
}

// LimitError is an error returned when Gitaly enforces request limits.
type LimitError struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// error_message provides context into why a limit was enforced.
	ErrorMessage string `protobuf:"bytes,1,opt,name=error_message,json=errorMessage,proto3" json:"error_message,omitempty"`
	// retry_after provides the duration after which a retry is safe.
	// 0 indicates non-retryable.
	RetryAfter    *durationpb.Duration `protobuf:"bytes,2,opt,name=retry_after,json=retryAfter,proto3" json:"retry_after,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *LimitError) Reset() {
	*x = LimitError{}
	mi := &file_errors_proto_msgTypes[12]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *LimitError) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*LimitError) ProtoMessage() {}

func (x *LimitError) ProtoReflect() protoreflect.Message {
	mi := &file_errors_proto_msgTypes[12]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use LimitError.ProtoReflect.Descriptor instead.
func (*LimitError) Descriptor() ([]byte, []int) {
	return file_errors_proto_rawDescGZIP(), []int{12}
}

func (x *LimitError) GetErrorMessage() string {
	if x != nil {
		return x.ErrorMessage
	}
	return ""
}

func (x *LimitError) GetRetryAfter() *durationpb.Duration {
	if x != nil {
		return x.RetryAfter
	}
	return nil
}

// CustomHookError is an error returned when Gitaly executes a custom hook and the hook returns
// a non-zero return code.
type CustomHookError struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// stdout is the standard output of the hook that has failed, if any. Data may be truncated.
	Stdout []byte `protobuf:"bytes,1,opt,name=stdout,proto3" json:"stdout,omitempty"`
	// stderr is the standard error of the hook that has failed, if any. Data may be truncated.
	Stderr []byte `protobuf:"bytes,2,opt,name=stderr,proto3" json:"stderr,omitempty"`
	// hook_type is the type of the hook.
	HookType      CustomHookError_HookType `protobuf:"varint,3,opt,name=hook_type,json=hookType,proto3,enum=gitaly.CustomHookError_HookType" json:"hook_type,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *CustomHookError) Reset() {
	*x = CustomHookError{}
	mi := &file_errors_proto_msgTypes[13]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *CustomHookError) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*CustomHookError) ProtoMessage() {}

func (x *CustomHookError) ProtoReflect() protoreflect.Message {
	mi := &file_errors_proto_msgTypes[13]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use CustomHookError.ProtoReflect.Descriptor instead.
func (*CustomHookError) Descriptor() ([]byte, []int) {
	return file_errors_proto_rawDescGZIP(), []int{13}
}

func (x *CustomHookError) GetStdout() []byte {
	if x != nil {
		return x.Stdout
	}
	return nil
}

func (x *CustomHookError) GetStderr() []byte {
	if x != nil {
		return x.Stderr
	}
	return nil
}

func (x *CustomHookError) GetHookType() CustomHookError_HookType {
	if x != nil {
		return x.HookType
	}
	return CustomHookError_HOOK_TYPE_UNSPECIFIED
}

// PathError is an error returned when there is an issue with the path provided.
type PathError struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// path is the file or directory path that triggered the error. The path may be
	// truncated due to size limits.
	Path []byte `protobuf:"bytes,1,opt,name=path,proto3" json:"path,omitempty"`
	// error_type is the type of path error that occurred.
	ErrorType     PathError_ErrorType `protobuf:"varint,2,opt,name=error_type,json=errorType,proto3,enum=gitaly.PathError_ErrorType" json:"error_type,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *PathError) Reset() {
	*x = PathError{}
	mi := &file_errors_proto_msgTypes[14]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *PathError) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*PathError) ProtoMessage() {}

func (x *PathError) ProtoReflect() protoreflect.Message {
	mi := &file_errors_proto_msgTypes[14]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use PathError.ProtoReflect.Descriptor instead.
func (*PathError) Descriptor() ([]byte, []int) {
	return file_errors_proto_rawDescGZIP(), []int{14}
}

func (x *PathError) GetPath() []byte {
	if x != nil {
		return x.Path
	}
	return nil
}

func (x *PathError) GetErrorType() PathError_ErrorType {
	if x != nil {
		return x.ErrorType
	}
	return PathError_ERROR_TYPE_UNSPECIFIED
}

// PathNotFoundError is an error returned when a given path cannot be found.
type PathNotFoundError struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// path is the path that could not be found.
	Path          []byte `protobuf:"bytes,1,opt,name=path,proto3" json:"path,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *PathNotFoundError) Reset() {
	*x = PathNotFoundError{}
	mi := &file_errors_proto_msgTypes[15]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *PathNotFoundError) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*PathNotFoundError) ProtoMessage() {}

func (x *PathNotFoundError) ProtoReflect() protoreflect.Message {
	mi := &file_errors_proto_msgTypes[15]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use PathNotFoundError.ProtoReflect.Descriptor instead.
func (*PathNotFoundError) Descriptor() ([]byte, []int) {
	return file_errors_proto_rawDescGZIP(), []int{15}
}

func (x *PathNotFoundError) GetPath() []byte {
	if x != nil {
		return x.Path
	}
	return nil
}

// AmbiguousReferenceError is an error returned when an reference is unknown.
type AmbiguousReferenceError struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// reference is the reference passed to the git command.
	Reference     []byte `protobuf:"bytes,1,opt,name=reference,proto3" json:"reference,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *AmbiguousReferenceError) Reset() {
	*x = AmbiguousReferenceError{}
	mi := &file_errors_proto_msgTypes[16]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *AmbiguousReferenceError) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*AmbiguousReferenceError) ProtoMessage() {}

func (x *AmbiguousReferenceError) ProtoReflect() protoreflect.Message {
	mi := &file_errors_proto_msgTypes[16]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use AmbiguousReferenceError.ProtoReflect.Descriptor instead.
func (*AmbiguousReferenceError) Descriptor() ([]byte, []int) {
	return file_errors_proto_rawDescGZIP(), []int{16}
}

func (x *AmbiguousReferenceError) GetReference() []byte {
	if x != nil {
		return x.Reference
	}
	return nil
}

// BadObjectError is an error returned when git cannot find a valid object based on its id.
type BadObjectError struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// bad_oid is the bad object id passed to the git command.
	BadOid        []byte `protobuf:"bytes,1,opt,name=bad_oid,json=badOid,proto3" json:"bad_oid,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *BadObjectError) Reset() {
	*x = BadObjectError{}
	mi := &file_errors_proto_msgTypes[17]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *BadObjectError) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*BadObjectError) ProtoMessage() {}

func (x *BadObjectError) ProtoReflect() protoreflect.Message {
	mi := &file_errors_proto_msgTypes[17]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use BadObjectError.ProtoReflect.Descriptor instead.
func (*BadObjectError) Descriptor() ([]byte, []int) {
	return file_errors_proto_rawDescGZIP(), []int{17}
}

func (x *BadObjectError) GetBadOid() []byte {
	if x != nil {
		return x.BadOid
	}
	return nil
}

// InvalidRevisionRange is an error returned when the range given to the git log command is invalid.
type InvalidRevisionRange struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// range is the commit range passed to git log command.
	Range         []byte `protobuf:"bytes,1,opt,name=range,proto3" json:"range,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *InvalidRevisionRange) Reset() {
	*x = InvalidRevisionRange{}
	mi := &file_errors_proto_msgTypes[18]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *InvalidRevisionRange) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*InvalidRevisionRange) ProtoMessage() {}

func (x *InvalidRevisionRange) ProtoReflect() protoreflect.Message {
	mi := &file_errors_proto_msgTypes[18]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use InvalidRevisionRange.ProtoReflect.Descriptor instead.
func (*InvalidRevisionRange) Descriptor() ([]byte, []int) {
	return file_errors_proto_rawDescGZIP(), []int{18}
}

func (x *InvalidRevisionRange) GetRange() []byte {
	if x != nil {
		return x.Range
	}
	return nil
}

// RemoteNotFoundError is an error returned when a repository is not found at given remote URL.
type RemoteNotFoundError struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *RemoteNotFoundError) Reset() {
	*x = RemoteNotFoundError{}
	mi := &file_errors_proto_msgTypes[19]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *RemoteNotFoundError) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*RemoteNotFoundError) ProtoMessage() {}

func (x *RemoteNotFoundError) ProtoReflect() protoreflect.Message {
	mi := &file_errors_proto_msgTypes[19]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use RemoteNotFoundError.ProtoReflect.Descriptor instead.
func (*RemoteNotFoundError) Descriptor() ([]byte, []int) {
	return file_errors_proto_rawDescGZIP(), []int{19}
}

var File_errors_proto protoreflect.FileDescriptor

const file_errors_proto_rawDesc = "" +
	"\n" +
	"\ferrors.proto\x12\x06gitaly\x1a\x1egoogle/protobuf/duration.proto\"\x86\x01\n" +
	"\x10AccessCheckError\x12#\n" +
	"\rerror_message\x18\x01 \x01(\tR\ferrorMessage\x12\x1a\n" +
	"\bprotocol\x18\x02 \x01(\tR\bprotocol\x12\x17\n" +
	"\auser_id\x18\x03 \x01(\tR\x06userId\x12\x18\n" +
	"\achanges\x18\x04 \x01(\fR\achanges\"\xbf\x02\n" +
	"\n" +
	"IndexError\x12\x12\n" +
	"\x04path\x18\x01 \x01(\fR\x04path\x12;\n" +
	"\n" +
	"error_type\x18\x02 \x01(\x0e2\x1c.gitaly.IndexError.ErrorTypeR\terrorType\"\xdf\x01\n" +
	"\tErrorType\x12\x1a\n" +
	"\x16ERROR_TYPE_UNSPECIFIED\x10\x00\x12\x19\n" +
	"\x15ERROR_TYPE_EMPTY_PATH\x10\x01\x12\x1b\n" +
	"\x17ERROR_TYPE_INVALID_PATH\x10\x02\x12\x1f\n" +
	"\x1bERROR_TYPE_DIRECTORY_EXISTS\x10\x03\x12\"\n" +
	"\x1eERROR_TYPE_DIRECTORY_TRAVERSAL\x10\x04\x12\x1a\n" +
	"\x16ERROR_TYPE_FILE_EXISTS\x10\x05\x12\x1d\n" +
	"\x19ERROR_TYPE_FILE_NOT_FOUND\x10\x06\"+\n" +
	"\x15InvalidRefFormatError\x12\x12\n" +
	"\x04refs\x18\x02 \x03(\fR\x04refs\"b\n" +
	"\x10NotAncestorError\x12'\n" +
	"\x0fparent_revision\x18\x01 \x01(\fR\x0eparentRevision\x12%\n" +
	"\x0echild_revision\x18\x02 \x01(\fR\rchildRevision\"\x1c\n" +
	"\x1aChangesAlreadyAppliedError\"w\n" +
	"\x12MergeConflictError\x12+\n" +
	"\x11conflicting_files\x18\x01 \x03(\fR\x10conflictingFiles\x124\n" +
	"\x16conflicting_commit_ids\x18\x02 \x03(\tR\x14conflictingCommitIds\"+\n" +
	"\x15ReferencesLockedError\x12\x12\n" +
	"\x04refs\x18\x01 \x03(\fR\x04refs\"O\n" +
	"\x14ReferenceExistsError\x12%\n" +
	"\x0ereference_name\x18\x01 \x01(\fR\rreferenceName\x12\x10\n" +
	"\x03oid\x18\x02 \x01(\tR\x03oid\"?\n" +
	"\x16ReferenceNotFoundError\x12%\n" +
	"\x0ereference_name\x18\x01 \x01(\fR\rreferenceName\"\x9c\x01\n" +
	"\x1bReferenceStateMismatchError\x12%\n" +
	"\x0ereference_name\x18\x01 \x01(\fR\rreferenceName\x12,\n" +
	"\x12expected_object_id\x18\x02 \x01(\fR\x10expectedObjectId\x12(\n" +
	"\x10actual_object_id\x18\x03 \x01(\fR\x0eactualObjectId\"o\n" +
	"\x14ReferenceUpdateError\x12%\n" +
	"\x0ereference_name\x18\x01 \x01(\fR\rreferenceName\x12\x17\n" +
	"\aold_oid\x18\x02 \x01(\tR\x06oldOid\x12\x17\n" +
	"\anew_oid\x18\x03 \x01(\tR\x06newOid\"2\n" +
	"\x14ResolveRevisionError\x12\x1a\n" +
	"\brevision\x18\x01 \x01(\fR\brevision\"m\n" +
	"\n" +
	"LimitError\x12#\n" +
	"\rerror_message\x18\x01 \x01(\tR\ferrorMessage\x12:\n" +
	"\vretry_after\x18\x02 \x01(\v2\x19.google.protobuf.DurationR\n" +
	"retryAfter\"\xf2\x01\n" +
	"\x0fCustomHookError\x12\x16\n" +
	"\x06stdout\x18\x01 \x01(\fR\x06stdout\x12\x16\n" +
	"\x06stderr\x18\x02 \x01(\fR\x06stderr\x12=\n" +
	"\thook_type\x18\x03 \x01(\x0e2 .gitaly.CustomHookError.HookTypeR\bhookType\"p\n" +
	"\bHookType\x12\x19\n" +
	"\x15HOOK_TYPE_UNSPECIFIED\x10\x00\x12\x18\n" +
	"\x14HOOK_TYPE_PRERECEIVE\x10\x01\x12\x14\n" +
	"\x10HOOK_TYPE_UPDATE\x10\x02\x12\x19\n" +
	"\x15HOOK_TYPE_POSTRECEIVE\x10\x03\"\xc2\x02\n" +
	"\tPathError\x12\x12\n" +
	"\x04path\x18\x01 \x01(\fR\x04path\x12:\n" +
	"\n" +
	"error_type\x18\x02 \x01(\x0e2\x1b.gitaly.PathError.ErrorTypeR\terrorType\"\xe4\x01\n" +
	"\tErrorType\x12\x1a\n" +
	"\x16ERROR_TYPE_UNSPECIFIED\x10\x00\x12\x19\n" +
	"\x15ERROR_TYPE_EMPTY_PATH\x10\x01\x12/\n" +
	"+ERROR_TYPE_RELATIVE_PATH_ESCAPES_REPOSITORY\x10\x02\x12\x1c\n" +
	"\x18ERROR_TYPE_ABSOLUTE_PATH\x10\x03\x12\x18\n" +
	"\x14ERROR_TYPE_LONG_PATH\x10\x04\x12\x1b\n" +
	"\x17ERROR_TYPE_INVALID_PATH\x10\x05\x12\x1a\n" +
	"\x16ERROR_TYPE_PATH_EXISTS\x10\x06\"'\n" +
	"\x11PathNotFoundError\x12\x12\n" +
	"\x04path\x18\x01 \x01(\fR\x04path\"7\n" +
	"\x17AmbiguousReferenceError\x12\x1c\n" +
	"\treference\x18\x01 \x01(\fR\treference\")\n" +
	"\x0eBadObjectError\x12\x17\n" +
	"\abad_oid\x18\x01 \x01(\fR\x06badOid\",\n" +
	"\x14InvalidRevisionRange\x12\x14\n" +
	"\x05range\x18\x01 \x01(\fR\x05range\"\x15\n" +
	"\x13RemoteNotFoundErrorB4Z2gitlab.com/gitlab-org/gitaly/v18/proto/go/gitalypbb\x06proto3"

var (
	file_errors_proto_rawDescOnce sync.Once
	file_errors_proto_rawDescData []byte
)

func file_errors_proto_rawDescGZIP() []byte {
	file_errors_proto_rawDescOnce.Do(func() {
		file_errors_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_errors_proto_rawDesc), len(file_errors_proto_rawDesc)))
	})
	return file_errors_proto_rawDescData
}

var file_errors_proto_enumTypes = make([]protoimpl.EnumInfo, 3)
var file_errors_proto_msgTypes = make([]protoimpl.MessageInfo, 20)
var file_errors_proto_goTypes = []any{
	(IndexError_ErrorType)(0),           // 0: gitaly.IndexError.ErrorType
	(CustomHookError_HookType)(0),       // 1: gitaly.CustomHookError.HookType
	(PathError_ErrorType)(0),            // 2: gitaly.PathError.ErrorType
	(*AccessCheckError)(nil),            // 3: gitaly.AccessCheckError
	(*IndexError)(nil),                  // 4: gitaly.IndexError
	(*InvalidRefFormatError)(nil),       // 5: gitaly.InvalidRefFormatError
	(*NotAncestorError)(nil),            // 6: gitaly.NotAncestorError
	(*ChangesAlreadyAppliedError)(nil),  // 7: gitaly.ChangesAlreadyAppliedError
	(*MergeConflictError)(nil),          // 8: gitaly.MergeConflictError
	(*ReferencesLockedError)(nil),       // 9: gitaly.ReferencesLockedError
	(*ReferenceExistsError)(nil),        // 10: gitaly.ReferenceExistsError
	(*ReferenceNotFoundError)(nil),      // 11: gitaly.ReferenceNotFoundError
	(*ReferenceStateMismatchError)(nil), // 12: gitaly.ReferenceStateMismatchError
	(*ReferenceUpdateError)(nil),        // 13: gitaly.ReferenceUpdateError
	(*ResolveRevisionError)(nil),        // 14: gitaly.ResolveRevisionError
	(*LimitError)(nil),                  // 15: gitaly.LimitError
	(*CustomHookError)(nil),             // 16: gitaly.CustomHookError
	(*PathError)(nil),                   // 17: gitaly.PathError
	(*PathNotFoundError)(nil),           // 18: gitaly.PathNotFoundError
	(*AmbiguousReferenceError)(nil),     // 19: gitaly.AmbiguousReferenceError
	(*BadObjectError)(nil),              // 20: gitaly.BadObjectError
	(*InvalidRevisionRange)(nil),        // 21: gitaly.InvalidRevisionRange
	(*RemoteNotFoundError)(nil),         // 22: gitaly.RemoteNotFoundError
	(*durationpb.Duration)(nil),         // 23: google.protobuf.Duration
}
var file_errors_proto_depIdxs = []int32{
	0,  // 0: gitaly.IndexError.error_type:type_name -> gitaly.IndexError.ErrorType
	23, // 1: gitaly.LimitError.retry_after:type_name -> google.protobuf.Duration
	1,  // 2: gitaly.CustomHookError.hook_type:type_name -> gitaly.CustomHookError.HookType
	2,  // 3: gitaly.PathError.error_type:type_name -> gitaly.PathError.ErrorType
	4,  // [4:4] is the sub-list for method output_type
	4,  // [4:4] is the sub-list for method input_type
	4,  // [4:4] is the sub-list for extension type_name
	4,  // [4:4] is the sub-list for extension extendee
	0,  // [0:4] is the sub-list for field type_name
}

func init() { file_errors_proto_init() }
func file_errors_proto_init() {
	if File_errors_proto != nil {
		return
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_errors_proto_rawDesc), len(file_errors_proto_rawDesc)),
			NumEnums:      3,
			NumMessages:   20,
			NumExtensions: 0,
			NumServices:   0,
		},
		GoTypes:           file_errors_proto_goTypes,
		DependencyIndexes: file_errors_proto_depIdxs,
		EnumInfos:         file_errors_proto_enumTypes,
		MessageInfos:      file_errors_proto_msgTypes,
	}.Build()
	File_errors_proto = out.File
	file_errors_proto_goTypes = nil
	file_errors_proto_depIdxs = nil
}
