// Code generated by protoc-gen-go-grpc. DO NOT EDIT.
// versions:
// - protoc-gen-go-grpc v1.6.1
// - protoc             v6.30.2
// source: testproto/valid.proto

package testproto

import (
	context "context"
	grpc "google.golang.org/grpc"
	codes "google.golang.org/grpc/codes"
	status "google.golang.org/grpc/status"
)

// This is a compile-time assertion to ensure that this generated file
// is compatible with the grpc package it is being compiled against.
// Requires gRPC-Go v1.64.0 or later.
const _ = grpc.SupportPackageIsVersion9

const (
	InterceptedService_TestMethod_FullMethodName = "/testproto.InterceptedService/TestMethod"
)

// InterceptedServiceClient is the client API for InterceptedService service.
//
// For semantics around ctx use and closing/ending streaming RPCs, please refer to https://pkg.go.dev/google.golang.org/grpc/?tab=doc#ClientConn.NewStream.
type InterceptedServiceClient interface {
	TestMethod(ctx context.Context, in *ValidRequest, opts ...grpc.CallOption) (*ValidResponse, error)
}

type interceptedServiceClient struct {
	cc grpc.ClientConnInterface
}

func NewInterceptedServiceClient(cc grpc.ClientConnInterface) InterceptedServiceClient {
	return &interceptedServiceClient{cc}
}

func (c *interceptedServiceClient) TestMethod(ctx context.Context, in *ValidRequest, opts ...grpc.CallOption) (*ValidResponse, error) {
	cOpts := append([]grpc.CallOption{grpc.StaticMethod()}, opts...)
	out := new(ValidResponse)
	err := c.cc.Invoke(ctx, InterceptedService_TestMethod_FullMethodName, in, out, cOpts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

// InterceptedServiceServer is the server API for InterceptedService service.
// All implementations must embed UnimplementedInterceptedServiceServer
// for forward compatibility.
type InterceptedServiceServer interface {
	TestMethod(context.Context, *ValidRequest) (*ValidResponse, error)
	mustEmbedUnimplementedInterceptedServiceServer()
}

// UnimplementedInterceptedServiceServer must be embedded to have
// forward compatible implementations.
//
// NOTE: this should be embedded by value instead of pointer to avoid a nil
// pointer dereference when methods are called.
type UnimplementedInterceptedServiceServer struct{}

func (UnimplementedInterceptedServiceServer) TestMethod(context.Context, *ValidRequest) (*ValidResponse, error) {
	return nil, status.Error(codes.Unimplemented, "method TestMethod not implemented")
}
func (UnimplementedInterceptedServiceServer) mustEmbedUnimplementedInterceptedServiceServer() {}
func (UnimplementedInterceptedServiceServer) testEmbeddedByValue()                            {}

// UnsafeInterceptedServiceServer may be embedded to opt out of forward compatibility for this service.
// Use of this interface is not recommended, as added methods to InterceptedServiceServer will
// result in compilation errors.
type UnsafeInterceptedServiceServer interface {
	mustEmbedUnimplementedInterceptedServiceServer()
}

func RegisterInterceptedServiceServer(s grpc.ServiceRegistrar, srv InterceptedServiceServer) {
	// If the following call panics, it indicates UnimplementedInterceptedServiceServer was
	// embedded by pointer and is nil.  This will cause panics if an
	// unimplemented method is ever invoked, so we test this at initialization
	// time to prevent it from happening at runtime later due to I/O.
	if t, ok := srv.(interface{ testEmbeddedByValue() }); ok {
		t.testEmbeddedByValue()
	}
	s.RegisterService(&InterceptedService_ServiceDesc, srv)
}

func _InterceptedService_TestMethod_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(ValidRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(InterceptedServiceServer).TestMethod(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: InterceptedService_TestMethod_FullMethodName,
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(InterceptedServiceServer).TestMethod(ctx, req.(*ValidRequest))
	}
	return interceptor(ctx, in, info, handler)
}

// InterceptedService_ServiceDesc is the grpc.ServiceDesc for InterceptedService service.
// It's only intended for direct use with grpc.RegisterService,
// and not to be introspected or modified (even as a copy)
var InterceptedService_ServiceDesc = grpc.ServiceDesc{
	ServiceName: "testproto.InterceptedService",
	HandlerType: (*InterceptedServiceServer)(nil),
	Methods: []grpc.MethodDesc{
		{
			MethodName: "TestMethod",
			Handler:    _InterceptedService_TestMethod_Handler,
		},
	},
	Streams:  []grpc.StreamDesc{},
	Metadata: "testproto/valid.proto",
}

const (
	ValidService_TestMethod_FullMethodName                             = "/testproto.ValidService/TestMethod"
	ValidService_TestMethod2_FullMethodName                            = "/testproto.ValidService/TestMethod2"
	ValidService_TestMethod3_FullMethodName                            = "/testproto.ValidService/TestMethod3"
	ValidService_TestMethod5_FullMethodName                            = "/testproto.ValidService/TestMethod5"
	ValidService_TestMethod6_FullMethodName                            = "/testproto.ValidService/TestMethod6"
	ValidService_TestMethod7_FullMethodName                            = "/testproto.ValidService/TestMethod7"
	ValidService_TestMethod8_FullMethodName                            = "/testproto.ValidService/TestMethod8"
	ValidService_TestMethod9_FullMethodName                            = "/testproto.ValidService/TestMethod9"
	ValidService_TestMaintenance_FullMethodName                        = "/testproto.ValidService/TestMaintenance"
	ValidService_TestMaintenanceWithExplicitScope_FullMethodName       = "/testproto.ValidService/TestMaintenanceWithExplicitScope"
	ValidService_TestMaintenanceWithNestedRequest_FullMethodName       = "/testproto.ValidService/TestMaintenanceWithNestedRequest"
	ValidService_TestMaintenanceWithNestedSharedRequest_FullMethodName = "/testproto.ValidService/TestMaintenanceWithNestedSharedRequest"
	ValidService_TestMutatorWithInnerNestedRequest_FullMethodName      = "/testproto.ValidService/TestMutatorWithInnerNestedRequest"
)

// ValidServiceClient is the client API for ValidService service.
//
// For semantics around ctx use and closing/ending streaming RPCs, please refer to https://pkg.go.dev/google.golang.org/grpc/?tab=doc#ClientConn.NewStream.
type ValidServiceClient interface {
	TestMethod(ctx context.Context, in *ValidRequest, opts ...grpc.CallOption) (*ValidResponse, error)
	TestMethod2(ctx context.Context, in *ValidRequest, opts ...grpc.CallOption) (*ValidResponse, error)
	TestMethod3(ctx context.Context, in *ValidRequest, opts ...grpc.CallOption) (*ValidResponse, error)
	TestMethod5(ctx context.Context, in *ValidNestedRequest, opts ...grpc.CallOption) (*ValidResponse, error)
	TestMethod6(ctx context.Context, in *ValidNestedSharedRequest, opts ...grpc.CallOption) (*ValidResponse, error)
	TestMethod7(ctx context.Context, in *ValidInnerNestedRequest, opts ...grpc.CallOption) (*ValidResponse, error)
	TestMethod8(ctx context.Context, in *ValidStorageRequest, opts ...grpc.CallOption) (*ValidResponse, error)
	TestMethod9(ctx context.Context, in *ValidStorageNestedRequest, opts ...grpc.CallOption) (*ValidResponse, error)
	TestMaintenance(ctx context.Context, in *ValidRequest, opts ...grpc.CallOption) (*ValidResponse, error)
	TestMaintenanceWithExplicitScope(ctx context.Context, in *ValidRequest, opts ...grpc.CallOption) (*ValidResponse, error)
	TestMaintenanceWithNestedRequest(ctx context.Context, in *ValidNestedRequest, opts ...grpc.CallOption) (*ValidResponse, error)
	TestMaintenanceWithNestedSharedRequest(ctx context.Context, in *ValidNestedSharedRequest, opts ...grpc.CallOption) (*ValidResponse, error)
	TestMutatorWithInnerNestedRequest(ctx context.Context, in *ValidInnerNestedRequest, opts ...grpc.CallOption) (*ValidResponse, error)
}

type validServiceClient struct {
	cc grpc.ClientConnInterface
}

func NewValidServiceClient(cc grpc.ClientConnInterface) ValidServiceClient {
	return &validServiceClient{cc}
}

func (c *validServiceClient) TestMethod(ctx context.Context, in *ValidRequest, opts ...grpc.CallOption) (*ValidResponse, error) {
	cOpts := append([]grpc.CallOption{grpc.StaticMethod()}, opts...)
	out := new(ValidResponse)
	err := c.cc.Invoke(ctx, ValidService_TestMethod_FullMethodName, in, out, cOpts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

func (c *validServiceClient) TestMethod2(ctx context.Context, in *ValidRequest, opts ...grpc.CallOption) (*ValidResponse, error) {
	cOpts := append([]grpc.CallOption{grpc.StaticMethod()}, opts...)
	out := new(ValidResponse)
	err := c.cc.Invoke(ctx, ValidService_TestMethod2_FullMethodName, in, out, cOpts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

func (c *validServiceClient) TestMethod3(ctx context.Context, in *ValidRequest, opts ...grpc.CallOption) (*ValidResponse, error) {
	cOpts := append([]grpc.CallOption{grpc.StaticMethod()}, opts...)
	out := new(ValidResponse)
	err := c.cc.Invoke(ctx, ValidService_TestMethod3_FullMethodName, in, out, cOpts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

func (c *validServiceClient) TestMethod5(ctx context.Context, in *ValidNestedRequest, opts ...grpc.CallOption) (*ValidResponse, error) {
	cOpts := append([]grpc.CallOption{grpc.StaticMethod()}, opts...)
	out := new(ValidResponse)
	err := c.cc.Invoke(ctx, ValidService_TestMethod5_FullMethodName, in, out, cOpts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

func (c *validServiceClient) TestMethod6(ctx context.Context, in *ValidNestedSharedRequest, opts ...grpc.CallOption) (*ValidResponse, error) {
	cOpts := append([]grpc.CallOption{grpc.StaticMethod()}, opts...)
	out := new(ValidResponse)
	err := c.cc.Invoke(ctx, ValidService_TestMethod6_FullMethodName, in, out, cOpts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

func (c *validServiceClient) TestMethod7(ctx context.Context, in *ValidInnerNestedRequest, opts ...grpc.CallOption) (*ValidResponse, error) {
	cOpts := append([]grpc.CallOption{grpc.StaticMethod()}, opts...)
	out := new(ValidResponse)
	err := c.cc.Invoke(ctx, ValidService_TestMethod7_FullMethodName, in, out, cOpts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

func (c *validServiceClient) TestMethod8(ctx context.Context, in *ValidStorageRequest, opts ...grpc.CallOption) (*ValidResponse, error) {
	cOpts := append([]grpc.CallOption{grpc.StaticMethod()}, opts...)
	out := new(ValidResponse)
	err := c.cc.Invoke(ctx, ValidService_TestMethod8_FullMethodName, in, out, cOpts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

func (c *validServiceClient) TestMethod9(ctx context.Context, in *ValidStorageNestedRequest, opts ...grpc.CallOption) (*ValidResponse, error) {
	cOpts := append([]grpc.CallOption{grpc.StaticMethod()}, opts...)
	out := new(ValidResponse)
	err := c.cc.Invoke(ctx, ValidService_TestMethod9_FullMethodName, in, out, cOpts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

func (c *validServiceClient) TestMaintenance(ctx context.Context, in *ValidRequest, opts ...grpc.CallOption) (*ValidResponse, error) {
	cOpts := append([]grpc.CallOption{grpc.StaticMethod()}, opts...)
	out := new(ValidResponse)
	err := c.cc.Invoke(ctx, ValidService_TestMaintenance_FullMethodName, in, out, cOpts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

func (c *validServiceClient) TestMaintenanceWithExplicitScope(ctx context.Context, in *ValidRequest, opts ...grpc.CallOption) (*ValidResponse, error) {
	cOpts := append([]grpc.CallOption{grpc.StaticMethod()}, opts...)
	out := new(ValidResponse)
	err := c.cc.Invoke(ctx, ValidService_TestMaintenanceWithExplicitScope_FullMethodName, in, out, cOpts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

func (c *validServiceClient) TestMaintenanceWithNestedRequest(ctx context.Context, in *ValidNestedRequest, opts ...grpc.CallOption) (*ValidResponse, error) {
	cOpts := append([]grpc.CallOption{grpc.StaticMethod()}, opts...)
	out := new(ValidResponse)
	err := c.cc.Invoke(ctx, ValidService_TestMaintenanceWithNestedRequest_FullMethodName, in, out, cOpts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

func (c *validServiceClient) TestMaintenanceWithNestedSharedRequest(ctx context.Context, in *ValidNestedSharedRequest, opts ...grpc.CallOption) (*ValidResponse, error) {
	cOpts := append([]grpc.CallOption{grpc.StaticMethod()}, opts...)
	out := new(ValidResponse)
	err := c.cc.Invoke(ctx, ValidService_TestMaintenanceWithNestedSharedRequest_FullMethodName, in, out, cOpts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

func (c *validServiceClient) TestMutatorWithInnerNestedRequest(ctx context.Context, in *ValidInnerNestedRequest, opts ...grpc.CallOption) (*ValidResponse, error) {
	cOpts := append([]grpc.CallOption{grpc.StaticMethod()}, opts...)
	out := new(ValidResponse)
	err := c.cc.Invoke(ctx, ValidService_TestMutatorWithInnerNestedRequest_FullMethodName, in, out, cOpts...)
	if err != nil {
		return nil, err
	}
	return out, nil
}

// ValidServiceServer is the server API for ValidService service.
// All implementations must embed UnimplementedValidServiceServer
// for forward compatibility.
type ValidServiceServer interface {
	TestMethod(context.Context, *ValidRequest) (*ValidResponse, error)
	TestMethod2(context.Context, *ValidRequest) (*ValidResponse, error)
	TestMethod3(context.Context, *ValidRequest) (*ValidResponse, error)
	TestMethod5(context.Context, *ValidNestedRequest) (*ValidResponse, error)
	TestMethod6(context.Context, *ValidNestedSharedRequest) (*ValidResponse, error)
	TestMethod7(context.Context, *ValidInnerNestedRequest) (*ValidResponse, error)
	TestMethod8(context.Context, *ValidStorageRequest) (*ValidResponse, error)
	TestMethod9(context.Context, *ValidStorageNestedRequest) (*ValidResponse, error)
	TestMaintenance(context.Context, *ValidRequest) (*ValidResponse, error)
	TestMaintenanceWithExplicitScope(context.Context, *ValidRequest) (*ValidResponse, error)
	TestMaintenanceWithNestedRequest(context.Context, *ValidNestedRequest) (*ValidResponse, error)
	TestMaintenanceWithNestedSharedRequest(context.Context, *ValidNestedSharedRequest) (*ValidResponse, error)
	TestMutatorWithInnerNestedRequest(context.Context, *ValidInnerNestedRequest) (*ValidResponse, error)
	mustEmbedUnimplementedValidServiceServer()
}

// UnimplementedValidServiceServer must be embedded to have
// forward compatible implementations.
//
// NOTE: this should be embedded by value instead of pointer to avoid a nil
// pointer dereference when methods are called.
type UnimplementedValidServiceServer struct{}

func (UnimplementedValidServiceServer) TestMethod(context.Context, *ValidRequest) (*ValidResponse, error) {
	return nil, status.Error(codes.Unimplemented, "method TestMethod not implemented")
}
func (UnimplementedValidServiceServer) TestMethod2(context.Context, *ValidRequest) (*ValidResponse, error) {
	return nil, status.Error(codes.Unimplemented, "method TestMethod2 not implemented")
}
func (UnimplementedValidServiceServer) TestMethod3(context.Context, *ValidRequest) (*ValidResponse, error) {
	return nil, status.Error(codes.Unimplemented, "method TestMethod3 not implemented")
}
func (UnimplementedValidServiceServer) TestMethod5(context.Context, *ValidNestedRequest) (*ValidResponse, error) {
	return nil, status.Error(codes.Unimplemented, "method TestMethod5 not implemented")
}
func (UnimplementedValidServiceServer) TestMethod6(context.Context, *ValidNestedSharedRequest) (*ValidResponse, error) {
	return nil, status.Error(codes.Unimplemented, "method TestMethod6 not implemented")
}
func (UnimplementedValidServiceServer) TestMethod7(context.Context, *ValidInnerNestedRequest) (*ValidResponse, error) {
	return nil, status.Error(codes.Unimplemented, "method TestMethod7 not implemented")
}
func (UnimplementedValidServiceServer) TestMethod8(context.Context, *ValidStorageRequest) (*ValidResponse, error) {
	return nil, status.Error(codes.Unimplemented, "method TestMethod8 not implemented")
}
func (UnimplementedValidServiceServer) TestMethod9(context.Context, *ValidStorageNestedRequest) (*ValidResponse, error) {
	return nil, status.Error(codes.Unimplemented, "method TestMethod9 not implemented")
}
func (UnimplementedValidServiceServer) TestMaintenance(context.Context, *ValidRequest) (*ValidResponse, error) {
	return nil, status.Error(codes.Unimplemented, "method TestMaintenance not implemented")
}
func (UnimplementedValidServiceServer) TestMaintenanceWithExplicitScope(context.Context, *ValidRequest) (*ValidResponse, error) {
	return nil, status.Error(codes.Unimplemented, "method TestMaintenanceWithExplicitScope not implemented")
}
func (UnimplementedValidServiceServer) TestMaintenanceWithNestedRequest(context.Context, *ValidNestedRequest) (*ValidResponse, error) {
	return nil, status.Error(codes.Unimplemented, "method TestMaintenanceWithNestedRequest not implemented")
}
func (UnimplementedValidServiceServer) TestMaintenanceWithNestedSharedRequest(context.Context, *ValidNestedSharedRequest) (*ValidResponse, error) {
	return nil, status.Error(codes.Unimplemented, "method TestMaintenanceWithNestedSharedRequest not implemented")
}
func (UnimplementedValidServiceServer) TestMutatorWithInnerNestedRequest(context.Context, *ValidInnerNestedRequest) (*ValidResponse, error) {
	return nil, status.Error(codes.Unimplemented, "method TestMutatorWithInnerNestedRequest not implemented")
}
func (UnimplementedValidServiceServer) mustEmbedUnimplementedValidServiceServer() {}
func (UnimplementedValidServiceServer) testEmbeddedByValue()                      {}

// UnsafeValidServiceServer may be embedded to opt out of forward compatibility for this service.
// Use of this interface is not recommended, as added methods to ValidServiceServer will
// result in compilation errors.
type UnsafeValidServiceServer interface {
	mustEmbedUnimplementedValidServiceServer()
}

func RegisterValidServiceServer(s grpc.ServiceRegistrar, srv ValidServiceServer) {
	// If the following call panics, it indicates UnimplementedValidServiceServer was
	// embedded by pointer and is nil.  This will cause panics if an
	// unimplemented method is ever invoked, so we test this at initialization
	// time to prevent it from happening at runtime later due to I/O.
	if t, ok := srv.(interface{ testEmbeddedByValue() }); ok {
		t.testEmbeddedByValue()
	}
	s.RegisterService(&ValidService_ServiceDesc, srv)
}

func _ValidService_TestMethod_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(ValidRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(ValidServiceServer).TestMethod(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: ValidService_TestMethod_FullMethodName,
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(ValidServiceServer).TestMethod(ctx, req.(*ValidRequest))
	}
	return interceptor(ctx, in, info, handler)
}

func _ValidService_TestMethod2_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(ValidRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(ValidServiceServer).TestMethod2(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: ValidService_TestMethod2_FullMethodName,
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(ValidServiceServer).TestMethod2(ctx, req.(*ValidRequest))
	}
	return interceptor(ctx, in, info, handler)
}

func _ValidService_TestMethod3_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(ValidRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(ValidServiceServer).TestMethod3(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: ValidService_TestMethod3_FullMethodName,
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(ValidServiceServer).TestMethod3(ctx, req.(*ValidRequest))
	}
	return interceptor(ctx, in, info, handler)
}

func _ValidService_TestMethod5_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(ValidNestedRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(ValidServiceServer).TestMethod5(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: ValidService_TestMethod5_FullMethodName,
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(ValidServiceServer).TestMethod5(ctx, req.(*ValidNestedRequest))
	}
	return interceptor(ctx, in, info, handler)
}

func _ValidService_TestMethod6_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(ValidNestedSharedRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(ValidServiceServer).TestMethod6(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: ValidService_TestMethod6_FullMethodName,
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(ValidServiceServer).TestMethod6(ctx, req.(*ValidNestedSharedRequest))
	}
	return interceptor(ctx, in, info, handler)
}

func _ValidService_TestMethod7_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(ValidInnerNestedRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(ValidServiceServer).TestMethod7(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: ValidService_TestMethod7_FullMethodName,
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(ValidServiceServer).TestMethod7(ctx, req.(*ValidInnerNestedRequest))
	}
	return interceptor(ctx, in, info, handler)
}

func _ValidService_TestMethod8_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(ValidStorageRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(ValidServiceServer).TestMethod8(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: ValidService_TestMethod8_FullMethodName,
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(ValidServiceServer).TestMethod8(ctx, req.(*ValidStorageRequest))
	}
	return interceptor(ctx, in, info, handler)
}

func _ValidService_TestMethod9_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(ValidStorageNestedRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(ValidServiceServer).TestMethod9(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: ValidService_TestMethod9_FullMethodName,
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(ValidServiceServer).TestMethod9(ctx, req.(*ValidStorageNestedRequest))
	}
	return interceptor(ctx, in, info, handler)
}

func _ValidService_TestMaintenance_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(ValidRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(ValidServiceServer).TestMaintenance(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: ValidService_TestMaintenance_FullMethodName,
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(ValidServiceServer).TestMaintenance(ctx, req.(*ValidRequest))
	}
	return interceptor(ctx, in, info, handler)
}

func _ValidService_TestMaintenanceWithExplicitScope_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(ValidRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(ValidServiceServer).TestMaintenanceWithExplicitScope(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: ValidService_TestMaintenanceWithExplicitScope_FullMethodName,
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(ValidServiceServer).TestMaintenanceWithExplicitScope(ctx, req.(*ValidRequest))
	}
	return interceptor(ctx, in, info, handler)
}

func _ValidService_TestMaintenanceWithNestedRequest_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(ValidNestedRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(ValidServiceServer).TestMaintenanceWithNestedRequest(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: ValidService_TestMaintenanceWithNestedRequest_FullMethodName,
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(ValidServiceServer).TestMaintenanceWithNestedRequest(ctx, req.(*ValidNestedRequest))
	}
	return interceptor(ctx, in, info, handler)
}

func _ValidService_TestMaintenanceWithNestedSharedRequest_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(ValidNestedSharedRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(ValidServiceServer).TestMaintenanceWithNestedSharedRequest(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: ValidService_TestMaintenanceWithNestedSharedRequest_FullMethodName,
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(ValidServiceServer).TestMaintenanceWithNestedSharedRequest(ctx, req.(*ValidNestedSharedRequest))
	}
	return interceptor(ctx, in, info, handler)
}

func _ValidService_TestMutatorWithInnerNestedRequest_Handler(srv interface{}, ctx context.Context, dec func(interface{}) error, interceptor grpc.UnaryServerInterceptor) (interface{}, error) {
	in := new(ValidInnerNestedRequest)
	if err := dec(in); err != nil {
		return nil, err
	}
	if interceptor == nil {
		return srv.(ValidServiceServer).TestMutatorWithInnerNestedRequest(ctx, in)
	}
	info := &grpc.UnaryServerInfo{
		Server:     srv,
		FullMethod: ValidService_TestMutatorWithInnerNestedRequest_FullMethodName,
	}
	handler := func(ctx context.Context, req interface{}) (interface{}, error) {
		return srv.(ValidServiceServer).TestMutatorWithInnerNestedRequest(ctx, req.(*ValidInnerNestedRequest))
	}
	return interceptor(ctx, in, info, handler)
}

// ValidService_ServiceDesc is the grpc.ServiceDesc for ValidService service.
// It's only intended for direct use with grpc.RegisterService,
// and not to be introspected or modified (even as a copy)
var ValidService_ServiceDesc = grpc.ServiceDesc{
	ServiceName: "testproto.ValidService",
	HandlerType: (*ValidServiceServer)(nil),
	Methods: []grpc.MethodDesc{
		{
			MethodName: "TestMethod",
			Handler:    _ValidService_TestMethod_Handler,
		},
		{
			MethodName: "TestMethod2",
			Handler:    _ValidService_TestMethod2_Handler,
		},
		{
			MethodName: "TestMethod3",
			Handler:    _ValidService_TestMethod3_Handler,
		},
		{
			MethodName: "TestMethod5",
			Handler:    _ValidService_TestMethod5_Handler,
		},
		{
			MethodName: "TestMethod6",
			Handler:    _ValidService_TestMethod6_Handler,
		},
		{
			MethodName: "TestMethod7",
			Handler:    _ValidService_TestMethod7_Handler,
		},
		{
			MethodName: "TestMethod8",
			Handler:    _ValidService_TestMethod8_Handler,
		},
		{
			MethodName: "TestMethod9",
			Handler:    _ValidService_TestMethod9_Handler,
		},
		{
			MethodName: "TestMaintenance",
			Handler:    _ValidService_TestMaintenance_Handler,
		},
		{
			MethodName: "TestMaintenanceWithExplicitScope",
			Handler:    _ValidService_TestMaintenanceWithExplicitScope_Handler,
		},
		{
			MethodName: "TestMaintenanceWithNestedRequest",
			Handler:    _ValidService_TestMaintenanceWithNestedRequest_Handler,
		},
		{
			MethodName: "TestMaintenanceWithNestedSharedRequest",
			Handler:    _ValidService_TestMaintenanceWithNestedSharedRequest_Handler,
		},
		{
			MethodName: "TestMutatorWithInnerNestedRequest",
			Handler:    _ValidService_TestMutatorWithInnerNestedRequest_Handler,
		},
	},
	Streams:  []grpc.StreamDesc{},
	Metadata: "testproto/valid.proto",
}
