// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v6.30.2
// source: testproto/valid.proto

package testproto

import (
	gitalypb "gitlab.com/gitlab-org/gitaly/v18/proto/go/gitalypb"
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

type ValidRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Destination   *gitalypb.Repository   `protobuf:"bytes,1,opt,name=destination,proto3" json:"destination,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ValidRequest) Reset() {
	*x = ValidRequest{}
	mi := &file_testproto_valid_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ValidRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ValidRequest) ProtoMessage() {}

func (x *ValidRequest) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_valid_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ValidRequest.ProtoReflect.Descriptor instead.
func (*ValidRequest) Descriptor() ([]byte, []int) {
	return file_testproto_valid_proto_rawDescGZIP(), []int{0}
}

func (x *ValidRequest) GetDestination() *gitalypb.Repository {
	if x != nil {
		return x.Destination
	}
	return nil
}

type ValidRequestWithoutRepo struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ValidRequestWithoutRepo) Reset() {
	*x = ValidRequestWithoutRepo{}
	mi := &file_testproto_valid_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ValidRequestWithoutRepo) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ValidRequestWithoutRepo) ProtoMessage() {}

func (x *ValidRequestWithoutRepo) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_valid_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ValidRequestWithoutRepo.ProtoReflect.Descriptor instead.
func (*ValidRequestWithoutRepo) Descriptor() ([]byte, []int) {
	return file_testproto_valid_proto_rawDescGZIP(), []int{1}
}

type ValidStorageRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	StorageName   string                 `protobuf:"bytes,1,opt,name=storage_name,json=storageName,proto3" json:"storage_name,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ValidStorageRequest) Reset() {
	*x = ValidStorageRequest{}
	mi := &file_testproto_valid_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ValidStorageRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ValidStorageRequest) ProtoMessage() {}

func (x *ValidStorageRequest) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_valid_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ValidStorageRequest.ProtoReflect.Descriptor instead.
func (*ValidStorageRequest) Descriptor() ([]byte, []int) {
	return file_testproto_valid_proto_rawDescGZIP(), []int{2}
}

func (x *ValidStorageRequest) GetStorageName() string {
	if x != nil {
		return x.StorageName
	}
	return ""
}

type ValidResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ValidResponse) Reset() {
	*x = ValidResponse{}
	mi := &file_testproto_valid_proto_msgTypes[3]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ValidResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ValidResponse) ProtoMessage() {}

func (x *ValidResponse) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_valid_proto_msgTypes[3]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ValidResponse.ProtoReflect.Descriptor instead.
func (*ValidResponse) Descriptor() ([]byte, []int) {
	return file_testproto_valid_proto_rawDescGZIP(), []int{3}
}

type ValidNestedRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	InnerMessage  *ValidRequest          `protobuf:"bytes,1,opt,name=inner_message,json=innerMessage,proto3" json:"inner_message,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ValidNestedRequest) Reset() {
	*x = ValidNestedRequest{}
	mi := &file_testproto_valid_proto_msgTypes[4]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ValidNestedRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ValidNestedRequest) ProtoMessage() {}

func (x *ValidNestedRequest) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_valid_proto_msgTypes[4]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ValidNestedRequest.ProtoReflect.Descriptor instead.
func (*ValidNestedRequest) Descriptor() ([]byte, []int) {
	return file_testproto_valid_proto_rawDescGZIP(), []int{4}
}

func (x *ValidNestedRequest) GetInnerMessage() *ValidRequest {
	if x != nil {
		return x.InnerMessage
	}
	return nil
}

type ValidStorageNestedRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	InnerMessage  *ValidStorageRequest   `protobuf:"bytes,1,opt,name=inner_message,json=innerMessage,proto3" json:"inner_message,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ValidStorageNestedRequest) Reset() {
	*x = ValidStorageNestedRequest{}
	mi := &file_testproto_valid_proto_msgTypes[5]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ValidStorageNestedRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ValidStorageNestedRequest) ProtoMessage() {}

func (x *ValidStorageNestedRequest) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_valid_proto_msgTypes[5]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ValidStorageNestedRequest.ProtoReflect.Descriptor instead.
func (*ValidStorageNestedRequest) Descriptor() ([]byte, []int) {
	return file_testproto_valid_proto_rawDescGZIP(), []int{5}
}

func (x *ValidStorageNestedRequest) GetInnerMessage() *ValidStorageRequest {
	if x != nil {
		return x.InnerMessage
	}
	return nil
}

type ValidNestedSharedRequest struct {
	state            protoimpl.MessageState `protogen:"open.v1"`
	NestedTargetRepo *gitalypb.ObjectPool   `protobuf:"bytes,1,opt,name=nested_target_repo,json=nestedTargetRepo,proto3" json:"nested_target_repo,omitempty"`
	unknownFields    protoimpl.UnknownFields
	sizeCache        protoimpl.SizeCache
}

func (x *ValidNestedSharedRequest) Reset() {
	*x = ValidNestedSharedRequest{}
	mi := &file_testproto_valid_proto_msgTypes[6]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ValidNestedSharedRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ValidNestedSharedRequest) ProtoMessage() {}

func (x *ValidNestedSharedRequest) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_valid_proto_msgTypes[6]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ValidNestedSharedRequest.ProtoReflect.Descriptor instead.
func (*ValidNestedSharedRequest) Descriptor() ([]byte, []int) {
	return file_testproto_valid_proto_rawDescGZIP(), []int{6}
}

func (x *ValidNestedSharedRequest) GetNestedTargetRepo() *gitalypb.ObjectPool {
	if x != nil {
		return x.NestedTargetRepo
	}
	return nil
}

type ValidInnerNestedRequest struct {
	state         protoimpl.MessageState          `protogen:"open.v1"`
	Header        *ValidInnerNestedRequest_Header `protobuf:"bytes,1,opt,name=header,proto3" json:"header,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ValidInnerNestedRequest) Reset() {
	*x = ValidInnerNestedRequest{}
	mi := &file_testproto_valid_proto_msgTypes[7]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ValidInnerNestedRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ValidInnerNestedRequest) ProtoMessage() {}

func (x *ValidInnerNestedRequest) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_valid_proto_msgTypes[7]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ValidInnerNestedRequest.ProtoReflect.Descriptor instead.
func (*ValidInnerNestedRequest) Descriptor() ([]byte, []int) {
	return file_testproto_valid_proto_rawDescGZIP(), []int{7}
}

func (x *ValidInnerNestedRequest) GetHeader() *ValidInnerNestedRequest_Header {
	if x != nil {
		return x.Header
	}
	return nil
}

type ValidStorageInnerNestedRequest struct {
	state         protoimpl.MessageState                 `protogen:"open.v1"`
	Header        *ValidStorageInnerNestedRequest_Header `protobuf:"bytes,1,opt,name=header,proto3" json:"header,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ValidStorageInnerNestedRequest) Reset() {
	*x = ValidStorageInnerNestedRequest{}
	mi := &file_testproto_valid_proto_msgTypes[8]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ValidStorageInnerNestedRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ValidStorageInnerNestedRequest) ProtoMessage() {}

func (x *ValidStorageInnerNestedRequest) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_valid_proto_msgTypes[8]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ValidStorageInnerNestedRequest.ProtoReflect.Descriptor instead.
func (*ValidStorageInnerNestedRequest) Descriptor() ([]byte, []int) {
	return file_testproto_valid_proto_rawDescGZIP(), []int{8}
}

func (x *ValidStorageInnerNestedRequest) GetHeader() *ValidStorageInnerNestedRequest_Header {
	if x != nil {
		return x.Header
	}
	return nil
}

type ValidInnerNestedRequest_Header struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Destination   *gitalypb.Repository   `protobuf:"bytes,1,opt,name=destination,proto3" json:"destination,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ValidInnerNestedRequest_Header) Reset() {
	*x = ValidInnerNestedRequest_Header{}
	mi := &file_testproto_valid_proto_msgTypes[9]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ValidInnerNestedRequest_Header) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ValidInnerNestedRequest_Header) ProtoMessage() {}

func (x *ValidInnerNestedRequest_Header) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_valid_proto_msgTypes[9]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ValidInnerNestedRequest_Header.ProtoReflect.Descriptor instead.
func (*ValidInnerNestedRequest_Header) Descriptor() ([]byte, []int) {
	return file_testproto_valid_proto_rawDescGZIP(), []int{7, 0}
}

func (x *ValidInnerNestedRequest_Header) GetDestination() *gitalypb.Repository {
	if x != nil {
		return x.Destination
	}
	return nil
}

type ValidStorageInnerNestedRequest_Header struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	StorageName   string                 `protobuf:"bytes,1,opt,name=storage_name,json=storageName,proto3" json:"storage_name,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ValidStorageInnerNestedRequest_Header) Reset() {
	*x = ValidStorageInnerNestedRequest_Header{}
	mi := &file_testproto_valid_proto_msgTypes[10]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ValidStorageInnerNestedRequest_Header) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ValidStorageInnerNestedRequest_Header) ProtoMessage() {}

func (x *ValidStorageInnerNestedRequest_Header) ProtoReflect() protoreflect.Message {
	mi := &file_testproto_valid_proto_msgTypes[10]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ValidStorageInnerNestedRequest_Header.ProtoReflect.Descriptor instead.
func (*ValidStorageInnerNestedRequest_Header) Descriptor() ([]byte, []int) {
	return file_testproto_valid_proto_rawDescGZIP(), []int{8, 0}
}

func (x *ValidStorageInnerNestedRequest_Header) GetStorageName() string {
	if x != nil {
		return x.StorageName
	}
	return ""
}

var File_testproto_valid_proto protoreflect.FileDescriptor

const file_testproto_valid_proto_rawDesc = "" +
	"\n" +
	"\x15testproto/valid.proto\x12\ttestproto\x1a\n" +
	"lint.proto\x1a\fshared.proto\"J\n" +
	"\fValidRequest\x12:\n" +
	"\vdestination\x18\x01 \x01(\v2\x12.gitaly.RepositoryB\x04\x98\xc6,\x01R\vdestination\"\x19\n" +
	"\x17ValidRequestWithoutRepo\">\n" +
	"\x13ValidStorageRequest\x12'\n" +
	"\fstorage_name\x18\x01 \x01(\tB\x04\x88\xc6,\x01R\vstorageName\"\x0f\n" +
	"\rValidResponse\"R\n" +
	"\x12ValidNestedRequest\x12<\n" +
	"\rinner_message\x18\x01 \x01(\v2\x17.testproto.ValidRequestR\finnerMessage\"`\n" +
	"\x19ValidStorageNestedRequest\x12C\n" +
	"\rinner_message\x18\x01 \x01(\v2\x1e.testproto.ValidStorageRequestR\finnerMessage\"b\n" +
	"\x18ValidNestedSharedRequest\x12F\n" +
	"\x12nested_target_repo\x18\x01 \x01(\v2\x12.gitaly.ObjectPoolB\x04\x98\xc6,\x01R\x10nestedTargetRepo\"\xa2\x01\n" +
	"\x17ValidInnerNestedRequest\x12A\n" +
	"\x06header\x18\x01 \x01(\v2).testproto.ValidInnerNestedRequest.HeaderR\x06header\x1aD\n" +
	"\x06Header\x12:\n" +
	"\vdestination\x18\x01 \x01(\v2\x12.gitaly.RepositoryB\x04\x98\xc6,\x01R\vdestination\"\x9d\x01\n" +
	"\x1eValidStorageInnerNestedRequest\x12H\n" +
	"\x06header\x18\x01 \x01(\v20.testproto.ValidStorageInnerNestedRequest.HeaderR\x06header\x1a1\n" +
	"\x06Header\x12'\n" +
	"\fstorage_name\x18\x01 \x01(\tB\x04\x88\xc6,\x01R\vstorageName2[\n" +
	"\x12InterceptedService\x12?\n" +
	"\n" +
	"TestMethod\x12\x17.testproto.ValidRequest\x1a\x18.testproto.ValidResponse\x1a\x04\xf0\x97(\x012\x80\t\n" +
	"\fValidService\x12G\n" +
	"\n" +
	"TestMethod\x12\x17.testproto.ValidRequest\x1a\x18.testproto.ValidResponse\"\x06\xfa\x97(\x02\b\x02\x12H\n" +
	"\vTestMethod2\x12\x17.testproto.ValidRequest\x1a\x18.testproto.ValidResponse\"\x06\xfa\x97(\x02\b\x01\x12H\n" +
	"\vTestMethod3\x12\x17.testproto.ValidRequest\x1a\x18.testproto.ValidResponse\"\x06\xfa\x97(\x02\b\x01\x12N\n" +
	"\vTestMethod5\x12\x1d.testproto.ValidNestedRequest\x1a\x18.testproto.ValidResponse\"\x06\xfa\x97(\x02\b\x01\x12T\n" +
	"\vTestMethod6\x12#.testproto.ValidNestedSharedRequest\x1a\x18.testproto.ValidResponse\"\x06\xfa\x97(\x02\b\x01\x12S\n" +
	"\vTestMethod7\x12\".testproto.ValidInnerNestedRequest\x1a\x18.testproto.ValidResponse\"\x06\xfa\x97(\x02\b\x01\x12Q\n" +
	"\vTestMethod8\x12\x1e.testproto.ValidStorageRequest\x1a\x18.testproto.ValidResponse\"\b\xfa\x97(\x04\b\x01\x10\x02\x12W\n" +
	"\vTestMethod9\x12$.testproto.ValidStorageNestedRequest\x1a\x18.testproto.ValidResponse\"\b\xfa\x97(\x04\b\x01\x10\x02\x12L\n" +
	"\x0fTestMaintenance\x12\x17.testproto.ValidRequest\x1a\x18.testproto.ValidResponse\"\x06\xfa\x97(\x02\b\x03\x12]\n" +
	" TestMaintenanceWithExplicitScope\x12\x17.testproto.ValidRequest\x1a\x18.testproto.ValidResponse\"\x06\xfa\x97(\x02\b\x03\x12c\n" +
	" TestMaintenanceWithNestedRequest\x12\x1d.testproto.ValidNestedRequest\x1a\x18.testproto.ValidResponse\"\x06\xfa\x97(\x02\b\x03\x12o\n" +
	"&TestMaintenanceWithNestedSharedRequest\x12#.testproto.ValidNestedSharedRequest\x1a\x18.testproto.ValidResponse\"\x06\xfa\x97(\x02\b\x03\x12i\n" +
	"!TestMutatorWithInnerNestedRequest\x12\".testproto.ValidInnerNestedRequest\x1a\x18.testproto.ValidResponse\"\x06\xfa\x97(\x02\b\x03B>Z<gitlab.com/gitlab-org/gitaly/v18/proto/go/gitalypb/testprotob\x06proto3"

var (
	file_testproto_valid_proto_rawDescOnce sync.Once
	file_testproto_valid_proto_rawDescData []byte
)

func file_testproto_valid_proto_rawDescGZIP() []byte {
	file_testproto_valid_proto_rawDescOnce.Do(func() {
		file_testproto_valid_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_testproto_valid_proto_rawDesc), len(file_testproto_valid_proto_rawDesc)))
	})
	return file_testproto_valid_proto_rawDescData
}

var file_testproto_valid_proto_msgTypes = make([]protoimpl.MessageInfo, 11)
var file_testproto_valid_proto_goTypes = []any{
	(*ValidRequest)(nil),                          // 0: testproto.ValidRequest
	(*ValidRequestWithoutRepo)(nil),               // 1: testproto.ValidRequestWithoutRepo
	(*ValidStorageRequest)(nil),                   // 2: testproto.ValidStorageRequest
	(*ValidResponse)(nil),                         // 3: testproto.ValidResponse
	(*ValidNestedRequest)(nil),                    // 4: testproto.ValidNestedRequest
	(*ValidStorageNestedRequest)(nil),             // 5: testproto.ValidStorageNestedRequest
	(*ValidNestedSharedRequest)(nil),              // 6: testproto.ValidNestedSharedRequest
	(*ValidInnerNestedRequest)(nil),               // 7: testproto.ValidInnerNestedRequest
	(*ValidStorageInnerNestedRequest)(nil),        // 8: testproto.ValidStorageInnerNestedRequest
	(*ValidInnerNestedRequest_Header)(nil),        // 9: testproto.ValidInnerNestedRequest.Header
	(*ValidStorageInnerNestedRequest_Header)(nil), // 10: testproto.ValidStorageInnerNestedRequest.Header
	(*gitalypb.Repository)(nil),                   // 11: gitaly.Repository
	(*gitalypb.ObjectPool)(nil),                   // 12: gitaly.ObjectPool
}
var file_testproto_valid_proto_depIdxs = []int32{
	11, // 0: testproto.ValidRequest.destination:type_name -> gitaly.Repository
	0,  // 1: testproto.ValidNestedRequest.inner_message:type_name -> testproto.ValidRequest
	2,  // 2: testproto.ValidStorageNestedRequest.inner_message:type_name -> testproto.ValidStorageRequest
	12, // 3: testproto.ValidNestedSharedRequest.nested_target_repo:type_name -> gitaly.ObjectPool
	9,  // 4: testproto.ValidInnerNestedRequest.header:type_name -> testproto.ValidInnerNestedRequest.Header
	10, // 5: testproto.ValidStorageInnerNestedRequest.header:type_name -> testproto.ValidStorageInnerNestedRequest.Header
	11, // 6: testproto.ValidInnerNestedRequest.Header.destination:type_name -> gitaly.Repository
	0,  // 7: testproto.InterceptedService.TestMethod:input_type -> testproto.ValidRequest
	0,  // 8: testproto.ValidService.TestMethod:input_type -> testproto.ValidRequest
	0,  // 9: testproto.ValidService.TestMethod2:input_type -> testproto.ValidRequest
	0,  // 10: testproto.ValidService.TestMethod3:input_type -> testproto.ValidRequest
	4,  // 11: testproto.ValidService.TestMethod5:input_type -> testproto.ValidNestedRequest
	6,  // 12: testproto.ValidService.TestMethod6:input_type -> testproto.ValidNestedSharedRequest
	7,  // 13: testproto.ValidService.TestMethod7:input_type -> testproto.ValidInnerNestedRequest
	2,  // 14: testproto.ValidService.TestMethod8:input_type -> testproto.ValidStorageRequest
	5,  // 15: testproto.ValidService.TestMethod9:input_type -> testproto.ValidStorageNestedRequest
	0,  // 16: testproto.ValidService.TestMaintenance:input_type -> testproto.ValidRequest
	0,  // 17: testproto.ValidService.TestMaintenanceWithExplicitScope:input_type -> testproto.ValidRequest
	4,  // 18: testproto.ValidService.TestMaintenanceWithNestedRequest:input_type -> testproto.ValidNestedRequest
	6,  // 19: testproto.ValidService.TestMaintenanceWithNestedSharedRequest:input_type -> testproto.ValidNestedSharedRequest
	7,  // 20: testproto.ValidService.TestMutatorWithInnerNestedRequest:input_type -> testproto.ValidInnerNestedRequest
	3,  // 21: testproto.InterceptedService.TestMethod:output_type -> testproto.ValidResponse
	3,  // 22: testproto.ValidService.TestMethod:output_type -> testproto.ValidResponse
	3,  // 23: testproto.ValidService.TestMethod2:output_type -> testproto.ValidResponse
	3,  // 24: testproto.ValidService.TestMethod3:output_type -> testproto.ValidResponse
	3,  // 25: testproto.ValidService.TestMethod5:output_type -> testproto.ValidResponse
	3,  // 26: testproto.ValidService.TestMethod6:output_type -> testproto.ValidResponse
	3,  // 27: testproto.ValidService.TestMethod7:output_type -> testproto.ValidResponse
	3,  // 28: testproto.ValidService.TestMethod8:output_type -> testproto.ValidResponse
	3,  // 29: testproto.ValidService.TestMethod9:output_type -> testproto.ValidResponse
	3,  // 30: testproto.ValidService.TestMaintenance:output_type -> testproto.ValidResponse
	3,  // 31: testproto.ValidService.TestMaintenanceWithExplicitScope:output_type -> testproto.ValidResponse
	3,  // 32: testproto.ValidService.TestMaintenanceWithNestedRequest:output_type -> testproto.ValidResponse
	3,  // 33: testproto.ValidService.TestMaintenanceWithNestedSharedRequest:output_type -> testproto.ValidResponse
	3,  // 34: testproto.ValidService.TestMutatorWithInnerNestedRequest:output_type -> testproto.ValidResponse
	21, // [21:35] is the sub-list for method output_type
	7,  // [7:21] is the sub-list for method input_type
	7,  // [7:7] is the sub-list for extension type_name
	7,  // [7:7] is the sub-list for extension extendee
	0,  // [0:7] is the sub-list for field type_name
}

func init() { file_testproto_valid_proto_init() }
func file_testproto_valid_proto_init() {
	if File_testproto_valid_proto != nil {
		return
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_testproto_valid_proto_rawDesc), len(file_testproto_valid_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   11,
			NumExtensions: 0,
			NumServices:   2,
		},
		GoTypes:           file_testproto_valid_proto_goTypes,
		DependencyIndexes: file_testproto_valid_proto_depIdxs,
		MessageInfos:      file_testproto_valid_proto_msgTypes,
	}.Build()
	File_testproto_valid_proto = out.File
	file_testproto_valid_proto_goTypes = nil
	file_testproto_valid_proto_depIdxs = nil
}
