// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v6.30.2
// source: internal.proto

package gitalypb

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	timestamppb "google.golang.org/protobuf/types/known/timestamppb"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

// WalkReposRequest ...
type WalkReposRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// storage_name ...
	StorageName   string `protobuf:"bytes,1,opt,name=storage_name,json=storageName,proto3" json:"storage_name,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *WalkReposRequest) Reset() {
	*x = WalkReposRequest{}
	mi := &file_internal_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *WalkReposRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*WalkReposRequest) ProtoMessage() {}

func (x *WalkReposRequest) ProtoReflect() protoreflect.Message {
	mi := &file_internal_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use WalkReposRequest.ProtoReflect.Descriptor instead.
func (*WalkReposRequest) Descriptor() ([]byte, []int) {
	return file_internal_proto_rawDescGZIP(), []int{0}
}

func (x *WalkReposRequest) GetStorageName() string {
	if x != nil {
		return x.StorageName
	}
	return ""
}

// WalkReposResponse ...
type WalkReposResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// relative_path ...
	RelativePath string `protobuf:"bytes,1,opt,name=relative_path,json=relativePath,proto3" json:"relative_path,omitempty"`
	// modification_time is the modification time of the repository directory.
	// This can be used as a proxy for when the repository was last
	// modified.
	ModificationTime *timestamppb.Timestamp `protobuf:"bytes,2,opt,name=modification_time,json=modificationTime,proto3" json:"modification_time,omitempty"`
	unknownFields    protoimpl.UnknownFields
	sizeCache        protoimpl.SizeCache
}

func (x *WalkReposResponse) Reset() {
	*x = WalkReposResponse{}
	mi := &file_internal_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *WalkReposResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*WalkReposResponse) ProtoMessage() {}

func (x *WalkReposResponse) ProtoReflect() protoreflect.Message {
	mi := &file_internal_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use WalkReposResponse.ProtoReflect.Descriptor instead.
func (*WalkReposResponse) Descriptor() ([]byte, []int) {
	return file_internal_proto_rawDescGZIP(), []int{1}
}

func (x *WalkReposResponse) GetRelativePath() string {
	if x != nil {
		return x.RelativePath
	}
	return ""
}

func (x *WalkReposResponse) GetModificationTime() *timestamppb.Timestamp {
	if x != nil {
		return x.ModificationTime
	}
	return nil
}

// ScanPoolMetadataRequest initiates a scan of a storage for repository-to-pool relationships
// by reading alternates files. It contains the name of the storage to scan.
type ScanPoolMetadataRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// storage_name is the name of the storage to scan.
	StorageName   string `protobuf:"bytes,1,opt,name=storage_name,json=storageName,proto3" json:"storage_name,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ScanPoolMetadataRequest) Reset() {
	*x = ScanPoolMetadataRequest{}
	mi := &file_internal_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ScanPoolMetadataRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ScanPoolMetadataRequest) ProtoMessage() {}

func (x *ScanPoolMetadataRequest) ProtoReflect() protoreflect.Message {
	mi := &file_internal_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ScanPoolMetadataRequest.ProtoReflect.Descriptor instead.
func (*ScanPoolMetadataRequest) Descriptor() ([]byte, []int) {
	return file_internal_proto_rawDescGZIP(), []int{2}
}

func (x *ScanPoolMetadataRequest) GetStorageName() string {
	if x != nil {
		return x.StorageName
	}
	return ""
}

// ScanPoolMetadataResponse contains a repository that is linked to an object pool.
type ScanPoolMetadataResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// relative_path is the relative path of the repository.
	RelativePath string `protobuf:"bytes,1,opt,name=relative_path,json=relativePath,proto3" json:"relative_path,omitempty"`
	// pool_disk_path is the disk path of the object pool the repository is linked to.
	PoolDiskPath string `protobuf:"bytes,2,opt,name=pool_disk_path,json=poolDiskPath,proto3" json:"pool_disk_path,omitempty"`
	// is_upstream indicates if this repository is the upstream/source repository for the pool.
	IsUpstream    bool `protobuf:"varint,3,opt,name=is_upstream,json=isUpstream,proto3" json:"is_upstream,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ScanPoolMetadataResponse) Reset() {
	*x = ScanPoolMetadataResponse{}
	mi := &file_internal_proto_msgTypes[3]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ScanPoolMetadataResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ScanPoolMetadataResponse) ProtoMessage() {}

func (x *ScanPoolMetadataResponse) ProtoReflect() protoreflect.Message {
	mi := &file_internal_proto_msgTypes[3]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ScanPoolMetadataResponse.ProtoReflect.Descriptor instead.
func (*ScanPoolMetadataResponse) Descriptor() ([]byte, []int) {
	return file_internal_proto_rawDescGZIP(), []int{3}
}

func (x *ScanPoolMetadataResponse) GetRelativePath() string {
	if x != nil {
		return x.RelativePath
	}
	return ""
}

func (x *ScanPoolMetadataResponse) GetPoolDiskPath() string {
	if x != nil {
		return x.PoolDiskPath
	}
	return ""
}

func (x *ScanPoolMetadataResponse) GetIsUpstream() bool {
	if x != nil {
		return x.IsUpstream
	}
	return false
}

// StorePoolMetadataRequest contains a single repository-to-pool relationship to store.
type StorePoolMetadataRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// storage_name is the name of the storage.
	StorageName string `protobuf:"bytes,1,opt,name=storage_name,json=storageName,proto3" json:"storage_name,omitempty"`
	// relative_path is the relative path of the repository.
	RelativePath string `protobuf:"bytes,2,opt,name=relative_path,json=relativePath,proto3" json:"relative_path,omitempty"`
	// pool_disk_path is the disk path of the object pool.
	PoolDiskPath string `protobuf:"bytes,3,opt,name=pool_disk_path,json=poolDiskPath,proto3" json:"pool_disk_path,omitempty"`
	// is_upstream indicates if this repository is the upstream/source repository for the pool.
	IsUpstream    bool `protobuf:"varint,4,opt,name=is_upstream,json=isUpstream,proto3" json:"is_upstream,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *StorePoolMetadataRequest) Reset() {
	*x = StorePoolMetadataRequest{}
	mi := &file_internal_proto_msgTypes[4]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *StorePoolMetadataRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*StorePoolMetadataRequest) ProtoMessage() {}

func (x *StorePoolMetadataRequest) ProtoReflect() protoreflect.Message {
	mi := &file_internal_proto_msgTypes[4]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use StorePoolMetadataRequest.ProtoReflect.Descriptor instead.
func (*StorePoolMetadataRequest) Descriptor() ([]byte, []int) {
	return file_internal_proto_rawDescGZIP(), []int{4}
}

func (x *StorePoolMetadataRequest) GetStorageName() string {
	if x != nil {
		return x.StorageName
	}
	return ""
}

func (x *StorePoolMetadataRequest) GetRelativePath() string {
	if x != nil {
		return x.RelativePath
	}
	return ""
}

func (x *StorePoolMetadataRequest) GetPoolDiskPath() string {
	if x != nil {
		return x.PoolDiskPath
	}
	return ""
}

func (x *StorePoolMetadataRequest) GetIsUpstream() bool {
	if x != nil {
		return x.IsUpstream
	}
	return false
}

// StorePoolMetadataResponse contains the result of storing pool metadata.
type StorePoolMetadataResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *StorePoolMetadataResponse) Reset() {
	*x = StorePoolMetadataResponse{}
	mi := &file_internal_proto_msgTypes[5]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *StorePoolMetadataResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*StorePoolMetadataResponse) ProtoMessage() {}

func (x *StorePoolMetadataResponse) ProtoReflect() protoreflect.Message {
	mi := &file_internal_proto_msgTypes[5]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use StorePoolMetadataResponse.ProtoReflect.Descriptor instead.
func (*StorePoolMetadataResponse) Descriptor() ([]byte, []int) {
	return file_internal_proto_rawDescGZIP(), []int{5}
}

// ListPoolMetadataRequest requests a list of pools from the metadata database.
type ListPoolMetadataRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// storage_name is the name of the storage to list pools for.
	StorageName   string `protobuf:"bytes,1,opt,name=storage_name,json=storageName,proto3" json:"storage_name,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ListPoolMetadataRequest) Reset() {
	*x = ListPoolMetadataRequest{}
	mi := &file_internal_proto_msgTypes[6]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ListPoolMetadataRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ListPoolMetadataRequest) ProtoMessage() {}

func (x *ListPoolMetadataRequest) ProtoReflect() protoreflect.Message {
	mi := &file_internal_proto_msgTypes[6]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ListPoolMetadataRequest.ProtoReflect.Descriptor instead.
func (*ListPoolMetadataRequest) Descriptor() ([]byte, []int) {
	return file_internal_proto_rawDescGZIP(), []int{6}
}

func (x *ListPoolMetadataRequest) GetStorageName() string {
	if x != nil {
		return x.StorageName
	}
	return ""
}

// ListPoolMetadataResponse contains a single pool from the metadata database.
type ListPoolMetadataResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// pool_disk_path is the disk path of the object pool.
	PoolDiskPath  string `protobuf:"bytes,1,opt,name=pool_disk_path,json=poolDiskPath,proto3" json:"pool_disk_path,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ListPoolMetadataResponse) Reset() {
	*x = ListPoolMetadataResponse{}
	mi := &file_internal_proto_msgTypes[7]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ListPoolMetadataResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ListPoolMetadataResponse) ProtoMessage() {}

func (x *ListPoolMetadataResponse) ProtoReflect() protoreflect.Message {
	mi := &file_internal_proto_msgTypes[7]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ListPoolMetadataResponse.ProtoReflect.Descriptor instead.
func (*ListPoolMetadataResponse) Descriptor() ([]byte, []int) {
	return file_internal_proto_rawDescGZIP(), []int{7}
}

func (x *ListPoolMetadataResponse) GetPoolDiskPath() string {
	if x != nil {
		return x.PoolDiskPath
	}
	return ""
}

// ListPoolUpstreamsRequest contains the storage name and a list of pool disk paths
// to query for upstream repositories via the Rails ObjectPoolMembers API.
type ListPoolUpstreamsRequest struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// storage_name is the name of the storage (or virtual storage). Set only in the first message
	// of the stream.
	StorageName string `protobuf:"bytes,1,opt,name=storage_name,json=storageName,proto3" json:"storage_name,omitempty"`
	// pool_disk_paths is a list of pool disk paths to look up upstreams for.
	PoolDiskPaths []string `protobuf:"bytes,2,rep,name=pool_disk_paths,json=poolDiskPaths,proto3" json:"pool_disk_paths,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ListPoolUpstreamsRequest) Reset() {
	*x = ListPoolUpstreamsRequest{}
	mi := &file_internal_proto_msgTypes[8]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ListPoolUpstreamsRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ListPoolUpstreamsRequest) ProtoMessage() {}

func (x *ListPoolUpstreamsRequest) ProtoReflect() protoreflect.Message {
	mi := &file_internal_proto_msgTypes[8]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ListPoolUpstreamsRequest.ProtoReflect.Descriptor instead.
func (*ListPoolUpstreamsRequest) Descriptor() ([]byte, []int) {
	return file_internal_proto_rawDescGZIP(), []int{8}
}

func (x *ListPoolUpstreamsRequest) GetStorageName() string {
	if x != nil {
		return x.StorageName
	}
	return ""
}

func (x *ListPoolUpstreamsRequest) GetPoolDiskPaths() []string {
	if x != nil {
		return x.PoolDiskPaths
	}
	return nil
}

// ListPoolUpstreamsResponse contains a mapping of pool disk paths to their
// upstream repository relative paths.
type ListPoolUpstreamsResponse struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// upstreams is a map of pool disk path to the upstream repository's relative path.
	// Pools without a public upstream are omitted from the map.
	Upstreams     map[string]string `protobuf:"bytes,1,rep,name=upstreams,proto3" json:"upstreams,omitempty" protobuf_key:"bytes,1,opt,name=key" protobuf_val:"bytes,2,opt,name=value"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ListPoolUpstreamsResponse) Reset() {
	*x = ListPoolUpstreamsResponse{}
	mi := &file_internal_proto_msgTypes[9]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ListPoolUpstreamsResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ListPoolUpstreamsResponse) ProtoMessage() {}

func (x *ListPoolUpstreamsResponse) ProtoReflect() protoreflect.Message {
	mi := &file_internal_proto_msgTypes[9]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ListPoolUpstreamsResponse.ProtoReflect.Descriptor instead.
func (*ListPoolUpstreamsResponse) Descriptor() ([]byte, []int) {
	return file_internal_proto_rawDescGZIP(), []int{9}
}

func (x *ListPoolUpstreamsResponse) GetUpstreams() map[string]string {
	if x != nil {
		return x.Upstreams
	}
	return nil
}

var File_internal_proto protoreflect.FileDescriptor

const file_internal_proto_rawDesc = "" +
	"\n" +
	"\x0einternal.proto\x12\x06gitaly\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\n" +
	"lint.proto\";\n" +
	"\x10WalkReposRequest\x12'\n" +
	"\fstorage_name\x18\x01 \x01(\tB\x04\x88\xc6,\x01R\vstorageName\"\x81\x01\n" +
	"\x11WalkReposResponse\x12#\n" +
	"\rrelative_path\x18\x01 \x01(\tR\frelativePath\x12G\n" +
	"\x11modification_time\x18\x02 \x01(\v2\x1a.google.protobuf.TimestampR\x10modificationTime\"B\n" +
	"\x17ScanPoolMetadataRequest\x12'\n" +
	"\fstorage_name\x18\x01 \x01(\tB\x04\x88\xc6,\x01R\vstorageName\"\x86\x01\n" +
	"\x18ScanPoolMetadataResponse\x12#\n" +
	"\rrelative_path\x18\x01 \x01(\tR\frelativePath\x12$\n" +
	"\x0epool_disk_path\x18\x02 \x01(\tR\fpoolDiskPath\x12\x1f\n" +
	"\vis_upstream\x18\x03 \x01(\bR\n" +
	"isUpstream\"\xaf\x01\n" +
	"\x18StorePoolMetadataRequest\x12'\n" +
	"\fstorage_name\x18\x01 \x01(\tB\x04\x88\xc6,\x01R\vstorageName\x12#\n" +
	"\rrelative_path\x18\x02 \x01(\tR\frelativePath\x12$\n" +
	"\x0epool_disk_path\x18\x03 \x01(\tR\fpoolDiskPath\x12\x1f\n" +
	"\vis_upstream\x18\x04 \x01(\bR\n" +
	"isUpstream\"\x1b\n" +
	"\x19StorePoolMetadataResponse\"B\n" +
	"\x17ListPoolMetadataRequest\x12'\n" +
	"\fstorage_name\x18\x01 \x01(\tB\x04\x88\xc6,\x01R\vstorageName\"@\n" +
	"\x18ListPoolMetadataResponse\x12$\n" +
	"\x0epool_disk_path\x18\x01 \x01(\tR\fpoolDiskPath\"k\n" +
	"\x18ListPoolUpstreamsRequest\x12'\n" +
	"\fstorage_name\x18\x01 \x01(\tB\x04\x88\xc6,\x01R\vstorageName\x12&\n" +
	"\x0fpool_disk_paths\x18\x02 \x03(\tR\rpoolDiskPaths\"\xa9\x01\n" +
	"\x19ListPoolUpstreamsResponse\x12N\n" +
	"\tupstreams\x18\x01 \x03(\v20.gitaly.ListPoolUpstreamsResponse.UpstreamsEntryR\tupstreams\x1a<\n" +
	"\x0eUpstreamsEntry\x12\x10\n" +
	"\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n" +
	"\x05value\x18\x02 \x01(\tR\x05value:\x028\x012\xf2\x03\n" +
	"\x0eInternalGitaly\x12L\n" +
	"\tWalkRepos\x12\x18.gitaly.WalkReposRequest\x1a\x19.gitaly.WalkReposResponse\"\b\xfa\x97(\x04\b\x02\x10\x020\x01\x12a\n" +
	"\x10ScanPoolMetadata\x12\x1f.gitaly.ScanPoolMetadataRequest\x1a .gitaly.ScanPoolMetadataResponse\"\b\xfa\x97(\x04\b\x02\x10\x020\x01\x12d\n" +
	"\x11StorePoolMetadata\x12 .gitaly.StorePoolMetadataRequest\x1a!.gitaly.StorePoolMetadataResponse\"\b\xfa\x97(\x04\b\x01\x10\x02(\x01\x12a\n" +
	"\x10ListPoolMetadata\x12\x1f.gitaly.ListPoolMetadataRequest\x1a .gitaly.ListPoolMetadataResponse\"\b\xfa\x97(\x04\b\x02\x10\x020\x01\x12f\n" +
	"\x11ListPoolUpstreams\x12 .gitaly.ListPoolUpstreamsRequest\x1a!.gitaly.ListPoolUpstreamsResponse\"\b\xfa\x97(\x04\b\x02\x10\x02(\x010\x01B4Z2gitlab.com/gitlab-org/gitaly/v18/proto/go/gitalypbb\x06proto3"

var (
	file_internal_proto_rawDescOnce sync.Once
	file_internal_proto_rawDescData []byte
)

func file_internal_proto_rawDescGZIP() []byte {
	file_internal_proto_rawDescOnce.Do(func() {
		file_internal_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_internal_proto_rawDesc), len(file_internal_proto_rawDesc)))
	})
	return file_internal_proto_rawDescData
}

var file_internal_proto_msgTypes = make([]protoimpl.MessageInfo, 11)
var file_internal_proto_goTypes = []any{
	(*WalkReposRequest)(nil),          // 0: gitaly.WalkReposRequest
	(*WalkReposResponse)(nil),         // 1: gitaly.WalkReposResponse
	(*ScanPoolMetadataRequest)(nil),   // 2: gitaly.ScanPoolMetadataRequest
	(*ScanPoolMetadataResponse)(nil),  // 3: gitaly.ScanPoolMetadataResponse
	(*StorePoolMetadataRequest)(nil),  // 4: gitaly.StorePoolMetadataRequest
	(*StorePoolMetadataResponse)(nil), // 5: gitaly.StorePoolMetadataResponse
	(*ListPoolMetadataRequest)(nil),   // 6: gitaly.ListPoolMetadataRequest
	(*ListPoolMetadataResponse)(nil),  // 7: gitaly.ListPoolMetadataResponse
	(*ListPoolUpstreamsRequest)(nil),  // 8: gitaly.ListPoolUpstreamsRequest
	(*ListPoolUpstreamsResponse)(nil), // 9: gitaly.ListPoolUpstreamsResponse
	nil,                               // 10: gitaly.ListPoolUpstreamsResponse.UpstreamsEntry
	(*timestamppb.Timestamp)(nil),     // 11: google.protobuf.Timestamp
}
var file_internal_proto_depIdxs = []int32{
	11, // 0: gitaly.WalkReposResponse.modification_time:type_name -> google.protobuf.Timestamp
	10, // 1: gitaly.ListPoolUpstreamsResponse.upstreams:type_name -> gitaly.ListPoolUpstreamsResponse.UpstreamsEntry
	0,  // 2: gitaly.InternalGitaly.WalkRepos:input_type -> gitaly.WalkReposRequest
	2,  // 3: gitaly.InternalGitaly.ScanPoolMetadata:input_type -> gitaly.ScanPoolMetadataRequest
	4,  // 4: gitaly.InternalGitaly.StorePoolMetadata:input_type -> gitaly.StorePoolMetadataRequest
	6,  // 5: gitaly.InternalGitaly.ListPoolMetadata:input_type -> gitaly.ListPoolMetadataRequest
	8,  // 6: gitaly.InternalGitaly.ListPoolUpstreams:input_type -> gitaly.ListPoolUpstreamsRequest
	1,  // 7: gitaly.InternalGitaly.WalkRepos:output_type -> gitaly.WalkReposResponse
	3,  // 8: gitaly.InternalGitaly.ScanPoolMetadata:output_type -> gitaly.ScanPoolMetadataResponse
	5,  // 9: gitaly.InternalGitaly.StorePoolMetadata:output_type -> gitaly.StorePoolMetadataResponse
	7,  // 10: gitaly.InternalGitaly.ListPoolMetadata:output_type -> gitaly.ListPoolMetadataResponse
	9,  // 11: gitaly.InternalGitaly.ListPoolUpstreams:output_type -> gitaly.ListPoolUpstreamsResponse
	7,  // [7:12] is the sub-list for method output_type
	2,  // [2:7] is the sub-list for method input_type
	2,  // [2:2] is the sub-list for extension type_name
	2,  // [2:2] is the sub-list for extension extendee
	0,  // [0:2] is the sub-list for field type_name
}

func init() { file_internal_proto_init() }
func file_internal_proto_init() {
	if File_internal_proto != nil {
		return
	}
	file_lint_proto_init()
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_internal_proto_rawDesc), len(file_internal_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   11,
			NumExtensions: 0,
			NumServices:   1,
		},
		GoTypes:           file_internal_proto_goTypes,
		DependencyIndexes: file_internal_proto_depIdxs,
		MessageInfos:      file_internal_proto_msgTypes,
	}.Build()
	File_internal_proto = out.File
	file_internal_proto_goTypes = nil
	file_internal_proto_depIdxs = nil
}
